/*  combinatorics.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#ifndef COMBINATORICS_H
#define COMBINATORICS_H

#include <vector>
#include <map>
#include <set>
#include <list>
#include <ostream>
#include "functions.h"
#include "yamlconfigurable.h"

namespace YAML {
class Emitter;
class Node;
}

namespace Reduze {

/// A permutation
class Permutation: public YAMLConfigurable {
public:
	static YAMLSpec yaml_spec() {
		YAMLSpec s;
		s.set_keyword("permutation");
		s.set_short_description("A permutation in cycle notation.");
		s.set_long_description(""//
					"A permutation of integers in cycle notation."
					" A permutation is defined as a sequence of sequences"
					" e.g. [[1, 2], [3, 4, 5]]. The identity permutation"
					" is the empty sequence [].");
		return s;
	}
	virtual YAMLSpec yaml_spec_link() const {
		return yaml_spec();
	}

	// constructors

	/// constructs identity permutation
	Permutation() {
	}
	/// constructs a permutation from a vector of cycles
	explicit Permutation(const std::vector<std::vector<int> >& cprod) :
		perm_(cprod) {
		init();
	}
	/// constructs a permutation which transforms 'first' to 'second'
	explicit Permutation(const std::map<int, int>& m);
	/// construct a permutation which transforms 'from' to 'to'
	Permutation(const std::vector<int>& from, const std::vector<int>& to);

	// getter
	const std::vector<std::vector<int> >& vector_rep() const {
		return perm_;
	}

	// invert and compose

	/// returns the inverse prmutation
	Permutation inverse() const;
	/// computes the product: x -> (p1 * p2)(x)  =  x -> p1(p2(x))
	/** gives the permutation that first maps x with p2 and then with p1 */
	static Permutation compose(const Permutation& p1, const Permutation& p2);

	// others

	/// swap the content
	void swap(Permutation& other);
	/// erase the numbers from the the cycles
	void erase(std::set<int> i);
	/// apply the permutation on the integer 'n'
	int apply(int n) const;

	// comparison

	bool is_identity() const {
		return perm_.empty();
	}

	inline bool operator ==(const Permutation& other) const {
		return perm_ == other.perm_;
	}
	inline bool operator !=(const Permutation& other) const {
		return perm_ != other.perm_;
	}
	inline bool operator <(const Permutation& other) const {
		return perm_ < other.perm_;
	}

	// output

	virtual void print(YAML::Emitter&) const;
	virtual void read(const YAML::Node&);
	friend std::ostream& operator<<(std::ostream& os, const Permutation& p);

private:
	/// some checks, calls unique()
	void init();
	/// transformation to disjoint cycles
	void unique();

private:
	std::vector<std::vector<int> > perm_;
};

inline void operator>>(const YAML::Node& n, Permutation& p) {
	p.read(n);
}

inline YAML::Emitter& operator<<(YAML::Emitter& ye, const Permutation& p) {
	p.print(ye);
	return ye;
}

//
//
//


class PermutationSet {
public:
	PermutationSet() {
	}
	explicit PermutationSet(const std::set<Permutation>& p) :
		pset_(p) {
	}

	void insert(const Permutation& p) {
		pset_.insert(p);
	}

	const std::set<Permutation>& pset() const {
		return pset_;
	}

	/// completes the set of permutations to a group (unity erased)
	void complete_permutations();

private:
	/// makes all pairs from this set and the set 's2'
	/** adds then the pairs to this set and defines 's2' as the set with the
	 * 	new pairs (the ones not already in this set) */
	void make_pairs(std::set<Permutation>& s2);

private:
	std::set<Permutation> pset_;
};

/// helper class to get all 'vector<T> A' from 'vector<set<T> > B' where the i-th T in A is from the i-th set of B
template<class T>
class combinations {
public:

	combinations(const std::vector<std::set<T> >& s) :
		good_(true), s_(s.size()), p_(s.size()) {
		if (s.empty())
			good_ = false;
		for (unsigned i = 0; i < s_.size(); ++i) {
			std::list<T> l(s[i].begin(), s[i].end());
			s_[i].swap(l);
			if (s_[i].empty())
				good_ = false;
			p_[i] = s_[i].begin();
		}
	}

	combinations(const std::vector<std::list<T> >& s) :
		good_(true), s_(s.size()), p_(s.size()) {
		if (s.empty())
			good_ = false;
		for (unsigned i = 0; i < s_.size(); ++i) {
			std::list<T> l(s[i].begin(), s[i].end());
			s_[i].swap(l);
			if (s_[i].empty())
				good_ = false;
			p_[i] = s_[i].begin();
		}
	}

	bool get_next(std::vector<T>& res) {
		if (good_) {
			std::vector<T> r;
			r.reserve(s_.size());
			typename std::vector<typename std::list<T>::const_iterator>::const_iterator
					it;
			for (it = p_.begin(); it != p_.end(); ++it)
				r.push_back(*(*it));
			r.swap(res);
			good_ = update(s_.size() - 1);
			return true;;
		}
		return false;
	}

private:
	bool update(int i) {
		if (i < 0)
			return false;
		ASSERT(p_[i] != s_[i].end());
		if (++p_[i] != s_[i].end())
			return true;
		p_[i] = s_[i].begin();
		return update(i - 1);
	}

private:
	bool good_;
	std::vector<std::list<T> > s_;
	std::vector<typename std::list<T>::const_iterator> p_;
};

} // namespace Reduze


/*
 This is the mail archive of the libstdc++@gcc.gnu.org mailing list for the libstdc++ project.
 Index Nav: 	[Date Index] [Subject Index] [Author Index] [Thread Index]
 Message Nav: 	[Date Prev] [Date Next] 	[Thread Prev] [Thread Next]
 Other format: 	[Raw text]
 Algorithm code: Combination opposite to std::permutation

 From: BenBear <cxj26424 at 163 dot com>
 To: libstdc++ at gcc dot gnu dot org
 Date: Thu, 21 Apr 2005 21:41:08 +0800
 Subject: Algorithm code: Combination opposite to std::permutation

 Hi, libstdc++-v3

 adjust_combination ()
 next_combination ()
 prev_combination ()

 I wrote a algorithm of combination, which is opposite to
 std::permutation.  It provides three operations, adjust_combination(),
 next_combination() and prev_combination().  The next_combination() and
 prev_combination() is just like the std::next_permutation() and
 std::perv_permutation().

 Each operation's arguments are three iterators, first, middle and last.
 I treat the range [first, middle) as the elements selected, while the
 range [middle, last) are the elements left.  The range [first, middle)
 is treated in dictionary ordering, like the std::permutation.  [middle,
 last) is maintained in a special ordering.

 A simple example:

 select 2 elements from {1, 2, 3, 4}
 f   m   l              // f: first,   m: middle,    l: last
 1 2, 3 4
 1 3, 4 2
 1 4, 3 2
 2 3, 4 1
 2 4, 3 1
 3 4, 2 1

 The first two number are the elements selected, in dictionary ordering.
 The next combination of {3 4}/{2 1} will be {1 2}/{3 4}.

 Ben Bear
 */

// Combination algorithm implementation

// Copyright (C) 2004, BenBear
//
// This file is an algorithm of the combination. This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License as published by the Free Software
// Foundation; either version 2, or (at your option) any later
// version.

// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#include <algorithm>

namespace benbear {

using namespace std;

template<typename BiIterator>
void sort_combination(BiIterator first, BiIterator last) {
	if (first == last) // no element
		return;

	BiIterator i = first;
	++i;
	if (i == last) // one element
		return;

	int half = distance(first, last) / 2; // half of the length
	BiIterator middle = first;
	advance(middle, half); // middle += half

	sort_combination(first, middle); // sort first part
	sort_combination(middle, last); // sort second part

	inplace_merge(first, middle, last); // merge two parts
}

/// sets up a combination according to the rules
template<typename BiIterator>
void adjust_combination(BiIterator first, BiIterator middle, BiIterator last) {
	// the front part or the back part have no elements
	if ((first == middle) || (middle == last))
		return;

	sort_combination(first, middle);
	sort_combination(middle, last);

	BiIterator b = middle;
	--b;
	BiIterator j = lower_bound(middle, last, *b);
	reverse(j, last);
	reverse(middle, last);
}

/// sets up a combination to its smallest (or largest) element
template<typename BiIterator>
void init_combination(BiIterator first, BiIterator middle, BiIterator last,
		bool min) {
	sort_combination(first, last);

	if (min == false) {
		// the max combination
		reverse(first, last);
		reverse(first, middle);
	}
}

template<typename BiIterator>
bool next_combination(BiIterator first, BiIterator middle, BiIterator last) {
	if ((first == middle) || (middle == last))
		return false;

	// last element of [first, middle)
	BiIterator b = middle;
	--b;

	if (*b < *middle) {
		BiIterator j = b;
		while ((++b != last) && (*j < *b)) {
			iter_swap(j, b);
			j = b;
		}
		return true;
	}

	BiIterator e = last;
	--e;
	while (e != middle) {
		BiIterator k = e;
		--k;
		if (!(*k < *e))
			e = k;
		else
			break;
	}

	BiIterator f = e;
	++f;
	while ((f != last) && !(*f < *e))
		++f;

	if (!(*first < *e)) {
		reverse(first, middle);
		reverse(first, last);
		return false;
	}

	if (*b < *e) {
		BiIterator bb = b;
		while ((++bb != e) && !(*b < *bb))
			;
		reverse(bb, f);
		reverse(b, f);
	} else {
		BiIterator i = b;
		while (!(*--i < *e))
			;

		BiIterator j = last;
		while (!(*i < *--j))
			;

		iter_swap(i, j);
		reverse(++i, middle);
		reverse(i, j);
	}
	return true;
}

template<typename BiIterator>
bool prev_combination(BiIterator first, BiIterator middle, BiIterator last) {
	if ((first == middle) || (middle == last))
		return false;

	BiIterator b = middle;
	--b;

	if (*middle < *b) {
		BiIterator i = upper_bound(first, middle, *middle);
		if (i != b)
			iter_swap(i, middle);
		else {
			BiIterator s = middle;
			while ((++s != last) && !(*s < *middle))
				;
			reverse(b, s);
		}

		return true;
	}

	BiIterator e = last;
	--e;
	while (e != middle) {
		BiIterator k = e;
		--k;
		if (!(*k < *e))
			e = k;
		else
			break;
	}

	BiIterator f = e;
	++f;
	while ((f != last) && !(*f < *e))
		++f;

	if (f == last) {
		reverse(first, last);
		reverse(first, middle);
		return false;
	}

	BiIterator i = upper_bound(first, middle, *f);
	if (i == b) {
		BiIterator s = f;
		while ((++s != last) && !(*s < *f))
			;

		reverse(b, f);
		reverse(b, s);
	} else {
		iter_swap(i, f);
		reverse(++i, f);
		reverse(i, middle);
	}
	return true;
}

} // end of namespace benbear

/*
 // for test:
 #include <iostream>
 #include <cstdlib>
 #include <vector>
 using namespace std;

 template<typename BiIterator>
 void show_array(BiIterator f, BiIterator m, BiIterator l) {
 while (f != m)
 cout << *f++;
 cout << " ";
 while (f != l)
 cout << *f++;
 cout << endl;
 }

 template<typename BiIterator>
 void combi(BiIterator f, BiIterator m, BiIterator l, bool next) {
 benbear::init_combination(f, m, l, next);
 int c = 0;
 cout << endl;
 do {
 show_array(f, m, l);
 c++;
 } while (next ? benbear::next_combination(f, m, l)
 : benbear::prev_combination(f, m, l));
 cout << "The end: ";
 show_array(f, m, l);
 cout << "There's: " << c << endl;
 }

 int main() {
 int a[9] = { 1, 2, 3, 4, 5, 6, 7, 8, 9 };
 bool dir = false;

 for (int i = 1; i < 10; i++) {
 random_shuffle(a, a + 7);
 sort(a, a + 3);
 benbear::adjust_combination(a, a + 3, a + 7);
 show_array(a, a + 3, a + 7);
 }
 system("Pause");

 for (int i = 1, j = 9; i < j; i++) {
 combi(a, a + i, a + j, dir);
 system("Pause");
 }

 sort(a, a + 9);
 benbear::sort_combination(a, a + 7);
 a[1] = 1;
 a[4] = 6;
 //a[6] = 6;
 for (int i = 1, j = 7; i < j; i++) {
 combi(a, a + i, a + j, dir);
 system("Pause");
 }

 int n;
 cout << "The n: ";
 cin >> n;
 vector<int> vec(n);
 cout << "please. input numbers: " << endl;
 for (int i = 0; i < n; i++)
 cin >> vec[i];

 for (int i = 1; i < n; i++) {
 combi(vec.begin(), vec.begin() + i, vec.end(), dir);
 system("Pause");
 }

 system("Pause");
 return 0;
 }
 */

#endif // COMBINATORICS_H
