/*  crossing.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "crossing.h"
#include "files.h"
#include "functions.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "integralfamily.h"
#include "equation.h"
#include "kinematics.h"
#include "sector.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// register type
namespace {
YAMLProxy<Crossing> dummy;
}

Crossing::Crossing(const Kinematics* kin) :
	kinematics_(kin), is_equivalent_to_identity_(true), has_symb2one_in_lhs_(
			false) {
}

// the permutation must transform the legs which are labeled from {1, ... n}
Crossing::Crossing(const Kinematics* kin, const Permutation& perm) :
	kinematics_(kin), permutation_(perm), is_equivalent_to_identity_(true),
			has_symb2one_in_lhs_(false) {

	const unsigned no_pat = subs_options::no_pattern;
	const lst& em = kinematics_->external_momenta();
	const lst& iem = kinematics_->independent_external_momenta();
	exmap em2alt, alt2em;
	lst alt;
	provide_alternative_symbols(em, alt, em2alt, alt2em);
	const exmap& rmc = kinematics_->rule_momentum_conservation();
	const exmap inv2sp = kinematics_->find_rules_invariants_to_sp();
	const exmap sp2inv = kinematics_->rules_sp_to_invariants();
	const lst& in_mom = kinematics_->incoming_momenta();
	const vector<vector<int> >& vec = permutation_.vector_rep();
	const int num_legs = em.nops();

	// verify all numbers *v of the permutation fulfill 1 <= *v <= num_legs
	for (vector<vector<int> >::const_iterator c = vec.begin(); c != vec.end(); ++c)
		for (vector<int>::const_iterator v = c->begin(); v != c->end(); ++v)
			if (*v < 1 || *v > num_legs)
				throw runtime_error("Invalid crossing: invalid leg number "
						+ to_string(*v) + " in " + to_string(permutation_));

	// determine name and the map to transform external momenta and kinematic invariants
	name_ = "";
	lst eqs;
	set<int> visited;
	for (vector<vector<int> >::const_iterator c = vec.begin(); c != vec.end(); ++c) {
		name_ += "x";
		for (vector<int>::const_iterator v = c->begin(); v != c->end();) {
			string vstr = to_string(*v);
			bool large_digit = *v > 10 ? true : false;
			visited.insert(*v);
			bool in1 = in_mom.has(em.op(*v - 1)), in2; // incoming ?
			ex mom1 = em.op(*v - 1).subs(rmc).expand(), mom2;
			++v;
			if (v == c->end()) {
				mom2 = em.op(*c->begin() - 1);
				in2 = in_mom.has(mom2); // incoming ?
			} else {
				mom2 = em.op(*v - 1);
				in2 = in_mom.has(mom2); // incoming ?
				if (large_digit)
					vstr += "p";
			}
			mom2 = mom2.subs(rmc, no_pat).expand();
			int sign = (in1 == in2 ? 1 : -1);
			ex tst = ScalarProduct(mom1, mom1) - ScalarProduct(mom2, mom2);
			tst = tst.subs(rmc, no_pat).expand().subs(sp2inv, no_pat).expand();
			if (!tst.is_zero())
				throw runtime_error(
						"Invalid crossing: external legs with different masses found: "
								+ to_string(permutation_));
			eqs.append((mom1.subs(em2alt, no_pat)) == sign * mom2);
			name_ += vstr;
		}
	}
	// append also the momenta which are not permuted
	for (int i = 1; i <= num_legs; ++i)
		if (visited.find(i) == visited.end()) {
			ex mom = em.op(i - 1).subs(rmc).expand();
			eqs.append((mom.subs(em2alt, no_pat)) == mom);
		}

	lst altiem = ex_to<lst> (ex(iem).subs(em2alt));
	eqs = ex_to<lst> (lsolve(eqs, altiem));
	rules_momenta_ = equations_to_substitutions(eqs, alt2em, true);
	ASSERT(permutation_.is_identity() || !rules_momenta_.empty());

	const ex inv1 = kinematics_->kinematic_invariants();
	const ex inv2 = //
					inv1.subs(inv2sp, no_pat).subs(rules_momenta_, no_pat).expand().subs(
							sp2inv, no_pat).expand();
	ASSERT(inv1.nops() == inv2.nops());
	for (unsigned i = 0; i < inv1.nops(); ++i) {
		ex i1 = inv1.op(i);
		ex i2 = inv2.op(i);
		ASSERT(is_a<symbol>(i1));
		if (i1.is_equal(i2))
			continue;
		rules_invariants_[i1] = i2;
	}

	// now verify derived rules
	// (might be wrong if numbers for invariants are used in user input)
	for (exmap::const_iterator r = sp2inv.begin(); r != sp2inv.end(); ++r) {
		ex lhsprime = r->first.subs(rules_momenta_, no_pat);
		ex rhsprime = r->second.subs(rules_invariants_, no_pat);
		ex diff = (lhsprime - rhsprime).expand().subs(sp2inv, no_pat);
		if (!diff.expand().is_zero())
			throw runtime_error(
					string("Failed to generate crossing:\n") +
					"Can't map crossing " + to_string(permutation_) +
					" to transformations of momenta and invariants.\n" +
					"This error is expected if numbers are used for kinematic"
					" invariants.\n"
					"Please try to disable generate_crossings in kinematics and"
					" avoid graph based jobs.");
	}

	// does the symbol to replace by one appear on lhs ?

	const symbol* symb2one = kinematics_->symbol_to_replace_by_one();
	if (symb2one)
		for (exmap::const_iterator r = rules_invariants_.begin(); r
				!= rules_invariants_.end(); ++r)
			if (r->first.is_equal(*symb2one)) {
				has_symb2one_in_lhs_ = true;
				break;
			}
	// is equivalent to identity crossing ?
	ex thisinvs = inv1.subs(rules_invariants_, no_pat);
	is_equivalent_to_identity_ = (thisinvs - inv1).expand().is_zero();
}

Crossing Crossing::inverse() const {
	return Crossing(kinematics_, permutation_.inverse());
}

Crossing Crossing::compose(const Crossing& c1, const Crossing& c2) {
	// \todo assert both kinematics are the same
	return Crossing(c1.kinematics_, Permutation::compose(c1.permutation_,
			c2.permutation_));
}

std::string Crossing::name_for_crossed_family(const IntegralFamily* ic) const {
	if (!ic->is_crossed())
		return ic->name() + name();
	ASSERT(dynamic_cast<const CrossedIntegralFamily*>(ic) != 0);
	const CrossedIntegralFamily* derived =
			static_cast<const CrossedIntegralFamily*> (ic);
	string crossing = Crossing::compose(*this, derived->crossing()).name();
	return ic->source_integralfamily()->name() + crossing;
}

const IntegralFamily* Crossing::transform(const IntegralFamily* f) const {
	string name = name_for_crossed_family(f);
	const IntegralFamily* ic = Files::instance()->integralfamily(name);
	return ic;
}

Sector Crossing::transform(const Sector& s) const {
	string name = name_for_crossed_family(s.integralfamily());
	const IntegralFamily* ic = Files::instance()->integralfamily(name);
	return Sector(ic, s.id());
}

INT Crossing::transform(const INT& i) const {
	string name = name_for_crossed_family(i.integralfamily());
	const IntegralFamily* ic = Files::instance()->integralfamily(name);
	return INT(ic, i.v());
}

LinearCombination Crossing::transform(const LinearCombination& id,
		bool transform_to_minimal_equiv) const {
	LinearCombination newid;
	for (LinearCombination::const_iterator t = id.begin(); t != id.end(); ++t) {
		INT i = transform(t->first);
		if (transform_to_minimal_equiv)
			i = Crossing::transform_to_minimal_equivalent(i);
		newid.insert(i, t->second.subs(rules_invariants_));
	}
	return newid;
}

Identity Crossing::transform(const Identity& id,
		bool transform_to_minimal_equiv) const {
	Identity newid;
	Identity::const_iterator t;
	const exmap& s2one = kinematics_->get_rule_symbol_to_replace_by_one();

	if (has_symb2one_in_lhs_) { // lhs of a rule has symb to replace by one
		Identity reconstruct = id;
		reconstruct.reconstruct_symbol_replaced_by_one();
		for (t = reconstruct.begin(); t != reconstruct.end(); ++t) {
			INT i = transform(t->first);
			if (transform_to_minimal_equiv)
				i = Crossing::transform_to_minimal_equivalent(i);
			newid.insert(i, t->second.subs(rules_invariants_).subs(s2one));
		}
	} else {
		for (t = id.begin(); t != id.end(); ++t) {
			INT i = transform(t->first);
			if (transform_to_minimal_equiv)
				i = Crossing::transform_to_minimal_equivalent(i);
			newid.insert(i, t->second.subs(rules_invariants_).subs(s2one));
		}
	}
	return newid;
}

GiNaC::ex Crossing::transform(const GiNaC::ex& e) const {
	return e.subs(rules_invariants_).subs(rules_momenta_);
}

pair<INT, Crossing> Crossing::uncross(const INT& integral) {
	const IntegralFamily* ic = integral.integralfamily();
	const CrossedIntegralFamily* cic =
			dynamic_cast<const CrossedIntegralFamily*> (ic);
	if (cic == 0) {
		return make_pair(integral, Crossing(ic->kinematics()));
	} else {
		INT sourceintegral(cic->source_integralfamily(), integral.v());
		return make_pair(sourceintegral, cic->crossing());
	}
}

INT Crossing::transform_to_minimal_equivalent(const INT& i) {
	if (!i.integralfamily()->is_crossed())
		return i;
	// the family of the integral (is crossed)
	const CrossedIntegralFamily * ic =
			static_cast<const CrossedIntegralFamily*> (i.integralfamily());
	// the source integral family
	const IntegralFamily* source = ic->source_integralfamily();
	Crossing c = ic->kinematics()->equivalent_crossings().at(ic->crossing());
	string name = c.name_for_crossed_family(source);
	return INT(Files::instance()->integralfamily(name), i.v());
}

bool Crossing::is_equivalent(const Crossing& other) const {
	const lst invs = kinematics_->kinematic_invariants();
	ex thisinvs = invs.subs(rules_invariants_);
	ex otherinvs = invs.subs(other.rules_invariants_);
	return (thisinvs - otherinvs).expand().is_zero();
}

Sector Crossing::to_minimal_crossing(const Sector& s) {
	if (!s.integralfamily()->is_crossed())
		return s;
	const map<Crossing, Crossing>& crossmap =
			s.integralfamily()->kinematics()->equivalent_crossings();
	pair<INT, Crossing> unx = Crossing::uncross(INT(s));
	const Sector snox = unx.first.get_sector();
	map<Crossing, Crossing>::const_iterator min = crossmap.find(unx.second);
	VERIFY(min != crossmap.end());
	return min->second.transform(snox);
}

INT Crossing::to_minimal_crossing(const INT& i) {
	if (!i.integralfamily()->is_crossed())
		return i;
	Sector smin = to_minimal_crossing(i.get_sector());
	if (smin == i.get_sector())
		return i;
	return INT(smin.integralfamily(), i.v());
}

void Crossing::print(YAML::Emitter& ye) const {
	using namespace YAML;
	ye << BeginMap;
	ye << Key << "permutation" << Value << permutation_;
	if (kinematics_) {
		ye << Key << "name" << Value << name_;
		ye << Key << "rules_momenta" << Value << Flow << rules_momenta_;
		ye << Key << "rules_invariants" << Value << Flow << rules_invariants_;
	}
	ye << EndMap;
}

void Crossing::read(const YAML::Node& node) {
	Permutation p;
	node["permutation"] >> p;
	VERIFY(kinematics_ != 0);
	*this = Crossing(kinematics_, p);
}

bool Crossing::operator<(const Crossing& other) const {
	if (is_equivalent_to_identity_ != other.is_equivalent_to_identity_)
		return is_equivalent_to_identity_;
	return permutation() < other.permutation();
}
bool Crossing::operator==(const Crossing& other) const {
	return permutation_ == other.permutation_;
}
bool Crossing::operator!=(const Crossing& other) const {
	return permutation_ != other.permutation_;
}

}
