/*  files.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#ifndef FILES_H_
#define FILES_H_

#include <string>
#include <map>
#include <set>
#include <list>
#include "yamlconfigurable.h"

namespace Reduze {

class Kinematics;
class IntegralFamily;
class SectorMappings;
class SectorReduzer;
class Job;
class JobQueue;
class Sector;
class OutFileData;
class FeynmanRules;
class FeynmanRulesGlobals;
class Paths;

/// Manager for storing objects to files and cached reading from files
class Files {
public:
	virtual ~Files();
	static Files* instance();

	void set_project_directory(const std::string& dir_name);
	const std::string& project_directory() const {
		return project_directory_;
	}
	// To these directories jobs may freely write, all other files
	// should be accessed via the get_filename*() names. Output should
	// always be written to a temporary file first, then renamed to the
	// final filename.
	std::string get_tmp_directory() const;
	std::string get_log_directory() const;
	std::string get_reductions_directory() const;
	std::string get_graph_directory() const;
	std::string get_graph_directory(const IntegralFamily* ic) const;
	std::string get_config_directory() const;
	std::string sectormappings_directory() const;

	// use these methods only if necessary, might be removed some day
	Kinematics* kinematics();
	Paths* paths();
	FeynmanRulesGlobals* feynmanrulesglobals();
	FeynmanRules* feynmanrules();
	IntegralFamily* integralfamily(const std::string& name);
	/// returns product IntegralFamily
	const IntegralFamily* integralfamily(const IntegralFamily* ic1,
			const IntegralFamily* ic2);
	std::list<IntegralFamily*> integralfamilies();
	SectorMappings* sectormappings(const std::string& name);
	bool is_empty_sectormappings_file(const std::string& name);

	std::string get_filename_paths() const;
	std::string get_filename_kinematics() const;
	std::string get_filename_feynmanrules() const;
	std::string get_filename_sectormappings(const std::string& name);
	std::string get_filename_sectorreduction(const Sector& sector);
	std::string get_filename_sectorreduction(const Sector& sector,
			const std::string& dir);

	std::string get_safe_log_filename_for_job(const Job* const job) const;

	void read(JobQueue& j, const std::string& filename);

	void save(const SectorMappings& m);

	/// clears cached sector mappings
	void clear_sectormappings();
	/// clears cached sector mappings and crossed sector mappings of family 'name'
	void clear_sectormappings(const std::string& name);
	/// verifies if the two files contain the same sector mappings for the input family
	/** symmetry shifts are not tested if they contain exactly the same shifts **/
	static void verify_equivalent_sector_mapping_files(
			const std::string& intfam_name, const std::string& secmap_file1,
			const std::string& secmap_file2);
	/// returns the name of the integral family if the file is a sector mapping file
	/** on failure an empty string is returned */
	static std::string family_in_sector_mapping_file(const std::string& filename);


private:
	Files();
	void load_integralfamilies();
	bool intfam_loaded_;
	std::string project_directory_;
	Kinematics* kinematics_;
	Paths* paths_;
	FeynmanRulesGlobals* feynmanrulesglobals_;
	FeynmanRules* feynmanrules_;
	std::map<std::string, IntegralFamily*> integralfamilies_;
	std::list<IntegralFamily*> ordered_integralfamilies_; // by id
	std::map<std::string, SectorMappings*> sectormappings_;
};

/// returns directory part of filename with "/" appended if not already present
std::string get_directory_of_filename(const std::string& filename);
/// returns filename with its directory stripped
std::string get_filename_without_directory(const std::string& filename);
/// returns filename with the given suffix stripped
std::string get_filename_without_suffix(const std::string& filename,
		const std::string& suffix);
/// returns filename with the given suffix stripped
std::string get_filename_without_suffix(const std::string& filename);
/// returns canonical form of filename, prepending "./" for relative paths
std::string get_canonical_filename(const std::string& filename);
/// returns a short version of the filename
std::string short_filename(const std::string& str);
/// returns the string with potentially dangerous special characters substituted
std::string safe_filename(const std::string& str);
/// concatenates directory and filename, inserting '/' in between if necessary
std::string cat_directory_and_filename(const std::string& dir,
		const std::string& filename);

/// returns a directory name with suffix .tmp/
std::string get_temporary_dir_name(const std::string& directory);
/// create a directory
void make_directory(const std::string& dirname);
/// tries to open the file and returns if this was successful
bool is_readable_file(const std::string& filename);
bool are_readable_files(const std::list<std::string>& filenames);
bool is_readable_directory(const std::string& path);

// operations of files
void rename(const std::string& fn1, const std::string& fn2);
void remove(const std::string& fn);
void remove(const std::list<std::string>& filenames);
void remove_empty_dir(const std::string& dir);
void remove_empty_dir(const std::list<std::string>& dir);

/// get the files in the directors
std::list<std::string> get_files_from_directory(const std::string& dirname);
/// removes all files in given directory and the directory itself
/** does not touch symbolic links, does not touch subdirectories **/
bool remove_directory_with_files(const std::string& dirname);
/// copy the content of file 'from' to file 'to'
void copy_file(const std::string& from, const std::string& to);

/// writes some settings like integral family definitions to output stream
void write_header(OutFileData& of);

/// paths to external programs
class Paths: public YAMLConfigurable {
public:
	static YAMLSpec yaml_spec() {
		YAMLSpec s;
		s.set_keyword("paths");
		s.set_short_description("Declaration of paths.");
		s.set_long_description(s.short_description());

		s.add_option("fermat", false, "string", ""//
					"The path to the Fermat executable. Currently,"
					" Fermat support is limited to handling reduction equations"
					" and does not cover advanced linear combinations with"
					" functions.");

		return s;
	}
	virtual YAMLSpec yaml_spec_link() const {
		return yaml_spec();
	}

	Paths() {
	}
	virtual ~Paths() {
	}
	virtual void read(const YAML::Node& node);
	virtual void print(YAML::Emitter& os) const;

	// getters
	const std::string& fermat() const;

private:
	std::string fermat_;
};

inline void operator>>(const YAML::Node& n, Paths& r) {
	r.read(n);
}
inline YAML::Emitter& operator<<(YAML::Emitter& ye, const Paths& r) {
	r.print(ye);
	return ye;
}

} // namespace Reduze

#endif /* FILES_H_ */
