/*  graph.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include <ginac/ginac.h>
#include "functions.h"
#include "graph.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "files.h"
#include "triconnected.h"
#include "kinematics.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

std::ostream& operator<<(std::ostream& os, const GraphEdge& edge) {
	os << "(" << edge.momentum_ << ", " << edge.squaredmass_ << ")";
	return os;
}

struct mass_is_less {
	bool operator()(const GiNaC::ex& e1, const GiNaC::ex& e2) {
		if (is_a < numeric > (e1) && is_a < numeric > (e2))
			return ex_to < numeric > (e1) < ex_to < numeric > (e2);
		if (is_a < numeric > (e1) && !is_a < numeric > (e2))
			return true;
		if (!is_a < numeric > (e1) && is_a < numeric > (e2))
			return false;
		ostringstream s1, s2;
		s1 << e1;
		s2 << e2;
		return s1.str() < s2.str();
	}
};

bool GraphLabel::operator<(const GraphLabel& other) const {
	if (masses_.nops() != other.masses_.nops())
		return masses_.nops() < other.masses_.nops();

	lst::const_iterator l1 = masses_.begin(), l2 = other.masses_.begin();
	for (; l1 != masses_.end() || l2 != other.masses_.end(); ++l1, ++l2)
		if (!l1->is_equal(*l2)) {
			mass_is_less mil;
			return mil(*l1, *l2);
		}

	return edge_label_ < other.edge_label_;
}

bool GraphLabel::operator==(const GraphLabel& other) const {
	if (masses_.nops() != other.masses_.nops())
		return false;
	lst::const_iterator l1 = masses_.begin(), l2 = other.masses_.begin();
	for (; l1 != masses_.end() || l2 != other.masses_.end(); ++l1, ++l2)
		if (!l1->is_equal(*l2))
			return false;

	return edge_label_ == other.edge_label_;
}

bool GraphLabel::operator!=(const GraphLabel& other) const {
	return !(*this == other);
}

std::ostream& operator<<(std::ostream& os, const GraphLabel& g) {
	os << g.masses_ << "\n";
	os << g.edge_label_;
	return os;
}
YAML::Emitter& operator<<(YAML::Emitter& ye, const GraphLabel& label) {
	ye << YAML::Flow << YAML::BeginSeq;
	ye << label.edge_label_;
	ye << label.masses_;
	ye << YAML::EndSeq;
	return ye;
}
void GraphLabel::read(const YAML::Node& n, const GiNaC::lst& symbs) {
	if (n.GetType() != YAML::CT_SEQUENCE || n.size() != 2)
		throw runtime_error(
				"expected a 2-element sequence " + position_info(n));
	n[0] >> edge_label_;
	Reduze::read<GiNaC::ex>(n[1], masses_, symbs);
}

//
// graph
//

Graph::Graph() :
		undefined(GiNaC::symbol("*")) {
}

Graph::Graph(const Topology& t) :
		Topology(t), undefined(GiNaC::symbol("*")) {
}

Graph::~Graph() {
}

void Graph::swap(Graph& other) {
	// swap topologies, enforce no call back which would clear Graph members
	static_cast<Topology&>(*this).swap(static_cast<Topology&>(other), false);
	subgraph_to_sector_.swap(other.subgraph_to_sector_);
	graph_to_sector_.swap(other.graph_to_sector_);
	graphedges_.swap(other.graphedges_);
	std::swap(undefined, other.undefined);
	std::swap(loop_momenta_, other.loop_momenta_);
}

const GraphEdge& Graph::graphedge(int edge_id) const {
	map<int, GraphEdge>::const_iterator it = graphedges_.find(edge_id);
	if (it == graphedges_.end())
		ABORT("Edge " << edge_id << " does not exist");
	return it->second;
}

// momenta

void Graph::set_loop_momenta(const GiNaC::lst& symbols) {
	loop_momenta_ = symbols;
}

const GiNaC::lst& Graph::loop_momenta() const {
	return loop_momenta_;
}

void Graph::clear_momentum(int edge) {
	set_momentum(edge, undefined);
}

void Graph::clear_internal_momenta() {
	for (map<int, GraphEdge>::iterator it = graphedges_.begin();
			it != graphedges_.end(); ++it)
		if (!is_external_edge(it->first))
			clear_momentum(it->first);
}

void Graph::set_momentum(int edge_id, const GiNaC::ex& mom) {
	if (!contains_edge(edge_id))
		ABORT("Invalid edge " << edge_id);
	map<int, GraphEdge>::iterator it = graphedges_.find(edge_id);
	if (it == graphedges_.end()) {
		GraphEdge ge(undefined, undefined);
		it = graphedges_.insert(make_pair(edge_id, ge)).first;
	}
	GraphEdge& et = it->second;

	// set the momentum but restore old value if momentum conservation violated
	GiNaC::ex old_momentum = et.momentum_;
	et.momentum_ = mom;

	const Edge& e = edge(edge_id);
	//const Edge& e = edges_[edge_id];
	// no restrictions on external nodes, but internal nodes must conserve mom.
	/// \todo: ensure total external momentum is zero for each connected comp.
	if ((!is_external_node(e.from) && !is_momentum_conserved(e.from)) || //
			(!is_external_node(e.to) && !is_momentum_conserved(e.to))) {
		et.momentum_ = old_momentum;
		throw runtime_error("Momentum violation in graph " + name());
	}
}

void Graph::complete_momentum(int node_id, bool recurse) {
	if (is_external_node(node_id))
		return;
	GiNaC::ex sum = 0;
	list<int> undef_edges;
	const list<Edge>& edges = edges_of_node(node_id);
	for (list<Edge>::const_iterator e = edges.begin(); e != edges.end(); ++e) {
		std::map<int, GraphEdge>::iterator ge = graphedges_.find(e->id);
		if (e->from == e->to) { // tadpole edge
			continue; // just ignore
		} else if (ge == graphedges_.end()
				|| ge->second.momentum_ == undefined) {
			if (!undef_edges.empty()) // more than one undefined momentum
				return;
			undef_edges.push_back(e->id);
		} else {
			if (e->from == node_id) // outgoing
				sum -= ge->second.momentum_;
			if (e->to == node_id) // incoming
				sum += ge->second.momentum_;

		}
	}
	if (undef_edges.empty())
		return; // nothing to do
	Edge undef = edge(undef_edges.front());
	// undef is either incoming or outgoing, no tadpole at this point
	bool outgoing = (undef.from == node_id);
	set_momentum(undef.id, (outgoing ? 1 : -1) * sum);
	if (recurse) {
		if (undef.from != node_id)
			complete_momentum(undef.from, true);
		if (undef.to != node_id)
			complete_momentum(undef.to, true);
	}
}

void Graph::complete_momenta() {
	for (set<int>::const_iterator n = nodes().begin(); n != nodes().end(); ++n)
		complete_momentum(*n, true);
}

GiNaC::ex Graph::get_momentum(int edge_id) const {
	if (!contains_edge(edge_id))
		ABORT("Invalid edge " << edge_id);
	map<int, GraphEdge>::const_iterator it = graphedges_.find(edge_id);
	return it == graphedges_.end() ? undefined : it->second.momentum_;
}

std::set<int> Graph::get_edges_undefined_momentum() const {
	set<int> result;
	const map<int, Edge>& es = this->edges();
	for (map<int, Edge>::const_iterator e = es.begin(); e != es.end(); ++e)
		if (graphedges_.find(e->first) == graphedges_.end()
				|| graphedges_.find(e->first)->second.momentum_ == undefined)
			result.insert(e->first);
	return result;
}

std::vector<Propagator> Graph::get_propagators_with_loop_momenta() const {
	vector<Propagator> props;
	props.reserve(edges().size());
	for (map<int, GraphEdge>::const_iterator e = graphedges_.begin();
			e != graphedges_.end(); ++e) {
		if (freeof(e->second.momentum_, loop_momenta_))
			continue;
		Propagator p(e->second.momentum_, e->second.squaredmass_);
		props.push_back(p);
	}
	return props;
}

bool Graph::is_momentum_conserved(int node) const {
	GiNaC::ex sum = 0;
	list < Edge > edges = edges_of_node(node);
	for (list<Edge>::const_iterator e = edges.begin(); e != edges.end(); ++e) {
		if (graphedges_.find(e->id) == graphedges_.end())
			return true;
		const GraphEdge& et = graphedges_.find(e->id)->second;
		if (et.momentum_ == undefined) // unassigned mom.
			return true;
		// note: an edge may be both, incoming and outgoing
		if (e->from == node) // outgoing edge
			sum -= et.momentum_;
		if (e->to == node) // incoming edge
			sum += et.momentum_;
	}
	return sum.expand().is_zero();
}

// squared mass

void Graph::set_squaredmass(int edge_id, const GiNaC::ex& squaredmass) {
	if (!contains_edge(edge_id))
		ABORT("Invalid edge " << edge_id);
	map<int, GraphEdge>::iterator it = graphedges_.find(edge_id);
	if (it == graphedges_.end()) {
		GraphEdge ge(undefined, undefined);
		it = graphedges_.insert(make_pair(edge_id, ge)).first;
	}
	it->second.squaredmass_ = squaredmass;
}

void Graph::clear_squaredmasses() {
	for (map<int, GraphEdge>::iterator it = graphedges_.begin();
			it != graphedges_.end(); ++it)
		if (!is_external_edge(it->first))
			it->second.squaredmass_ = undefined;
}

GiNaC::ex Graph::get_squaredmass(int edge_id) const {
	if (!contains_edge(edge_id))
		ABORT("Invalid edge " << edge_id);
	map<int, GraphEdge>::const_iterator it = graphedges_.find(edge_id);
	return it == graphedges_.end() ? undefined : it->second.squaredmass_;
}

bool Graph::is_undefined_square_mass(int edge_id) const {
	ex ms = get_squaredmass(edge_id);
	if (ms.is_equal(undefined))
		return true;
	return false;
}

GiNaC::lst Graph::get_different_masses() const {
	// verify all masses set
	set<ex, mass_is_less> ms;
	map<int, GraphEdge>::const_iterator e;
	for (e = graphedges_.begin(); e != graphedges_.end(); ++e) {
		if (e->second.squaredmass_.is_equal(undefined))
			throw runtime_error(
					"graph " + name() + " has undefined square masses");
		ms.insert(e->second.squaredmass_);
	}
	lst masses;
	for (set<ex, mass_is_less>::const_iterator m = ms.begin(); m != ms.end();
			++m)
		masses.append(*m);
	return masses;
}

// graph construction

void Graph::insert_graph_edge_to_node(const GraphEdge& graphedge, int node,
		int edge_id, std::list<Graph>& result, bool stop_on_first_match) const {
	if (is_external_node(node)) // no cleaving of external vertices
		return;

	list < Edge > edges = edges_of_node(node);
	if (edges.size() < 4) // split vertex only if there are more than 3 edges attached
		return;
	// cannot treat self loops
	for (list<Edge>::const_iterator e = edges.begin(); e != edges.end(); ++e)
		if (e->from == e->to)
			ABORT("self loop encountered in diagram " << name());

	int max_iteration = edges.size() / 2;
	for (int iter = 2; iter <= max_iteration; ++iter) {
		edges.sort();
		list<Edge>::iterator middle = edges.begin();
		advance(middle, iter);
		do {
			ex mom1(0), mom2(0);
			set<Edge> move;
			list<Edge>::iterator e = edges.begin();
			for (; e != middle; ++e) {
				int sign = e->to == node ? 1 : -1;
				mom1 += sign * get_momentum(e->id);
				move.insert(*e);
			}
			for (; e != edges.end(); ++e) {
				int sign = e->to == node ? 1 : -1;
				mom2 += sign * get_momentum(e->id);
			}
			ASSERT((mom1 + mom2).expand().is_zero());
			int sign(0);
			if ((graphedge.momentum_ + mom2).expand().is_zero())
				sign = 1;
			else if ((-graphedge.momentum_ + mom2).expand().is_zero())
				sign = -1;
			if (sign != 0) {
				Graph graph(*this);
				pair<int, int> new_node_edge = graph.cleave(node, move, sign,
						edge_id);
				graph.set_momentum(new_node_edge.second, graphedge.momentum_);
				graph.set_squaredmass(new_node_edge.second,
						graphedge.squaredmass_);
				result.push_back(graph);
				if (stop_on_first_match)
					return;
			}
		} while (benbear::next_combination(edges.begin(), middle, edges.end()));
	}
}

void Graph::insert_graph_edge(const GraphEdge& graphedge, int edge_id,
		std::list<Graph>& result) const {
	const set<int>& nodes = this->nodes();
	for (set<int>::const_iterator v = nodes.begin(); v != nodes.end(); ++v) {
		list<Graph> tmp;
		insert_graph_edge_to_node(graphedge, *v, edge_id, tmp, false);
		result.splice(result.end(), tmp);
	}
}

bool Graph::insert_graph_edge(const GraphEdge& graphedge, int edge_id) {
	const set<int>& nodes = this->nodes();
	for (set<int>::const_iterator v = nodes.begin(); v != nodes.end(); ++v) {
		list<Graph> graphs;
		insert_graph_edge_to_node(graphedge, *v, edge_id, graphs, true);
		if (!graphs.empty()) {
			graphs.back().swap(*this);
			return true;
		}
	}
	return false;
}

void Graph::insert_graph_edges(
		const std::vector<std::pair<int, GraphEdge> >& graphedges) {
	if (graphedges.empty())
		return;
	int num_edges = num_internal_edges();

	vector<int> pints(graphedges.size(), 0);
	for (unsigned int i = 1; i < graphedges.size(); i++)
		pints[i] = i;

	stack<Graph> backups;
	vector<int>::iterator p = pints.begin();
	backups.push(*this);
	while (p != pints.end()) {
		Graph work = backups.top(); // copy
		if (work.insert_graph_edge(graphedges[*p].second,
				graphedges[*p].first)) {
			backups.push(work);
			++p;
		} else if (p == pints.begin()) {
			break;
		} else {
			sort(p, pints.end(), greater<int>());
			do {
				--p;
				backups.pop();
			} while (*p > *(p + 1) && p != pints.begin());
			if (!next_permutation(pints.begin(), pints.end()))
				break;
		}
	}
	if (p == pints.end()) {
		*this = backups.top();
		ASSERT(num_edges + graphedges.size() == num_internal_edges());
		return;
	}
	throw runtime_error("failed to insert propagators to graph " + name());
}

// get the minimal graph with w.r.t a canonical label calculated without permuting ext. nodes
Graph get_minimal_graph_fixed_ext_nodes(const std::list<Graph>& graphs) {
	list<Graph>::const_iterator gmin = graphs.end();
	GraphLabel label, lmin;
	for (list<Graph>::const_iterator g = graphs.begin(); g != graphs.end();
			++g) {
		label = g->find_canonical_graph_label(false).first;
		if (g == graphs.begin() || label < lmin) {
			lmin = label;
			gmin = g;
		}
	}
	VERIFY(gmin != graphs.end());
	return *gmin;
}

void Graph::insert_graph_edges_unique(
		const std::vector<std::pair<int, GraphEdge> >& graphedges) {
	if (graphedges.empty()) {
		remove_degree_2_nodes();
		disconnect_vacuum_components();
		return;
	}
	//int num_edges = num_internal_edges();

	list<Graph> result;
	result.push_back(*this);
	vector<pair<int, GraphEdge> >::const_iterator e;
	for (e = graphedges.begin(); e != graphedges.end(); ++e) {
		list<Graph> tmp;
		for (list<Graph>::iterator g = result.begin(); g != result.end(); ++g)
			g->insert_graph_edge(e->second, e->first, tmp);
		if (tmp.empty())
			throw runtime_error(
					"failed to insert propagators to graph " + name());
		else
			result.swap(tmp);
	}
	// find the minimal graphs with permutations of external nodes
	GraphLabel min;
	list<Graph> minimals;
	for (list<Graph>::iterator g = result.begin(); g != result.end(); ++g) {
		Graph graph = *g;
		graph.remove_degree_2_nodes();
		graph.disconnect_vacuum_components();
		GraphLabel label = graph.find_canonical_graph_label().first;
		if (g == result.begin() || label < min) {
			min = label;
			minimals.clear();
			minimals.push_back(graph);
		} else if (label == min) {
			minimals.push_back(graph);
		}
	}
	VERIFY(!minimals.empty());
	// find the minimal graph with fixed external nodes
	if (num_external_nodes() > 0)
		*this = get_minimal_graph_fixed_ext_nodes(minimals);
	else
		*this = *minimals.begin();
	//ASSERT(num_edges + graphedges.size() == num_internal_edges());
}

void Graph::contract_multiple_internal_propagators() {
	exset props;
	const map<int, GraphEdge> ges = graphedges_; // copy
	for (map<int, GraphEdge>::const_iterator e = ges.begin(); e != ges.end();
			++e) {
		if (is_external_edge(e->first))
			continue;
		ex mom = e->second.momentum_;
		ex sqmass = e->second.squaredmass_;
		if (mom.is_equal(undefined) || sqmass.is_equal(undefined))
			continue;
		ex p = Propagator(mom, sqmass);
		if (!props.insert(p).second)
			contract_edge(e->first);
	}
}

void Graph::disconnect_vacuum_components() {
	list<Topology> other;
	list<Topology> vac = biconnected_vacuum_components(other);
	other.splice(other.end(), vac);

	Graph graph;
	graph.set_name(name());
	graph.set_loop_momenta(loop_momenta());
	set<int> inserted;
	for (list<Topology>::const_iterator t = other.begin(); t != other.end();
			++t) {
		const map<int, Edge>& es = t->edges();
		for (map<int, Edge>::const_iterator e = es.begin(); e != es.end(); ++e)
			inserted.insert(graph.insert_edge(e->second));
	}
	for (set<int>::const_iterator e = inserted.begin(); e != inserted.end();
			++e) {
		graph.set_momentum(*e, get_momentum(*e));
		graph.set_squaredmass(*e, get_squaredmass(*e));
	}
	graph.swap(*this);
}

// canonical label

// color massive (internal) edges by prolonging them
std::pair<Topology, GiNaC::lst> Graph::mass_colored_topology() const {
	// verify no edges of degree 2
	const set<int>& ns1 = nodes();
	for (set<int>::const_iterator n = ns1.begin(); n != ns1.end(); ++n) {
		const list<Edge>& edges = edges_of_node(*n);
		if (edges.size() == 2 && !is_external_edge(edges.begin()->id)
				&& !is_external_edge(edges.rbegin()->id)) {
			throw runtime_error(
					"node " + to_string(*n) + " of graph " + name()
							+ " has degree 2");
		}
	}
	//  number different masses
	lst masses = get_different_masses();
	map<ex, int, mass_is_less> label_by_mass;
	int lauf = 0;
	for (lst::const_iterator m = masses.begin(); m != masses.end(); ++m)
		label_by_mass[*m] = lauf++;
	// construct topology with prolonged edges according to masses
	Topology colored(*this);
	map<int, GraphEdge>::const_iterator e;
	for (e = graphedges_.begin(); e != graphedges_.end(); ++e) {
		int l = label_by_mass[e->second.squaredmass_];
		for (int i = 0; i < l; ++i)
			colored.insert_degree_2_node(e->first);
	}
	return make_pair(colored, masses);
}

pair<GraphLabel, GraphRelabeling> Graph::find_canonical_graph_label(
		bool permute_external_nodes) const {
	pair<Topology, lst> colored = mass_colored_topology();
	pair<TopologyLabel, TopologyRelabeling> tr =
			colored.first.find_canonical_topology_label(permute_external_nodes);
	GraphLabel label;
	label.masses_ = colored.second;
	label.edge_label_ = tr.first;
	return make_pair(label, tr.second);
}

std::list<std::map<int, int> > Graph::find_node_symmetry_group() const {
	// symmetry group of the mass colored topology
	pair<Topology, lst> mct = mass_colored_topology();
	list < map<int, int> > perms = mct.first.find_node_symmetry_group();

	// remove additional nodes introduced by the coloring
	const set<int>& color_nodes = mct.first.nodes();
	set<int> delete_nodes;
	set<int>::const_iterator n;
	for (n = color_nodes.begin(); n != color_nodes.end(); ++n)
		if (!contains_node(*n))
			delete_nodes.insert(delete_nodes.end(), *n);
	const set<int>& nodes = this->nodes();
	for (list<map<int, int> >::iterator p = perms.begin(); p != perms.end();
			++p) {
		Permutation permutation(*p);
		permutation.erase(delete_nodes);
		map<int, int> new_p;
		for (n = nodes.begin(); n != nodes.end(); ++n)
			new_p.insert(new_p.end(), make_pair(*n, permutation.apply(*n)));
		p->swap(new_p);
	}
	perms.sort();
	perms.unique();
	VERIFY(!perms.empty());

	return perms;
}

/// merges external nodes into one, returns new node id
int identify_external_nodes(Topology& t) {
	VERIFY(!t.nodes().empty());
	int new_node = std::max(1, *t.nodes().rbegin() + 1);
	t.insert_node(new_node);
	set<int> ns = t.nodes();
	for (set<int>::const_iterator n = ns.begin(); n != ns.end(); ++n)
		if (t.is_external_node(*n))
			t.identify_nodes(*n, new_node);
	return new_node;
}

/// splits identified external nodes into external nodes again, returns true on success
/** function succeeds only if the edges of 'g' with the same edge ids as
 *  the external edges from 'nonvac' are adjacent to one node with the same
 *  directions.
 */
bool split_identified_external_nodes(Graph& g, const Graph& nonvac) {
	int new_node = -1;
	list<Edge> ext_edges = nonvac.external_edges();
	int num_ext_edges = ext_edges.size();
	const map<int, list<Edge> >& edges_of_node = g.edges_of_nodes();
	map<int, list<Edge> >::const_reverse_iterator n;
	for (n = edges_of_node.rbegin(); n != edges_of_node.rend(); ++n) {
		const list<Edge>& ge = n->second;
		int count = 0;
		set<int> new_nodes;
		for (list<Edge>::const_iterator it = ge.begin(); it != ge.end(); ++it) {
			if (nonvac.is_external_edge(it->id)) {
				if (nonvac.is_external_node(nonvac.edge(it->id).to)) // it->id has been an outgoing edge
					new_nodes.insert(it->to);
				else
					new_nodes.insert(it->from);
				++count;
				if (new_nodes.size() > 1)
					break;
			}
		}
		if (count == num_ext_edges && new_nodes.size() == 1) {
			new_node = *new_nodes.begin();
			break;
		}
	}
	if (n == edges_of_node.rend())
		return false;

	for (list<Edge>::const_iterator e = ext_edges.begin(); e != ext_edges.end();
			++e) {
		int id = e->id;
		Edge edge = g.edge(id);
		GraphEdge gedge = g.graphedge(id);
		g.remove_edge(id);
		if (nonvac.is_external_node(e->to)) // outgoing
			g.insert_edge(Edge(edge.from, e->to, id));
		else
			g.insert_edge(Edge(e->from, edge.to, id));
		g.set_momentum(id, gedge.momentum_);
		g.set_squaredmass(id, gedge.squaredmass_);
	}
	g.remove_isolated_node(new_node);
	return true;
}

void reverse_all_edges(Topology& t) {
	map<int, Edge> es = t.edges();
	for (map<int, Edge>::const_iterator e = es.begin(); e != es.end(); ++e)
		t.reverse_edge(e->first);
}

void Graph::minimize_by_twists() {
#ifdef DEBUG
	static int count = 0;
	++count;
	print_dot("g_" + to_string(count) + "_a.dot");
#endif
	const list<Edge> ext_edges = external_edges();
	const int num_ext_edges = ext_edges.size();
	Graph gvac(*this);
	if (num_ext_edges > 0)
		identify_external_nodes(gvac);
#ifdef DEBUG
	gvac.print_dot("g_" + to_string(count) + "_b.dot");
#endif
	string origname = name();
	string twistname = name();
	try {
		bool equiv_edges = (gvac.get_different_masses().nops() == 1);
		if (num_ext_edges > 0)
			equiv_edges = false;
		TriConnectedComponents tcc(gvac, equiv_edges);
		pair<GraphLabel, Graph> best;
		list < pair<Topology, set<int> > > twists =
				tcc.find_topologies_of_matroid();
		if (twists.empty())
			throw "empty twists";
		list<pair<Topology, set<int> > >::const_iterator t;
		bool first = true;
		// all minimal graphs (canonical label calculated with permuting external nodes)
		list<Graph> minimals;
		for (t = twists.begin(); t != twists.end(); ++t) {
			twistname = t->first.name();
			Topology tp(t->first); // twisted topo with corrected orientations
			const set<int>& rev = t->second; // reverse edges to match momenta
			for (set<int>::const_iterator r = rev.begin(); r != rev.end(); ++r)
				tp.reverse_edge(*r);
			Graph g(gvac);
			static_cast<Topology&>(g) = tp;
#ifdef DEBUG
			static int count_twist = 0;
			++count_twist;
			g.print_dot("g_" + to_string(count) + "_twisted_"
					+ to_string(count_twist) + ".dot");
#endif
			// sanity check: momentum conserved (leave in production code for now!)
			for (set<int>::const_iterator n = g.nodes().begin();
					n != g.nodes().end(); ++n)
				if (/*!g.is_external_node(*n) &&*/!g.is_momentum_conserved(*n))
					throw "momentum conservation";
			// split external node
			if (num_ext_edges > 0) {
				if (!split_identified_external_nodes(g, *this)) {
					reverse_all_edges(tp);
					static_cast<Topology&>(g) = tp;
					if (!split_identified_external_nodes(g, *this))
						continue;
				}
			}
#ifdef DEBUG
			// currently used only for graphs with all momenta defined,
			// additional check: all momenta still defined after twisting
			map<int, Edge>::const_iterator e;
			for (e = g.edges().begin(); e != g.edges().end(); ++e)
			if (g.graphedges_.find(e->first) == g.graphedges_.end())
			throw "no momentum defined for edge";
#endif
			GraphLabel label = g.find_canonical_graph_label().first;
			if (first || label < best.first) {
				best = make_pair(label, g);
				first = false;
				minimals.clear();
				minimals.push_back(g);
			} else if (label == best.first) {
				minimals.push_back(g);
			}
		}
		if (first)
			throw "original graph not contained in the twisted graphs";
		// find the minimal graph with fixed external nodes
		if (num_ext_edges > 0)
			*this = get_minimal_graph_fixed_ext_nodes(minimals);
		else
			*this = *minimals.begin();
		set_name(origname);
#ifdef DEBUG
		print_dot("g_" + to_string(count) + "_c.dot");
#endif
	} catch (const char* reason) {
		ERROR("Twisting graph '" + origname + "' / '" + twistname + "' " + ""
		"failed for an internal reason\n(" + reason + ")." + ""
		"\nYou could try disabling twisting/matroid related features"
		"\ncompletely (or just for components with external momenta)"
		"\nand using the option construct_minimal_graphs in the jobs"
		"\n(e.g. in setup_sector_mappings)."
		"\nPlease let us know about the problem.");
	}
}

//void Graph::minimize_by_twists(/*bool tweak_external_nodes*/) {
//#ifdef DEBUG
//	static int count = 0;
//	++count;
//	print_postscript("g_" + to_string(count) + "_a.ps");
//#endif
//	string origname = name();
//	string twistname = name();
//	try {
//		bool equiv_edges = (get_different_masses().nops() == 1);
//		TriConnectedComponents tcc(*this, equiv_edges);
//		pair<GraphLabel, Graph> best;
//		list<pair<Topology, set<int> > > twists =
//				tcc.find_topologies_of_matroid();
//		list<pair<Topology, set<int> > >::const_iterator t;
//		for (t = twists.begin(); t != twists.end(); ++t) {
//			twistname = t->first.name();
//			Graph g(*this);
//			static_cast<Topology&> (g) = t->first;
//			map<int, GraphEdge>::iterator ge;
//			for (ge = g.graphedges_.begin(); ge != g.graphedges_.end(); ++ge)
//				// reverse direction of some momenta
//				if (ge->second.momentum_ == undefined)
//					continue; //throw "edge with undefined momentum";
//				else if (t->second.find(ge->first) != t->second.end())
//					ge->second.momentum_ = -ge->second.momentum_;
//			// sanity check: momentum conserved (leave in production code for now!)
//			for (set<int>::const_iterator n = g.nodes().begin(); n
//					!= g.nodes().end(); ++n)
//				if (!g.is_external_node(*n) && !g.is_momentum_conserved(*n))
//					throw "momentum conservation";
//#ifdef DEBUG
//			// currently used only for graphs with all momenta defined,
//			// additional check: all momenta still defined after twisting
//			map<int, Edge>::const_iterator e;
//			for (e = edges().begin(); e != edges().end(); ++e)
//			if (graphedges_.find(e->first) == graphedges_.end()) {
//				throw "no momentum defined for edge";
//			}
//#endif
//			GraphLabel label = g.find_canonical_graph_label().first;
//			if (t == twists.begin() || label < best.first)
//				best = make_pair(label, g);
//		}
//		*this = best.second;
//		set_name(origname);
//#ifdef DEBUG
//		print_postscript("g_" + to_string(count) + "_b.ps");
//#endif
//	} catch (const char* reason) {
//		ERROR("Twisting graph '" + origname + "' / '" + twistname + "' " +
//				"failed for an internal reason\n(" + reason + ")." +
//				"\nYou could try disabling twisting/matroid related features"
//				"\ncompletely (or just for components with external momenta)."
//				"\nPlease let us know about the problem.");
//	}
//}

GiNaC::ex Graph::find_shift(const Graph& graph1,
		const std::pair<GraphLabel, GraphRelabeling>& label1,
		const Graph& graph2,
		const std::pair<GraphLabel, GraphRelabeling>& label2,
		GiNaC::exmap& shift, const Kinematics* kin) {

	Crossing crossing(kin);
	list < map<int, int> > dummysym;
	GiNaC::ex det = Graph::find_shift(graph1, label1, graph2, label2, shift,
			crossing, dummysym);
	GiNaC::exmap mom_rule = crossing.rules_momenta();
	// apply crossing to the shift
	shift = substitute_rhs(shift, mom_rule);
	// add the permutation of the external momenta to the shift
	shift.insert(mom_rule.begin(), mom_rule.end());

	return det;
}

// removes nodes not contained in the graph
vector<int> clean_node_permutation(const vector<int>& node_perm,
		const Graph& graph) {
	vector<int> result;
	result.reserve(graph.nodes().size());
	for (unsigned i = 0; i < node_perm.size(); ++i)
		if (graph.contains_node(node_perm[i]))
			result.push_back(node_perm[i]);
	VERIFY(result.size() == graph.num_nodes());

	//	cerr << "node permutation of graph " << graph.name() << endl;
	//	for (unsigned i = 0; i < result.size(); ++i)
	//		cerr << i << " <-- " << result[i] << endl;
	//	static_cast<const Topology&> (graph).Topology::print_dot("topo_A_"
	//			+ graph.name() + ".dot");

	return result;
}

// get the crossing from graph 'g1' to 'g2' with given node permutations
Crossing get_crossing(const std::map<int, int>& node_perm, const Graph& g1,
		const Graph& g2, const Kinematics* kin) {
	map<int, int> ext_edge_perm;
	map<int, int>::const_iterator it;
	for (it = node_perm.begin(); it != node_perm.end(); ++it) {
		VERIFY(g1.contains_node(it->first) && g2.contains_node(it->second));
		bool is_external = g1.is_external_node(it->first);
		bool is_external2 = g2.is_external_node(it->second);
		VERIFY(is_external == is_external2);
		if (is_external) {
			const list<Edge>& edges1 = g1.edges_of_node(it->first);
			const list<Edge>& edges2 = g2.edges_of_node(it->second);
			VERIFY(edges1.size() == 1 && edges2.size() == 1);
			ext_edge_perm[edges1.begin()->id] = edges2.begin()->id;
		}
	}
	VERIFY(kin->external_momenta().nops() == ext_edge_perm.size());
	return Crossing(kin, Permutation(ext_edge_perm));
}

// finds the node permutation and the minimal crossing needed for finding the shift
// from 'graph1' to 'graph2' (graphs must be isomorphic)
// with the symmetry group of 'graph2' the node permutation is transformed
// such that the crossing gets minimal
std::pair<std::map<int, int>, Crossing> get_minimal_crossing(
		const Graph& graph1, const std::vector<int>& np1, //
		const Graph& graph2, const std::vector<int>& np2, //
		const std::list<std::map<int, int> >& sym2, const Kinematics* kin) {
	const map<Crossing, Crossing>& crossmap = kin->equivalent_crossings();
	map<int, int> node_perm;
	Crossing crossing(kin);
	VERIFY(!sym2.empty());
	list<map<int, int> >::const_iterator s2;
	for (s2 = sym2.begin(); s2 != sym2.end(); ++s2) {
		vector<int> new_np2(graph2.num_nodes());
		for (unsigned i = 0; i < new_np2.size(); ++i)
			new_np2[i] = s2->at(np2[i]);
		map<int, int> tmp;
		tmp = TopologyRelabeling::find_node_permutation(np1, new_np2);
		Crossing c = get_crossing(tmp, graph1, graph2, kin);
		if (s2 == sym2.begin() || crossmap.at(c) < crossmap.at(crossing)
				|| (crossmap.at(c) == crossmap.at(crossing) && c < crossing)) {
			crossing = c;
			node_perm = tmp;
			if (crossing.is_identity())
				break;
		}
	}
	return make_pair(node_perm, crossing);
}

GiNaC::ex Graph::find_shift(const Graph& graph1,
		const std::pair<GraphLabel, GraphRelabeling>& label1,
		const Graph& graph2,
		const std::pair<GraphLabel, GraphRelabeling>& label2,
		GiNaC::exmap& shift, Crossing& crossing,
		const std::list<std::map<int, int> >& sym2) {
#ifdef DEBUG
	graph1.print_postscript("shift_g1.ps");
	graph2.print_postscript("shift_g2.ps");
#endif
	const Kinematics* kin = crossing.kinematics();
	VERIFY(kin != 0);

	// assert the graphs are isomorphic
	string gname1 = "'" + graph1.name() + "'";
	string gname2 = "'" + graph2.name() + "'";
	if (label1.first != label2.first)
		throw runtime_error(
				"graphs " + gname1 + " and " + gname2
						+ " don't have the same canonical labeling.");

	// assert the loop momenta are set
	const lst& srcloops = graph1.loop_momenta();
	const lst& destloops = graph2.loop_momenta();
	if (srcloops.nops() != destloops.nops()) {
		ostringstream srcl, destl;
		srcl << srcloops;
		destl << destloops;
		throw runtime_error("not the same number of loop momenta set:\n " //
		+ gname1 + ": " + srcl.str() + "\n " //
				+ gname2 + ": " + destl.str());
	}
	// assert the propagator momenta are set
	if (!graph1.get_edges_undefined_momentum().empty())
		throw runtime_error("graph " + gname1 + " has undefined momenta.");
	if (!graph2.get_edges_undefined_momentum().empty())
		throw runtime_error("graph " + gname2 + " has undefined momenta.");

	// find the node permutation and the crossing

	map<int, int> node_perm;
	vector<int> np1, np2;
	np1 = clean_node_permutation(label1.second.node_permutation(), graph1);
	np2 = clean_node_permutation(label2.second.node_permutation(), graph2);
	if (!sym2.empty() && graph2.num_external_nodes() > 0) {
		pair<map<int, int>, Crossing> c = get_minimal_crossing(graph1, np1,
				graph2, np2, sym2, kin);
		node_perm = c.first;
		crossing = c.second;
	} else {
		node_perm = TopologyRelabeling::find_node_permutation(np1, np2);
		crossing = get_crossing(node_perm, graph1, graph2, kin);
	}
	exmap mom_rule = crossing.rules_momenta();

	// find the mapping of the cycle edges (compare masses, stop on first match)

	set<int> cycles = graph1.find_cycles(), visited;
	if (cycles.size() != srcloops.nops())
		throw runtime_error(
				"wrong number of loop momenta set in graph " + gname1);
	map<int, pair<int, int> > cycle_edge_perm;
	for (set<int>::const_iterator c = cycles.begin(); c != cycles.end(); ++c) {
		const Edge& edge = graph1.edge(*c);
		int from = node_perm.at(edge.from);
		int to = node_perm.at(edge.to);
		ex sqmass = graph1.get_squaredmass(edge.id);
		const list<Edge>& oes = graph2.edges_of_node(from);
		int sign = 0;
		list<Edge>::const_iterator e2 = oes.begin();
		for (; e2 != oes.end(); ++e2) {
			if (visited.find(e2->id) != visited.end())
				continue;
			ex sqmass2 = graph2.get_squaredmass(e2->id);
			if (!sqmass2.is_equal(sqmass))
				continue;
			if (e2->to == to && e2->from == from)
				sign = 1;
			else if (e2->from == to && e2->to == from)
				sign = -1;
			if (sign != 0) { // for multiple edges any edge will do the job
				visited.insert(e2->id);
				break;
			}
		}
		VERIFY(e2 != oes.end());
		cycle_edge_perm[edge.id] = make_pair(sign, e2->id);
	}

	// determine the shift of the loop momenta which transforms graph1 to graph2

	lst tmpsrcloops, equations;
	exmap loop2tmp, tmp2loop;
	provide_alternative_symbols(srcloops, tmpsrcloops, loop2tmp, tmp2loop);
	map<int, pair<int, int> >::const_iterator p;
	for (p = cycle_edge_perm.begin(); p != cycle_edge_perm.end(); ++p) {
		const GraphEdge& g1 = graph1.graphedge(p->first);
		const GraphEdge& g2 = graph2.graphedge(p->second.second);
		ex lhs = g1.momentum_;
		ex rhs = p->second.first * g2.momentum_;
		lhs = lhs.subs(mom_rule, subs_options::no_pattern);
		lhs = lhs.subs(loop2tmp, subs_options::no_pattern);
		equations.append(lhs == rhs);
	}
	equations = ex_to < lst > (lsolve(equations, tmpsrcloops));

	VERIFY(equations.nops() == destloops.nops());
	ex det = abs(jacobi_determinant(equations, destloops));
	shift = equations_to_substitutions(equations, tmp2loop, true);

	Crossing inverse_cross = crossing.inverse();
	shift = substitute_rhs(shift, inverse_cross.rules_momenta());

	VERIFY(!det.is_zero());
	return det;
}

namespace {

/** find all possible ways to map edges between nodes n1 and n2 of graph1 to
 *  edges between nodes of graph2 (mapped by node_perm)
 */

std::list<std::map<int, std::pair<int, int> > > get_edge_mappings(
		const Graph& graph1, //
		const Graph& graph2, //
		const std::map<int, int>& node_perm, //
		int n1, int n2) {

	// mappings of edge id1 to pair (orientation, edge id2)
	list < map<int, pair<int, int> > > result;

	// pairs of orientation and edge ids by their mass
	map<ex, vector<pair<int, int> >, ex_is_less> edges_from_by_mass,
			edges_to_by_mass;

	// sort edges by their masses and find orientation

	list<Edge> edges_from = graph1.find_edges_between(n1, n2);
	list<Edge>::const_iterator e;
	for (e = edges_from.begin(); e != edges_from.end(); ++e) {
		int sign = e->from == n1 ? 1 : -1;
		ex sqm = graph1.get_squaredmass(e->id);
		edges_from_by_mass[sqm].push_back(make_pair(sign, e->id));
	}

	int n1_to = node_perm.at(n1);
	int n2_to = node_perm.at(n2);
	list<Edge> edges_to = graph2.find_edges_between(n1_to, n2_to);
	for (e = edges_to.begin(); e != edges_to.end(); ++e) {
		int sign = e->from == n1_to ? 1 : -1;
		ex sqm = graph2.get_squaredmass(e->id);
		edges_to_by_mass[sqm].push_back(make_pair(sign, e->id));
	}

	// find permutations of the multi-edges with the same mass

	// each entry of the vector corresponds to a mass
	vector < list<map<int, pair<int, int> > >
			> trafos_by_mass(edges_from_by_mass.size());

	VERIFY(edges_from_by_mass.size() == edges_to_by_mass.size());
	map<ex, vector<pair<int, int> >, ex_is_less>::iterator it;
	int mass_number = 0;
	for (it = edges_to_by_mass.begin(); it != edges_to_by_mass.end(); ++it) {
		vector < pair<int, int> > &to = it->second;
		const vector<pair<int, int> >& from = edges_from_by_mass[it->first];
		VERIFY(to.size() == from.size());
		std::sort(to.begin(), to.end());
		do {
			map<int, pair<int, int> > m;
			for (unsigned i = 0; i < to.size(); ++i) {
				int sign = from[i].first * to[i].first;
				m[from[i].second] = make_pair(sign, to[i].second);
			}
			trafos_by_mass[mass_number].push_back(map<int, pair<int, int> >());
			trafos_by_mass[mass_number].back().swap(m);
		} while (std::next_permutation(to.begin(), to.end()));
		++mass_number;
	}

	// find combinations of the edge-transformations with different masses

	combinations<map<int, pair<int, int> > > combs(trafos_by_mass);
	vector < map<int, pair<int, int> > > r;
	while (combs.get_next(r)) {
		map<int, pair<int, int> > merged;
		for (unsigned i = 0; i < r.size(); ++i)
			merged.insert(r[i].begin(), r[i].end());
		result.push_back(map<int, pair<int, int> >());
		result.back().swap(merged);
	}

	return result;
}

}

void Graph::find_symmetry_shifts(std::list<GiNaC::exmap>& shifts,
		const Kinematics* kin) const {
	LOGX("\nFinding symmetry shifts of graph " << name() << ":");
	// find the node symmetry group of the graph with permutation of external
	// nodes allowed in order to find crossings equivalent to the identity
	list < map<int, int> > perms = find_node_symmetry_group();
	//LOGX("  Found " << perms.size() << " symmetries (from node permutations) in the graph");
	find_symmetry_shifts(shifts, kin, perms);
}

void Graph::find_symmetry_shifts(std::list<GiNaC::exmap>& shifts,
		const Kinematics* kin,
		const std::list<std::map<int, int> >& perms) const {
	using GiNaC::exmap;
	LOGX(
			"  Found " << perms.size() << " symmetries (from node permutations) in the graph");
	vector<Crossing> the_crossings;
	set<int> cycles = find_cycles();
	list < map<int, pair<int, int> > > cycle_mappings;

	set < map<int, int> > testset;

	for (list<map<int, int> >::const_iterator p = perms.begin();
			p != perms.end(); ++p) {

		// crossing
		const map<int, int>& node_perm = *p;
		Crossing crossing = get_crossing(node_perm, *this, *this, kin);
		if (!crossing.is_equivalent_to_identity()) {
			LOGX(
					"  omitting crossing: " << crossing.rules_momenta() << "(not equivalent to identity crossing)");
			continue;
		};

		// all edge mappings of the fundamental cycle

		set < pair<int, int> > visited;
		vector < list<map<int, pair<int, int> > > > edge_mappings;
		for (set<int>::const_iterator c = cycles.begin(); c != cycles.end();
				++c) {
			const Edge& e = edge(*c); // filter multi-edges, take only one
			int n1 = std::min(e.from, e.to);
			int n2 = std::max(e.from, e.to);
			if (!visited.insert(make_pair(n1, n2)).second)
				continue;
			edge_mappings.push_back(
					get_edge_mappings(*this, *this, *p, n1, n2));
		}

		// all combinations of the edge mappings of the cycles

		combinations<map<int, pair<int, int> > > combs(edge_mappings);
		vector < map<int, pair<int, int> > > r;
		while (combs.get_next(r)) {
			map<int, pair<int, int> > merged;
			for (unsigned i = 0; i < r.size(); ++i)
				merged.insert(r[i].begin(), r[i].end());

			map<int, pair<int, int> >::const_iterator n;
			map<int, int> test;
			for (n = merged.begin(); n != merged.end(); ++n)
				test[n->first] = n->second.second;

			// filter out equal mappings
			if (testset.insert(test).second) {
				cycle_mappings.push_back(merged);
				the_crossings.push_back(crossing);
			}
		}
	}

	//	// delete non-fundamental entries
	//	list<map<int, pair<int, int> > >::iterator it;
	//	for (it = cycle_mappings.begin(); it != cycle_mappings.end(); ++it)
	//		for (map<int, pair<int, int> >::iterator m = it->begin(); m
	//				!= it->end();)
	//			if (cycles.find(m->first) == cycles.end())
	//				it->erase(m++);
	//			else
	//				++m;

	list<map<int, pair<int, int> > >::const_iterator ml;
	unsigned lauf = 0;
	VERIFY(the_crossings.size() == cycle_mappings.size());
	for (ml = cycle_mappings.begin(); ml != cycle_mappings.end();
			++ml, ++lauf) {

		const Crossing& crossing = the_crossings[lauf];
		exmap mom_rule = crossing.rules_momenta();

		// determine the shift of the loop momenta which transforms graph1 to graph2

		exmap shift;
		lst srcloops = loop_momenta_;
		lst tmpsrcloops, equations;
		exmap loop2tmp, tmp2loop;
		provide_alternative_symbols(srcloops, tmpsrcloops, loop2tmp, tmp2loop);
		map<int, pair<int, int> >::const_iterator p;
		// we have one equation too much
		for (p = ml->begin(); p != ml->end(); ++p) {
			const GraphEdge& g1 = graphedge(p->first);
			const GraphEdge& g2 = graphedge(p->second.second);
			ex lhs = g1.momentum_;
			ex rhs = p->second.first * g2.momentum_;
			lhs = lhs.subs(mom_rule, subs_options::no_pattern);
			lhs = lhs.subs(loop2tmp, subs_options::no_pattern);
			equations.append(lhs == rhs);
		}
		equations = ex_to < lst > (lsolve(equations, tmpsrcloops));
		VERIFY(equations.nops() == srcloops.nops());
		ex det = abs(jacobi_determinant(equations, srcloops));
		VERIFY(!det.is_zero());
		shift = equations_to_substitutions(equations, tmp2loop, true);
		shift.insert(mom_rule.begin(), mom_rule.end());
		// add the permutation of the external momenta to the shift
		if (det.is_equal(1)) {
			if (!shift.empty())
				shifts.push_back(shift);
		} else {
			LOG(
					"Reject symmetry shift with |det| " << det << ", shift: " << shift);
		}
	}
	LOGX(
			"  Found " << shifts.size() << " symmetry shifts (permutations of multi-edges performed)");
}

bool Graph::match_to_integralfamily(const IntegralFamily* ic) {
	using namespace GiNaC;
	const bool analyze_subtopos = false;
	const vector<Propagator>& propagators = ic->propagators();
	if (ic->loop_momenta().nops() != loop_momenta_.nops()) {
		LOGXX("mismatch in number of loops");
		return false;
	}
	// temp names for our loop momenta in case they coincide with those from ic
	lst tmp_moms;
	exmap mom2tmp, tmp2mom;
	provide_alternative_symbols(loop_momenta_, tmp_moms, mom2tmp, tmp2mom);
	// try to find shifts from of loop momenta such that propagators match
	list < pair<int, pair<int, lst> > > shifts; // {(subtopo,(sector,eqs)),..}
	shifts.push_back(make_pair(0, make_pair(0, lst()))); // one empty element
	map<int, GraphEdge>::const_iterator eit;
	int loop_edge = 0; // loop edge numbered like 0,1,...
	for (eit = graphedges_.begin(); eit != graphedges_.end(); ++eit) {
		const GraphEdge& e = eit->second;
		if (e.momentum_ == undefined || e.squaredmass_ == undefined) {
			WARNING("Graph " << name() << " has undefined kinematics");
			return false;
		} else if (freeof(e.momentum_, loop_momenta_)) { // edge not part of a loop
			continue;
		}
		list < pair<int, pair<int, lst> > > new_shifts;
		while (!shifts.empty()) {
			// consider next edge and all possible propagators
			lst shift = ex_to < lst > (shifts.front().second.second);
			int sub = shifts.front().first;
			int sec = shifts.front().second.first;
			if (analyze_subtopos) // incl. subtopos (current edge unmatched)
				new_shifts.splice(new_shifts.end(), shifts, shifts.begin());
			else
				// no subtopologies
				shifts.pop_front();
			int new_sub = sub | (1 << loop_edge); // new subtopology id
			// try to match current edge
			for (int i = 0; i < (int) propagators.size(); ++i) {
				try {
					if (e.squaredmass_ != propagators[i].squaredmass())
						continue;
					int new_sec = sec | (1 << i); // new sector id
					ex o = e.momentum_.subs(mom2tmp); // old momentum
					ex n = propagators[i].momentum(); // new momentum
					lst s1 = shift, s2 = shift;
					s1 = ex_to < lst > (lsolve(s1.append(o == n), tmp_moms));
					s2 = ex_to < lst > (lsolve(s2.append(o == -n), tmp_moms));
					if (s1.nops() > 0)
						new_shifts.push_back(
								make_pair(new_sub, make_pair(new_sec, s1)));
					if (s2.nops() > 0)
						new_shifts.push_back(
								make_pair(new_sub, make_pair(new_sec, s2)));
				} catch (exception& e) {
					// ignore propagators without well-defined momentum
				}
			}
		}
		shifts.swap(new_shifts);
		++loop_edge;
	}
	// select shifts with jacobian 1 and provide substitution maps
	bool found = false;
	list<pair<int, pair<int, lst> > >::iterator l;
	for (l = shifts.begin(); l != shifts.end(); ++l) {
		int sub = l->first;
		Sector sec(ic, l->second.first);
		lst sol = l->second.second;
		ex det = abs(jacobi_determinant(sol, ic->loop_momenta()).expand());
		if (det == 0)
			continue;
		if (det != 1)
			ABORT("Don't know how to handle Jacobian determinant " << det);
		exmap subst = equations_to_substitutions(sol, tmp2mom);
		if (sub == (1 << loop_edge) - 1) { // match for full graph
			found = true;
			graph_to_sector_[sec] = subst;
		}
		subgraph_to_sector_[sub][sec] = subst;
	}
	return found;
}

void Graph::print_sector_matches() const {
	map<int, map<Sector, GiNaC::exmap> >::const_iterator i;
	for (i = subgraph_to_sector_.begin(); i != subgraph_to_sector_.end(); ++i) {
		map<Sector, GiNaC::exmap>::const_iterator j;
		LOGNX("     sectors for subtopo " << i->first << ": ");
		for (j = i->second.begin(); j != i->second.end(); ++j) {
			LOGNX(j->first << " ");
		}
		LOGX("");
	}
}

const std::pair<const Sector, GiNaC::exmap>* Graph::shift_to_sector() const {
	if (graph_to_sector_.empty())
		return 0;
	return &(*graph_to_sector_.begin());
}

// output

YAML::Emitter& operator<<(YAML::Emitter& ye, const Graph& g) {
	using namespace YAML;
	ye << BeginMap;
	ye << Key << "name" << Value << g.name();
	ye << Key << "edges" << Value << Flow << BeginSeq;
	const map<int, Edge>& es = g.edges();
	for (map<int, Edge>::const_iterator e = es.begin(); e != es.end(); ++e) {
		ye << BeginSeq;
		ye << e->second.from << e->second.to << e->second.id
				<< g.get_momentum(e->second.id)
				<< g.get_squaredmass(e->second.id);
		ye << EndSeq;
	}
	ye << EndSeq;

	ye << Key << "nodes" << Value << Flow << BeginSeq;
	const set<int>& node = g.nodes();
	for (set<int>::const_iterator n = node.begin(); n != node.end(); ++n)
		ye << *n;
	ye << EndSeq;

	ye << EndMap;

	return ye;
}

void Graph::read(const YAML::Node& n, const GiNaC::lst& loopmoms,
		const GiNaC::lst& ext_kin) {
	using namespace YAML;
	if (n.GetType() != CT_MAP || n.size() != 3)
		throw runtime_error("node is not a 3-element map " + position_info(n));
	string name;
	n["name"] >> name;
	set_name(name);
	set_loop_momenta(loopmoms);
	const Node& edge_node = n["edges"];
	map<int, pair<ex, ex> > mom_mass;
	for (Iterator i = edge_node.begin(); i != edge_node.end(); ++i) {
		if (i->GetType() != CT_SEQUENCE || i->size() != 5)
			throw runtime_error(
					"node is not a 5-element sequence " + position_info(*i));
		Edge edge;
		(*i)[0] >> edge.from;
		(*i)[1] >> edge.to;
		(*i)[2] >> edge.id;
		insert_edge(edge);
		ex mom, mass;
		Reduze::read((*i)[3], mom, add_lst(loopmoms, ext_kin));
		Reduze::read((*i)[4], mass, ext_kin);
		mom_mass[edge.id] = make_pair(mom, mass);
	}
	for (map<int, pair<ex, ex> >::const_iterator i = mom_mass.begin();
			i != mom_mass.end(); ++i) {
		set_momentum(i->first, i->second.first);
		set_squaredmass(i->first, i->second.second);
	}

	const Node& node_node = n["nodes"];
	if (node_node.GetType() != CT_SEQUENCE)
		throw runtime_error(
				"node is not a sequence " + position_info(node_node));
	for (Iterator i = node_node.begin(); i != node_node.end(); ++i) {
		int n;
		*i >> n;
		insert_node(n);
	}
}

void Graph::print_dot(const std::string& filename) const {
	ofstream of(filename.c_str());
	print_dot(of);
	of.close();
}

void Graph::print_dot(std::ostream& of) const {
	bool directed = true;
	of << (directed ? "digraph" : "graph") << " ";
	of << (name() != "graph" ? name() : "graph0") << " {\n";

	of << "  label=\"" << name() << "\";\n";
	of << "  rankdir=\"LR\";" << "\n";

	set<int> innodes, outnodes;
	of << "  subgraph edges {\n";
	const map<int, Edge>& es = edges();
	for (map<int, Edge>::const_iterator ep = es.begin(); ep != es.end(); ++ep) {
		const Edge& e = ep->second;
		if (e.from < 0)
			innodes.insert(e.from);
		if (e.to < 0)
			outnodes.insert(e.to);
		of << "    " << e.from << (directed ? " -> " : " -- ") << e.to;
		if (graphedges_.find(e.id) != graphedges_.end()) {
			const GraphEdge& ge = graphedges_.find(e.id)->second;
			if (ge.momentum_ != undefined || ge.squaredmass_ != undefined)
				of << " [label=\"(" << ge.momentum_ << "," << ge.squaredmass_
						<< ";" << ep->first << ")\"]";
		}
		of << ";\n";
	}
	of << "  }\n";

	set<int>::const_reverse_iterator n;
	if (!innodes.empty()) {
		of << "  subgraph incoming { rank=\"source\"; ";
		for (n = innodes.rbegin(); n != innodes.rend(); ++n)
			of << *n << "; ";
		of << "}\n";
	}
	if (!outnodes.empty()) {
		of << "  subgraph outgoing { rank=\"sink\"; ";
		for (n = outnodes.rbegin(); n != outnodes.rend(); ++n)
			of << *n << "; ";
		of << "}\n";
	}

	// nodes: small filled circles without label
	for (set<int>::const_iterator n = nodes().begin(); n != nodes().end(); ++n)
		of << *n << " [shape=point];\n";

	of << "}\n";

	of << "\n";
}

void Graph::removed_edge(int id) {
	graphedges_.erase(id);
}

void Graph::swapped() {
	subgraph_to_sector_.clear();
	graph_to_sector_.clear();
	graphedges_.clear();
	loop_momenta_.remove_all();
}

//
//
//

SectorGL::SectorGL(const YAML::Node& node) :
		sector_(node["sector"]) {
	const IntegralFamily* ic = sector_.integralfamily();
	graph_.read(node["graph"], ic->loop_momenta(),
			ic->kinematics()->get_external_momenta_and_kinematics());
	GraphLabel l;
	l.read(node["label"], ic->kinematics()->kinematic_invariants());
	GraphRelabeling r;
	node["relabeling"] >> r;
	labels_ = make_pair(l, r);
}

YAML::Emitter& operator<<(YAML::Emitter& ye, const SectorGL& s) {
	using namespace YAML;
	ye << BeginMap;
	ye << Key << "sector" << Value << s.sector_;
	ye << Key << "graph" << Value << s.graph_;
	ye << Key << "label" << Value << s.labels_.first;
	ye << Key << "relabeling" << Value << s.labels_.second;
	ye << EndMap;
	return ye;
}

} // namespace Reduze

