/*  identitygenerator.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "identitygenerator.h"
#include "shiftoperators.h"
#include "propagator.h"
#include "kinematics.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "functions.h"
#include "files.h"
#include "equation.h"
#include "equationlist.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "filedata.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// register type
namespace {
YAMLProxy<IdentityGeneratorOptions> dummy;
}

// IdentityGeneratorOptions


void IdentityGeneratorOptions::set_map_on_coeff(const GiNaC::exmap& m) {
	map_on_coeff_ = m;
}
void IdentityGeneratorOptions::set_transform_integrals_to_equivalent(bool b) {
	transform_integrals_to_equivalent_ = b;
	if (transform_integrals_to_equivalent_)
		map_on_integral_ = &SectorMappings::get_equivalent_or_unique_zero;
	else
		map_on_integral_ = 0; // has been set to non zero value in init()
}
void IdentityGeneratorOptions::set_solve_equations(bool b) {
	solve_equations_ = b;
}
void IdentityGeneratorOptions::set_subsectors_to_zero(bool b) {
	set_subsectors_to_zero_ = b;
	if (set_subsectors_to_zero_)
		eliminate_subsectors_by_shift_ = false;
}
void IdentityGeneratorOptions::set_eliminate_subsectors_by_shift(bool b) {
	eliminate_subsectors_by_shift_ = b;
	if (set_subsectors_to_zero_)
		eliminate_subsectors_by_shift_ = false;
}
void IdentityGeneratorOptions::init() {
	if (transform_integrals_to_equivalent_)
		map_on_integral_ = &SectorMappings::get_equivalent_or_unique_zero;
}

void IdentityGeneratorOptions::print(YAML::Emitter& os) const {
	using namespace YAML;
	os << BeginMap;
	os << Key << "transform_integrals_to_equivalent" << Value
			<< transform_integrals_to_equivalent_;
	//	os << Key << "solve_equations" << Value << solve_equations_;
	os << Key << "set_subsectors_to_zero" << Value << set_subsectors_to_zero_;
	os << Key << "eliminate_subsectors_by_shift" << Value
			<< eliminate_subsectors_by_shift_;
	os << Key << "map_on_coeff" << Value << Flow << map_on_coeff_;
	os << EndMap;
}

void IdentityGeneratorOptions::read(const YAML::Node& node) {
	verify_yaml_spec(node);
	const YAML::Node* n;
	if ((n = node.FindValue("transform_integrals_to_equivalent")))
		*n >> transform_integrals_to_equivalent_;
	//	if ((n = node.FindValue("solve_equations")))
	//		*n >> solve_equations_;
	if ((n = node.FindValue("set_subsectors_to_zero")))
		*n >> set_subsectors_to_zero_;
	if ((n = node.FindValue("eliminate_subsectors_by_shift")))
		*n >> eliminate_subsectors_by_shift_;
	if ((n = node.FindValue("map_on_coeff"))) {
		lst l = Files::instance()->kinematics()->kinematic_invariants_and_dimension();
		Reduze::read(*n, map_on_coeff_, l);
	}
	init();
}

// IdentityGenerator

std::map<std::string, GenericIdentityGenerator>
		IdentityGenerator::generic_generators_;

void IdentityGenerator::generate(const INT& in, IdentityList& out,
		const string& type) {
	IdentityList identities;
	if (type == "sector_relations")
		genid_shift(in, identities);
	else if (type == "sector_symmetries")
		genid_symmetry(in, identities);
	else
		genid_ibp_lorentz(in, identities, type);

	if (options_.eliminate_subsectors_by_shift_) {
		set<INT> ints;
		identities.find_subsector_INT(static_cast<int> (in.t()), ints);
		IdentityList replace;
		genid_shift(ints, replace);
		identities.replace(replace);
	}
	out.splice(identities);
}

void IdentityGenerator::generate(const set<INT>& s, IdentityList& out,
		const string& type) {
	set<INT>::const_iterator it;
	for (it = s.begin(); it != s.end(); ++it)
		generate(*it, out, type);
}

void IdentityGenerator::genid_ibp_lorentz(const INT& in, IdentityList& result,
		const string& type) {
	const IntegralFamily* f = in.integralfamily();
	const LinearCombinationGenericList& eqs = generic_equations(f, type);
	IdentityList identities;
	LinearCombinationGenericList::const_iterator it;
	for (it = eqs.begin(); it != eqs.end(); ++it) {
		LinearCombination lincomb = it->get_linear_combination(in,
				options_.set_subsectors_to_zero_);
		identities.push_back(Identity());
		identities.back().swap_terms(lincomb);
	}
	apply_remaining_options(identities);
	result.splice(identities);
}

Identity IdentityGenerator::get_shift_identity(const INT& integin,
		bool replace_symbol_to_replace_by_one, bool set_subsecs_to_zero) {
	// minimal crossing, permutation symmetries, obvious zero
	INT imin = SectorMappings::get_equivalent_or_unique_zero(integin);

	LinearCombination rhs;

	// check if we find a shift to uncrossed version of same sector
	bool found_shift = false;
	if (imin.integralfamily()->is_crossed()) {
		const SectorMappings* m = Files::instance()->sectormappings(
				imin.integralfamily()->name());
		map<int, Sector>::const_iterator target = m->sector_relations().find(
				imin.id());
		if (target != m->sector_relations().end()) {
			const exmap& shift =
					m->sector_relation_shifts().find(imin.id())->second;
			rhs = imin.apply_shift(shift, target->second.integralfamily(), -1,
					replace_symbol_to_replace_by_one, set_subsecs_to_zero);
			found_shift = true;
		}
	}

	// search shift (crossed families inherit from uncrossed ones)
	if (!found_shift) {
		pair<INT, Crossing> unx = Crossing::uncross(imin);
		INT inox = unx.first;
		const string famnox = inox.integralfamily()->name();
		const SectorMappings* m = Files::instance()->sectormappings(famnox);
		map<int, Sector>::const_iterator target = //
				m->sector_relations().find(inox.id());
		if (target != m->sector_relations().end()) {
			const exmap& shift =
					m->sector_relation_shifts().find(inox.id())->second;
			LinearCombination rhsnox = inox.apply_shift(shift,
					target->second.integralfamily(), -1,
					replace_symbol_to_replace_by_one, set_subsecs_to_zero);
			// cross before recursion to find minimal cross in case of chained crossings
			rhs = unx.second.transform(rhsnox);
			if (replace_symbol_to_replace_by_one) {
				const Kinematics* kin = Files::instance()->kinematics();
				const exmap& s2o = kin->get_rule_symbol_to_replace_by_one();
				rhs.apply(s2o, subs_options::no_pattern);
			}
			found_shift = true;
		}
	}
	// also provide other simplifications
	if (!found_shift) {
		rhs.insert(imin, -1, false);
	}

	rhs.remove_zero_integral();
	if (found_shift) {
		IdentityList replace;
		for (LinearCombination::const_iterator t = rhs.begin(); t != rhs.end(); ++t)
			replace.push_back(get_shift_identity(t->first,
					replace_symbol_to_replace_by_one, set_subsecs_to_zero));
		for (IdentityList::const_iterator r = replace.begin(); r
				!= replace.end(); ++r)
			rhs.replace(*r);
	}

	// build up identity
	Identity outeq;
	// rhs: minus the shifted expression
	outeq.swap_terms(rhs);
	// lhs: the original integral
	outeq.insert(integin, 1);
	return outeq;
}

void IdentityGenerator::genid_shift(const INT& integral, IdentityList& out,
		bool no_options) {
	IdentityList ident;
	ident.push_back(get_shift_identity(integral, true,
			options_.set_subsectors_to_zero_));
	if (!no_options)
		apply_remaining_options(ident);
	if (!ident.back().empty())
		out.splice(ident);
}

void IdentityGenerator::genid_shift(const set<INT>& in, IdentityList& out) {
	IdentityList identities;
	set<INT>::const_iterator it;
	for (it = in.begin(); it != in.end(); ++it)
		genid_shift(*it, identities, true);
	apply_remaining_options(identities);
	out.splice(identities);
}

void IdentityGenerator::genid_symmetry(const INT& integral, IdentityList& out,
		bool no_options) {
	IdentityList identities;
	int id = integral.id();
	const SectorMappings* m = Files::instance()->sectormappings(
			integral.integralfamily()->name());
	const map<int, list<exmap> >& all_shifts = m->sector_symmetries();
	map<int, list<exmap> >::const_iterator it = all_shifts.find(id);
	if (it == all_shifts.end())
		return;
	const list<exmap>& shifts = it->second;
	for (list<exmap>::const_iterator s = shifts.begin(); s != shifts.end(); ++s) {
		// rhs
		LinearCombination lc = integral.apply_shift(*s,
				integral.integralfamily(), -1, true,
				options_.set_subsectors_to_zero_);
		// rhs
		identities.push_back(Identity());
		identities.back().swap_terms(lc);
		// lhs
		identities.back().insert(integral, 1);
	}

	if (!no_options)
		apply_remaining_options(identities);

	out.splice(identities);
}

const LinearCombinationGenericList& IdentityGenerator::generic_equations(
		const IntegralFamily* f, const std::string& type) {
	map<string, GenericIdentityGenerator>::const_iterator g;
	g = generic_generators_.find(f->name());
	if (g == generic_generators_.end()) {
		generic_generators_.insert(make_pair(f->name(),
				GenericIdentityGenerator(f)));
		return generic_equations(f, type);
	}
	return g->second.generic_equations(type);
}

void IdentityGenerator::apply_remaining_options(IdentityList& identities) const {
	if (options_.map_on_integral_)
		identities.transform_integrals(options_.map_on_integral_);
	if (!options_.map_on_coeff_.empty())
		identities.apply(options_.map_on_coeff_);
	identities.remove_zero_integral();
	if (options_.solve_equations_)
		identities.solve();
	const Kinematics* kin = Files::instance()->kinematics();
	const exmap& s2o = kin->get_rule_symbol_to_replace_by_one();
	identities.apply(s2o, subs_options::no_pattern);
}

// generic IBP generation

GenericIdentityGenerator::GenericIdentityGenerator(const IntegralFamily* f) :
	integralfamily_(f) {
	ASSERT(f != 0);
	construct_shift_operators();
	construct_generic_identities();
}

const LinearCombinationGenericList& GenericIdentityGenerator::generic_equations(
		const std::string& type) const {
	map<string, LinearCombinationGenericList>::const_iterator e;
	e = generic_equations_.find(type);
	if (e == generic_equations_.end())
		ABORT("Unknown generic equations of type " << type);
	return e->second;
}

void GenericIdentityGenerator::construct_shift_operators() {
	// determine effect of q_i d/dq_k on general integrand,
	// expressed in terms of shift operators
	LOGXX("determine differential operators: q_i * d/dq_k * " << INTGeneric(integralfamily_));
	vector<int> v(2, 0);
	const Kinematics* kin = integralfamily_->kinematics();
	const lst& loop_mom = integralfamily_->loop_momenta();
	const lst& iext_mom = kin->independent_external_momenta();
	const lst all_mom = add_lst(loop_mom, iext_mom);
	const vector<Propagator>& propagators = integralfamily_->propagators();
	const exmap& s2i = kin->rules_sp_to_invariants();
	const exmap& s2p = integralfamily_->rules_sp_to_prop();
	for (size_t i = 0; i < all_mom.nops(); i++) {
		for (size_t k = 0; k < all_mom.nops(); k++) {
			ASSERT(is_a<symbol>(all_mom[i]) && is_a<symbol>(all_mom[k]));
			const symbol qi = ex_to<symbol> (all_mom[i]);
			const symbol qk = ex_to<symbol> (all_mom[k]);
			v[0] = i;
			v[1] = k;
			LOGXX("(" << qi << " * d/d" << qk << ") * " << INTGeneric(integralfamily_));
			derivative_ops_[v] = OPSUM();
			for (size_t j = 0; j < propagators.size(); j++) {
				// qi (d/dqk)propj^aj = (aj propj^(aj-1)) * (qi dpropj/dqk)
				ex aj = integralfamily_->propagator_exponents()[j];
				OP yjm = OP(-(j + 1)); // decrements aj by one
				ex qi_dpropj_dqk = propagators[j].derivative_contracted(qi, qk);
				qi_dpropj_dqk = qi_dpropj_dqk.expand();
				int opt = subs_options::algebraic;
				qi_dpropj_dqk = qi_dpropj_dqk.subs(s2p, opt);
				qi_dpropj_dqk = qi_dpropj_dqk.subs(s2i, opt);
				OPSUM opsum(yjm, aj);
				opsum *= ex_to_OPSUM(qi_dpropj_dqk, propagators);
				derivative_ops_[v] += opsum;
			}
			LOGXX("  = " << derivative_ops_[v]);
		}
	}
}

void GenericIdentityGenerator::construct_generic_identities() {
	// references to the data
	LinearCombinationGenericList& ibp /*     */= generic_equations_["ibp"];
	LinearCombinationGenericList& ibp_dim /* */= generic_equations_["ibp_dim"];
	LinearCombinationGenericList& ibp_dim_free =
			generic_equations_["ibp_dim_free"];
	LinearCombinationGenericList& ibp_lee /* */= generic_equations_["ibp_lee"];
	LinearCombinationGenericList& lorentz /* */= generic_equations_["lorentz"];

	const Kinematics* kin = integralfamily_->kinematics();
	const lst& loop_mom = integralfamily_->loop_momenta();
	const lst& iext_mom = kin->independent_external_momenta();
	const exmap& s2i = kin->rules_sp_to_invariants();
	const size_t num_loop_mom = loop_mom.nops();
	const size_t num_iext_mom = iext_mom.nops();
	const lst all_mom = add_lst(loop_mom, iext_mom);

	// the generic integral
	const INTGeneric gint(integralfamily_);

	// generate the ibp_
	// (k,i)  ==>  d/dq_k * q_i  (X) == dim (X) + q_i * d/dq_k (X)
	map<vector<int> , OPSUM> ibp_operators;
	vector<int> v(2, 0), dv(2, 0);
	// derivative:
	for (size_t k = 0; k < num_loop_mom; k++) {
		v[0] = dv[1] = k;
		// multiply:
		for (size_t i = 0; i < num_loop_mom + num_iext_mom; i++) {
			v[1] = dv[0] = i;
			OPSUM opsum = derivative_ops_[dv]; // (i, k)
			if (i == k)
				opsum += OPSUM(OPPROD(), kin->dimension());
			LinearCombinationGeneric equibp = gint.shift(opsum);
			ibp.push_back(equibp);
			if (i != k)
				ibp_dim_free.push_back(equibp);
			else
				ibp_dim.push_back(equibp);
			ibp_operators[v] = opsum;
		}
	}
	if (num_loop_mom > 1) {
		// d/dk_i k_{i+1} , k_{l+1} == k_1, k_i = loop mom  ==> NoLM equations
		for (size_t i = 0; i < num_loop_mom; i++) {
			v[0] = i;
			v[1] = (i + 1) % num_loop_mom;
			ibp_lee.push_back(gint.shift(ibp_operators[v]));
		}
		// d/dk_1 p_j , k_1 = loop mom, p_j ext. mom.   ==> NoEM equations
		for (size_t i = 0; i < num_iext_mom; i++) {
			v[0] = 1;
			v[1] = num_loop_mom + i;
			ibp_lee.push_back(gint.shift(ibp_operators[v]));
		}
		// Sum_{i=0}^{NoLM-1} d/dk_i k_i , k_i = loop mom.   ==> 1 equation
		OPSUM opsum;
		for (size_t i = 0; i < num_loop_mom; i++) {
			v[0] = v[1] = i;
			opsum += ibp_operators[v];
		}
		ibp_lee.push_back(gint.shift(opsum));
	}

	// generic lorentz identities

	//  Akj - Ajk; 0 <= k < NoEM,  k+1 <= j < NoEM
	// 	Akj = Sum_{i=0}^{NoEM-1}  ( p_j*p_i   p_k*d/dp_i   -   p_k*p_i   p_j*d/dp_i    )
	vector<int> v1(2, 0), v2(2, 0);
	for (size_t k = num_loop_mom; k < num_iext_mom + num_loop_mom; k++) {
		v1[0] = k;
		for (size_t j = k + 1; j < num_iext_mom + num_loop_mom; j++) {
			v2[0] = j;
			OPSUM opsum;
			// calculate Akj and Ajk
			for (size_t i = num_loop_mom; i < num_iext_mom + num_loop_mom; i++) {
				ex mom_i = all_mom[i];
				ex mom_j = all_mom[j];
				ex mom_k = all_mom[k];
				int opts = subs_options::algebraic;
				ex pjpi = ScalarProduct(mom_j, mom_i).eval().subs(s2i, opts);
				ex pkpi = ScalarProduct(mom_k, mom_i).eval().subs(s2i, opts);
				if (!freeof(pkpi, iext_mom) || !freeof(pjpi, iext_mom))
					ABORT("Failed to generate lorentz identities\n"
							<< pjpi << '\n' << pkpi << '\n' << s2i);
				v1[1] = v2[1] = i;
				OPSUM Akj = OPSUM(OPPROD(), pjpi) * derivative_ops_[v1];
				OPSUM Ajk = OPSUM(OPPROD(), pkpi) * derivative_ops_[v2];
				opsum += Akj - Ajk;
			}
			lorentz.push_back(gint.shift(opsum));
		}
	}
}

LinearCombination GenericIdentityGenerator::get_derivative_wrt_external_momenta(
		const INT& integral, int i, int k) {
	VERIFY(*integral.integralfamily() == *integralfamily_);
	vector<int> v(2);
	v[0] = integralfamily_->loop_momenta().nops() + i;
	v[1] = integralfamily_->loop_momenta().nops() + k;
	INTGeneric genint(integral.integralfamily());
	LinearCombinationGeneric genlc = genint.shift(derivative_ops_[v]);
	return genlc.get_linear_combination(integral);
}

// application of e to integrand is implied (e integrand = OPSUM integrand)
OPSUM ex_to_OPSUM(const GiNaC::ex& e, const vector<Propagator>& propagators) {
	exmap s;
	OPSUM res; // is zero !!!
	if (is_a<add> (e)) {
		for (size_t i = 0; i < e.nops(); ++i)
			res += ex_to_OPSUM(e.op(i), propagators);
	} else if (is_a<mul> (e)) {
		res = OPSUM(OPPROD(), 1); // is one
		for (size_t i = 0; i < e.nops(); ++i)
			res *= ex_to_OPSUM(e.op(i), propagators);
	} else if (e.has(Propagator(wild(1), wild(2), wild(3)))) {
		ex base, expo;
		if (e.match(pow(Propagator(wild(1), wild(2), wild(3)), wild(4)), s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s),//
					wild(3).subs(s));
			expo = wild(4).subs(s);
		} else if (e.match(Propagator(wild(1), wild(2), wild(3)), s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s), //
					wild(3).subs(s));
			expo = 1;
		} else {
			ABORT("Unknown structure containing propagator: " << e);
		}
		// check integer expo > 0 or expo < 0
		if (!(expo.info(info_flags::posint) || expo.info(info_flags::negint)))
			ABORT("Failed to parse propagator exponent in " << e);
		int n = ex_to<numeric> (expo).to_int();
		size_t i;
		for (i = 0; i < propagators.size(); ++i)
			if (base.match(propagators[i]))
				break;
		if (i == propagators.size())
			ABORT("Unknown propagator " << base);
		res = OPSUM(OPPROD(), 1);
		for (size_t j = 0; j < static_cast<size_t> (abs(n)); ++j)
			res *= OPSUM(OPPROD(OP(n > 0 ? (i + 1) : -(i + 1))), 1);
	} else if (!e.has(Propagator(wild(1), wild(2), wild(3)))) {
		res = OPSUM(OPPROD(), e);
	} else {
		ABORT("Failed to parse " << e);
	}
	return res;
}

} // namespace Reduze
