/*  integralfamily.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "integralfamily.h"
#include "kinematics.h"
#include "equationlist.h"
#include "int.h"
#include "functions.h"
#include "sector.h"
#include "ginacutils.h"
#include "yamlutils.h"

using namespace std;

namespace Reduze {

namespace {
YAMLProxy<IntegralFamily> dummy;
}
/*
 struct IntegralFamilyPrivate {
 /// integrals which are preferred
 std::set<INT> preferred_integrals_;
 };

 IntegralFamily& IntegralFamily::operator=(const IntegralFamily& o) {
 if (&o == this)
 return *this;
 delete d;
 d = new IntegralFamilyPrivate(o.d);
 return *this;
 }
 */

IntegralFamily::IntegralFamily() :
		id_(0), kinematics_(0) {
}

IntegralFamily::IntegralFamily(const Kinematics* kin) :
		id_(0), kinematics_(kin) {
}

/* see class Files
 IntegralFamily::IntegralFamily(const IntegralFamily* ic1,
 const IntegralFamily* ic2) :
 Kinematics(*Files::instance()->kinematics()) {
 using namespace GiNaC;

 string name1 = ic1->name();
 string name2 = ic2->name();

 LOGX("Creating product of integral families: " << name1 << ", " << name2);

 name_ = name1 + "_X_" + name2;

 // new symbols for the loop momenta
 exmap old2newA, old2newB;
 lst loop_momenta1 = ic1->loop_momenta();
 lst loop_momenta2 = ic2->loop_momenta();
 int num_mom1 = loop_momenta1.nops();
 int num_mom2 = loop_momenta2.nops();
 int num_mom = num_mom1 + num_mom2;

 for (int i = 1; i <= num_mom1; ++i) {
 symbol s(get_loop_momentum_prefix() + to_string(i));
 loop_momenta_.append(s);
 old2newA[loop_momenta1.op(i - 1)] = s;
 }

 for (int i = num_mom1 + 1; i <= num_mom; ++i) {
 symbol s(get_loop_momentum_prefix() + to_string(i));
 loop_momenta_.append(s);
 old2newB[loop_momenta2.op(i - num_mom1 - 1)] = s;
 }

 vector<Propagator> propagators1 = ic1->get_propagators();
 vector<Propagator> propagators2 = ic2->get_propagators();
 int num_props = propagators1.size() + propagators2.size();
 propagators_.reserve(num_props + num_mom1 * num_mom2);
 vector<Propagator>::const_iterator it;
 for (it = propagators1.begin(); it != propagators1.end(); ++it)
 propagators_.push_back(ex_to<Propagator> (it->subs(old2newA)));
 for (it = propagators2.begin(); it != propagators2.end(); ++it)
 propagators_.push_back(ex_to<Propagator> (it->subs(old2newB)));

 // additional propagators
 for (int i = 0; i < num_mom1; ++i) {
 for (int j = 0; i < num_mom2; ++j) {
 ex mom1 = loop_momenta1.op(i).subs(old2newA);
 ex mom2 = loop_momenta2.op(j).subs(old2newB);
 ex mom = mom1 - mom2;
 ex sqmass = 0;
 propagators_.push_back(Propagator(mom, sqmass));
 }
 }

 int num_props1 = propagators1.size();
 permutation_symmetries_ = ic1->permutation_symmetries_;
 map<int, vector<vector<int> > > permutation_symmetries_2 = ic2->permutation_symmetries_;
 map<int, vector<vector<int> > >::iterator mit;
 for (mit = permutation_symmetries_2.begin(); mit != permutation_symmetries_2.end(); ++mit) {
 vector<vector<int> >::iterator it1;
 for (it1 = mit->second.begin(); it1 != mit->second.end(); ++it1) {
 vector<int>::iterator it2;
 for (it2 = it1->begin(); it2 != it1->end(); ++it2) {
 *it2 += num_props1;
 }
 }
 permutation_symmetries_[permutation_symmetries_.size() + 1] = mit->second;
 }

 init();
 check();
 LOGX("Product integral family: " << name() << " created");
 }
 */

IntegralFamily::IntegralFamily(const Kinematics* kin, const std::string& name,
		const GiNaC::lst& loop_momenta,
		const std::vector<Propagator>& propagators) :
		name_(name), id_(0), kinematics_(kin), loop_momenta_(loop_momenta), propagators_(
				propagators) {
	init();
}

IntegralFamily::~IntegralFamily() {
}

void IntegralFamily::set_id(int id) {
	id_ = id;
}

GiNaC::lst IntegralFamily::get_all_momenta() const {
	return add_lst(loop_momenta_, kinematics_->external_momenta());
}

int IntegralFamily::get_t(int sector_id) const {
	return Sector::t_of_id(sector_id);
}

size_t IntegralFamily::get_num_scalar_products() const {
	int NoLM = loop_momenta().nops();
	int NoEM = kinematics_->independent_external_momenta().nops();
	int res = (NoLM + 1) * NoLM;
	res /= 2;
	res += NoLM * NoEM;
	return res;
}

set<int> IntegralFamily::get_immediate_subsectors(int sector_id) const {
	set<int> subsectors;
	unsigned u_id = sector_id;
	for (unsigned i = 0; i < num_propagators(); ++i)
		subsectors.insert(u_id & ~(1 << i));
	subsectors.erase(u_id);
	return subsectors;
}

std::set<int> IntegralFamily::get_subsector_equivalents(int sector_id,
		bool recurse) const {
	set<int> res;
	stack<int> todo;
	int next = get_equivalent(sector_id);
	do {
		set<int> subs = get_immediate_subsectors(next);
		for (set<int>::const_iterator s = subs.begin(); s != subs.end(); ++s) {
			int equiv = get_equivalent(*s);
			if (res.insert(equiv).second)
				todo.push(equiv);
		}
	} while (recurse && !todo.empty() && (next = todo.top(), todo.pop(), true));
	return res;
}

void IntegralFamily::read(const YAML::Node& node) {
	using namespace GiNaC;

	LOGXX("Set up integral family from YAML config");
	verify_yaml_spec(node);

	node["name"] >> name_;
	propagators_.clear();
	cut_propagators_.clear();

	list < string > loop_momenta_str;
	node["loop_momenta"] >> loop_momenta_str;
	loop_momenta_ = create_symbols(loop_momenta_str);

	LOGXX("Parsing propagators");
	if (node.FindValue("propagators") != 0) {
		const YAML::Node& props_node = node["propagators"];
		for (YAML::Iterator n = props_node.begin(); n != props_node.end();
				++n) {
			const YAML::Node* np = 0;
			string type;
			if (n->GetType() == YAML::CT_SEQUENCE && n->size() == 2) {
				type = "standard";
				np = &*n;
			} else if (n->GetType() == YAML::CT_MAP && n->size() == 1) {
				n->begin().first() >> type;
				LOG("got type " + type);
				np = &n->begin().second();
			} else {
				throw runtime_error(
						"unknown entry in propagators " + position_info(*n));
			}
			string str;
			ex k1, k2, msq;
			const GiNaC::exmap& mom_conserv = //
					kinematics_->rule_momentum_conservation();
			if (type == "standard") { // standard propagator with k^2
				(*np)[0] >> str;
				k1 = k2 = ex(str, get_all_momenta()).subs(mom_conserv);
				(*np)[1] >> str;
				msq = ex(str, kinematics_->kinematic_invariants());
				propagators_.push_back(Propagator(k1, k2, msq));
			} else if (type == "bilinear") { // generalised propagator with k1.k2
				(*np)[0][0] >> str;
				k1 = ex(str, get_all_momenta()).subs(mom_conserv);
				(*np)[0][1] >> str;
				k2 = ex(str, get_all_momenta()).subs(mom_conserv);
				(*np)[1] >> str;
				msq = ex(str, kinematics_->kinematic_invariants());
				propagators_.push_back(Propagator(k1, k2, msq));
			} else {
				throw runtime_error(
						"unknown propagator type '" + type + "' "
								+ position_info(*n));
			}
		}
	}

	LOGXX("Parsing permutation symmetries");
	if (node.FindValue("permutation_symmetries")) {
		const YAML::Node& plist = node["permutation_symmetries"];
		if (plist.GetType() != YAML::CT_SEQUENCE)
			throw runtime_error(
					"Node is not a sequence " + position_info(plist));
		for (YAML::Iterator p = plist.begin(); p != plist.end(); ++p) {
			Permutation permutation;
			*p >> permutation;
			vector < vector<int> > perm = permutation.vector_rep();
			// internal numbering of the propagators starts with 0
			add_to_each_element(perm, -1);
			permutation_symmetries_[permutation_symmetries_.size() + 1] = perm;
		}
	}

	init();
}

void IntegralFamily::init() {
	using namespace GiNaC;
	try {
		LOGXX("Checking propagators");
		check_propagators();

		LOGXX("Init global propagator permutations");
		LOGXX("Verifying global propagator permutations");
		verify_permutation_symmetries();
		// complete the global permutations
		PermutationSet aset;
		map<int, vector<vector<int> > >::const_iterator p;
		for (p = permutation_symmetries_.begin();
				p != permutation_symmetries_.end(); ++p)
			aset.insert(Permutation(p->second));
		LOGXX(
				aset.pset().size() << " user-defined global permutation symmetries");
		aset.complete_permutations();
		LOGXX(
				aset.pset().size() << " global permutation symmetries after completion.");
		permutation_symmetries_.clear();
		const set<Permutation>& pset = aset.pset();
		int lauf(0);
		for (set<Permutation>::const_iterator s = pset.begin(); s != pset.end();
				++s)
			permutation_symmetries_[++lauf] = s->vector_rep();
#ifdef DEBUG
		LOGXX("Checking global permutation symmetries after completing the permutations");
		verify_permutation_symmetries();
#endif

		LOGXX("Init symbols for generic propagator exponents");
		propagator_exponent_symbols_.remove_all();
		propagator_exponents_.assign(num_propagators(), 0);
		for (size_t i = 0; i < num_propagators(); ++i) {
			propagator_exponents_[i] = GiNaC::symbol("a" + to_string(i + 1));
			propagator_exponent_symbols_.append(propagator_exponents_[i]);
		}

		LOGXX("Init mapping from scalar products to propagators");
		lst eqns; // equations: 1/Propagator(k, m2) == (ScalarProduct(k,k)-m2)
		for (size_t i = 0; i < num_propagators(); ++i) {
			// get propagator <-> sum of scalar product relation
			ex p = propagators_[i].to_explicite_form();
			p = p.expand().subs(kinematics_->rules_sp_to_invariants());
			eqns.append((1 / propagators_[i]) == 1 / p);
		}
		try {
			LOGXX("solving:\n" <<eqns);
			rules_sp_to_prop_ = lsolve_pattern(eqns,
					ScalarProduct(wild(1), wild(2)));
			LOGXX("solved");
		} catch (exception& e) {
			throw runtime_error(
					string("Can't determine scalar products from propagators\n")
							+ e.what());
		}
		exmap::iterator r;
		for (r = rules_sp_to_prop_.begin(); r != rules_sp_to_prop_.end(); ++r)
			r->second = r->second.expand(); // expand (propi+propj)/(propi*propj)

	} catch (exception& e) {
		throw std::runtime_error(
				"invalid integral family: " + name_ + ":\n" + e.what());
	}
}

void IntegralFamily::print(YAML::Emitter& os) const {
	using namespace YAML;
	os << BeginMap;

	os << Key << "name" << Value << name_;
	os << Key << "loop_momenta" << Value << loop_momenta_;

	os << Key << "propagators" << Value << BeginSeq;
	for (size_t i = 0; i < propagators_.size(); ++i) {
		GiNaC::ex k1 = propagators_[i].scalarproduct().op(0);
		GiNaC::ex k2 = propagators_[i].scalarproduct().op(1);
		// quick hack for now: we have only two types of propagators
		if (k1.is_equal(k2)) {
			os << Flow << BeginMap << Key << "standard" << Value;
			os << BeginSeq << k1 << propagators_[i].squaredmass() << EndSeq;
			os << EndMap;
		} else {
			os << Flow << BeginMap << Key << "bilinear" << Value;
			os << BeginSeq << BeginSeq << k1 << k2 << EndSeq;
			os << propagators_[i].squaredmass() << EndSeq;
			os << EndMap;
		}
	}
	os << EndSeq;

	os << Key << "permutation_symmetries" << Value << BeginSeq;
	map<int, vector<vector<int> > >::const_iterator s;
	for (s = permutation_symmetries_.begin();
			s != permutation_symmetries_.end(); ++s) {
		os << Flow << BeginSeq;
		vector<vector<int> >::const_iterator c;
		for (c = s->second.begin(); c != s->second.end(); ++c) {
			os << BeginSeq;
			vector<int>::const_iterator i;
			for (i = c->begin(); i != c->end(); ++i)
				// YAML config counts propagators from 1, we count from 0
				os << (*i + 1);
			os << EndSeq;
		}
		os << EndSeq;
	}
	os << EndSeq;
	os << EndMap;
}

const Kinematics* IntegralFamily::kinematics() const {
	return kinematics_;
}

int IntegralFamily::get_equivalent(int sector_id) const {
	return (INT(this, sector_id)).get_equivalent().id();
}

std::set<int> IntegralFamily::get_trivial_zero_sector_equivalents() const {
	// \todo optimize this function
	int num_loops = loop_momenta().nops();
	if (num_loops == 0)
		return set<int>();
	int max = max_sector_id();
	set<int> equivs = get_subsector_equivalents(max, true);
	equivs.insert(max);
	for (set<int>::iterator s = equivs.begin(); s != equivs.end();)
		if (Sector::t_of_id(*s) >= num_loops)
			equivs.erase(s++);
		else
			++s;
	return equivs;
}

bool IntegralFamily::match_integrand(const GiNaC::ex& integrand, INT& integral,
		GiNaC::ex& coeff) const {
	using namespace GiNaC;
	integral = INT(this);
	coeff = 1;
	lst loop_moms = loop_momenta();
	lst powers;
	if (is_a < mul > (integrand)) {
		for (size_t i = 0; i < integrand.nops(); ++i)
			powers.append(integrand.op(i));
	} else if (is_a < power > (integrand) || is_a < Propagator > (integrand)
			|| freeof(integrand, loop_moms)) {
		powers.append(integrand);
	} else {
		return false;
	}
	vector<int8> v(propagators_.size(), 0);
	for (lst::const_iterator p = powers.begin(); p != powers.end(); ++p) {
		if (freeof(*p, loop_momenta())) {
			// currently we assume that all propagators should be matched
			if (p->has(Propagator(wild(1), wild(2), wild(3))))
				ABORT(
						"encountered " << *p << " with propagator free of " << name_ << " loop momenta");
			coeff *= *p;
		} else {
			exmap s;
			ex base, expo;
			if (p->match(pow(Propagator(wild(1), wild(2), wild(3)), wild(4)),
					s)) {
				base = Propagator(wild(1).subs(s), wild(2).subs(s),
						wild(3).subs(s));
				expo = wild(4).subs(s);
			} else if (p->match(Propagator(wild(1), wild(2), wild(3)), s)) {
				base = Propagator(wild(1).subs(s), wild(2).subs(s), //
				wild(3).subs(s));
				expo = 1;
			} else {
				return false;
			}
			ASSERT(!expo.is_zero());
			if (!(expo.info(info_flags::posint || expo.info(info_flags::negint))))
				return false;
			//if (!(expo.info(info_flags::integer)))
			//	return false;
			int num_expo = ex_to < numeric > (expo).to_int();
			size_t i;
			for (i = 0; i < propagators_.size(); ++i) {
				// if (base.match(propagators_[i])) {
				//ASSERT(is_a<Propagator>(base));
				if (base == propagators_[i]) {
					v[i] += num_expo;
					break;
				}
			}
			//LOG("problematic: " << base);
			if (i == propagators_.size())
				return false;
		}
	}
	integral = (INT(this, v)).get_equivalent();
	return true;
}

void print_global_permutation_symmetry(std::ostream& os,
		const std::vector<std::vector<int> >& cycle) {
	using namespace YAML;
	Emitter ye;
	ye << Flow << BeginSeq;
	vector<vector<int> >::const_iterator c;
	for (c = cycle.begin(); c != cycle.end(); ++c) {
		ye << BeginSeq;
		vector<int>::const_iterator i;
		for (i = c->begin(); i != c->end(); ++i)
			// YAML config counts propagators from 1, we count from 0
			ye << (*i + 1);
		ye << EndSeq;
	}
	ye << EndSeq;
	os << ye.c_str();
}

bool IntegralFamily::have_non_standard_propagators() const {
	vector<Propagator>::const_iterator p;
	for (p = propagators_.begin(); p != propagators_.end(); ++p)
		if (!p->has_unique_momentum())
			return true;
	return false;
}

void IntegralFamily::verify_permutation_symmetries() const {
	using namespace GiNaC;
	const exmap& mom_cons = kinematics()->rule_momentum_conservation();
	const lst& iextm = kinematics()->independent_external_momenta();
	exmap sym2alt, alt2sym;
	lst altloopsym, altextsym;
	provide_alternative_symbols(loop_momenta(), altloopsym, sym2alt, alt2sym);
	provide_alternative_symbols(iextm, altextsym, sym2alt, alt2sym);

	map<int, vector<vector<int> > >::const_iterator p;
	p = permutation_symmetries_.begin();
	for (; p != permutation_symmetries_.end(); ++p) {
		ostringstream yamlperm_stream;
		print_global_permutation_symmetry(yamlperm_stream, p->second);
		string yaml_perm = yamlperm_stream.str();
		string yaml_perm_error = "invalid permutation symmetry: "
				+ yamlperm_stream.str() + "\n";

		// check the propagators in the permutation exist
		vector<vector<int> >::const_iterator c;
		for (c = p->second.begin(); c != p->second.end(); ++c)
			for (vector<int>::const_iterator n = c->begin(); n != c->end(); ++n)
				if (*n < 0 || *n >= (int) propagators_.size())
					throw runtime_error(
							yaml_perm_error + "propagator " + to_string(*n + 1)
									+ " does not exist");

		// 2 equations for each relation prop1 -> prop2 (+/- sign)
		vector < list<ex> > eqns_list;
		// consistency check of the permutations and setup of the equations
		for (c = p->second.begin(); c != p->second.end(); ++c) {
			VERIFY(c->size() > 1);
			for (vector<int>::const_iterator n = c->begin(); n != c->end();) {
				//lhs
				bool is_lhs_cut =
						cut_propagators_.find(*n) != cut_propagators_.end() ? true :
								false;
				bool is_lhs_standard =
						propagators_[*n].op(0).is_equal(
								propagators_[*n].op(1)) ? true : false;
				ex mom_lhs;
				if (is_lhs_standard)
					mom_lhs = propagators_[*n].momentum().subs(mom_cons);
				ex sqmass_lhs = propagators_[*n].squaredmass();

				++n;
				n = (n == c->end() ? c->begin() : n);

				// rhs
				bool is_rhs_cut =
						cut_propagators_.find(*n) != cut_propagators_.end() ? true :
								false;
				bool is_rhs_standard =
						propagators_[*n].op(0).is_equal(
								propagators_[*n].op(1)) ? true : false;
				ex mom_rhs;
				if (is_rhs_standard)
					mom_rhs = propagators_[*n].momentum().subs(mom_cons).subs(
							sym2alt);
				ex sqmass_rhs = propagators_[*n].squaredmass();

				// throw exception for impossible cases
				if (is_lhs_cut != is_rhs_cut)
					throw runtime_error(
							yaml_perm_error
									+ "permutes cut with non-cut propagators");
				if (is_lhs_standard != is_rhs_standard)
					throw runtime_error(
							yaml_perm_error
									+ "permutes standard with non-standard propagators");
				if (!sqmass_lhs.is_equal(sqmass_rhs))
					throw runtime_error(
							yaml_perm_error
									+ "permutes propagators with different masses");

				eqns_list.push_back(list<ex>());
				eqns_list.back().push_back(relational(mom_lhs, mom_rhs));
				eqns_list.back().push_back(relational(mom_lhs, -mom_rhs));
				if (n == c->begin())
					break;
			}
		} // loop over cycles in the permutation

		// no real verification of the symmetry when non-standard propagators are present
		if (have_non_standard_propagators()) {
			WARNING(
					"Non-standard propagators encountered: cannot verify" << " permutation symmetry: " << yaml_perm << "\n");
			continue;
		}

		// add the equations of the momenta of the props not involved in the permutation
		set<int> permuted_props;
		for (c = p->second.begin(); c != p->second.end(); ++c)
			permuted_props.insert(c->begin(), c->end());
		for (unsigned i = 0; i < propagators_.size(); ++i) {
			if (permuted_props.find(i) == permuted_props.end()) {
				ex mom = propagators_[i].momentum().subs(mom_cons);
				eqns_list.push_back(list<ex>());
				eqns_list.back().push_back(relational(mom, mom.subs(sym2alt)));
				eqns_list.back().push_back(relational(mom, -mom.subs(sym2alt)));
			}
		}

		// find a shift to verify the permutation

		combinations<ex> combs(eqns_list);
		vector<ex> eqns_vec;
		bool perm_verified = false;
		while (combs.get_next(eqns_vec)) {
			lst eqns, sol_loop, sol_ext;
			for (vector<ex>::const_iterator v = eqns_vec.begin();
					v != eqns_vec.end(); ++v)
				eqns.append(*v);
			const lst sym = add_lst(loop_momenta(), iextm);
			LOGXX(
					"Finding solutions for " << sym << " from equations " << eqns);
			ex solved = GiNaC::lsolve(eqns, sym);
			LOGXX("solution: " << solved);
			VERIFY(GiNaC::is_a<lst>(solved));
			if (solved.nops() == 0)
				continue;
			GiNaC::const_iterator r;
			for (r = solved.begin(); r != solved.end(); ++r) {
				VERIFY(GiNaC::is_a<relational>(*r));
				GiNaC::relational rel = GiNaC::ex_to < relational > (*r);
				VERIFY(GiNaC::is_a<symbol>(rel.lhs()));
				if (kinematics()->external_momenta().has(rel.lhs()))
					if (freeof(rel.rhs(), loop_momenta()))
						sol_ext.append(rel);
					else
						break;
				else
					sol_loop.append(rel);
			}
			if (r != solved.end()) {
				LOGXX(
						"   invalid solution: found loop momenta in right hand side of external momenta transformations.");
				continue;
			}
			ex det = GiNaC::abs(jacobi_determinant(sol_loop, altloopsym));
			if (!det.is_equal(1)) {
				LOGXX("   invalid solution: |det| = " << det);
				continue;
			}
			list<Crossing> crossings = kinematics()->ordered_crossings();
			crossings.push_front(Crossing(kinematics()));
			list<Crossing>::const_iterator cr;
			for (cr = crossings.begin(); cr != crossings.end(); ++cr) {
				if (!cr->is_equivalent_to_identity())
					continue;
				ex identity = ex(sol_ext).subs(cr->rules_momenta()).subs(
						alt2sym);
				GiNaC::const_iterator it;
				for (it = identity.begin(); it != identity.end(); ++it)
					if (!(it->op(0) - it->op(1)).expand().is_zero())
						break;
				if (it == identity.end())
					break; // valid crossing
			}
			if (cr == crossings.end()) {
				LOGXX("   invalid solution: invalid crossing.");
				continue;
			}
			LOGXX("Verified global permutation symmetry: " << yaml_perm);
			LOGXX("  Input equations from propagators:");
			LOGXX("  " << ex(eqns).subs(alt2sym));
			LOGXX("  Shift found (leaves kinematic invariants unchanged):");
			LOGXX("  " << ex(solved).subs(alt2sym));
			perm_verified = true;
			break;
		}
		if (!perm_verified)
			throw runtime_error(
					yaml_perm_error
							+ "no shift found to verify the permutation");
	} // loop over all global permutations
}

void IntegralFamily::check_propagators() const {
	//using namespace GiNaC;
	stringstream msg;
	size_t NoP = num_propagators();
	size_t NoLM = loop_momenta().nops();
	size_t NoEM = kinematics_->independent_external_momenta().nops();
	// check for multiple defined propagators
	GiNaC::lst props_unique;
	for (size_t i = 0; i < NoP; ++i)
		props_unique.append(propagators_[i]);
	props_unique.sort().unique();
	size_t NoP_unique = props_unique.nops();
	if (NoP != NoP_unique) {
		stringstream ss;
		for (size_t i = 0; i < NoP; ++i)
			ss << propagators_[i] << " ";
		throw runtime_error(
				"Multiple definitions of the same propagator: " + ss.str());
	}
	// check for the correct number of propagators
	if (NoP != get_num_scalar_products()) {
		msg << "With " << NoLM << " loop momenta " << loop_momenta() << " and "
				<< NoEM << " independent external momenta "
				<< kinematics_->independent_external_momenta()
				<< " one needs an integral family with "
				<< get_num_scalar_products() << " propagators.";
		throw runtime_error(msg.str());
	}
	// check if propagators contain loop momenta
	for (size_t i = 0; i < NoP; ++i) {
		const Propagator& p = propagators_[i];
		if (!freeof(p.squaredmass(), get_all_momenta())) {
			msg << "Propagator " << p << " has momentum dependent mass^2.";
			throw runtime_error(msg.str());
		} else if (freeof(p, loop_momenta())) {
			msg << "Propagator " << p << " contains no loop momenta\n" << "("
					<< loop_momenta() << ")";
			throw runtime_error(msg.str());
		}
	}
	// check if mass dimensions of propagators is 2
	for (size_t i = 0; i < NoP; ++i) {
		try {
			GiNaC::ex msq = propagators_[i].squaredmass();
			if (!msq.is_zero() && kinematics_->find_mass_dimension(msq) != 2)
				WARNING(
						"integral family " << name_ << ":" << " squared mass of propagator " << i << " specified as " << propagators_[i].squaredmass() << " has mass dimension different from 2");
		} catch (exception& e) {
			WARNING(
					"can't determine mass dimension of " << propagators_[i].squaredmass());
		}
	}
}

string IntegralFamily::t_tree_nice_string(const set<int>& sector_ids) const {
	ostringstream result;
	map<int, set<int> > ids_by_t;
	for (set<int>::const_iterator s = sector_ids.begin(); s != sector_ids.end();
			++s)
		ids_by_t[get_t(*s)].insert(*s);
	map<int, set<int> >::const_iterator t;
	for (t = ids_by_t.begin(); t != ids_by_t.end(); ++t) {
		result << "t = " << t->first << ":\n";
		const set<int>& ids = t->second;
		for (set<int>::const_iterator i = ids.begin(); i != ids.end();) {
			result << *i;
			if (++i != ids.end())
				result << ", ";
		}
		result << "\n";
	}
	return result.str();
}

// CrossedIntegralFamily members

CrossedIntegralFamily::CrossedIntegralFamily(const IntegralFamily* source_ic,
		const Crossing& crossing) : //
		IntegralFamily(source_ic->kinematics()), //
		source_integralfamily_(source_ic), //
		crossing_(crossing) {
	using namespace GiNaC;

	string name = crossing.name_for_crossed_family(source_ic);

	vector<Propagator> props = source_ic->propagators();
	vector<Propagator>::iterator p;
	for (p = props.begin(); p != props.end(); ++p) {
		ex tmp = p->subs(crossing.rules_momenta());
		tmp = tmp.eval(); // bring Propagator into canonical form
		ASSERT(is_a<Propagator>(tmp));
		*p = ex_to < Propagator > (tmp);
	}

	*dynamic_cast<IntegralFamily*>(this) = //
			IntegralFamily(source_ic->kinematics(), name,
					source_ic->loop_momenta(), props);
}

CrossedIntegralFamily::~CrossedIntegralFamily() {
}

void sequel(vector<int8> & vec, int8 sum, int8 min, int level) {
	// calculate the value of the first entry: sum - values of second ... last entries
	int8 first = sum;
	int length = static_cast<int>(vec.size());
	for (int i = 1; i < length; ++i) {
		first -= vec[i];
	}
	if (level == 0) {
		vec[0] = first;
		return;
	}
	if (first > min) {
		// increase entry at position level and set first entry
		vec[level]++;
		vec[0] = first - 1;
		return;
	} else {
		// set last entry to min and call the function on the next smaller level
		vec[level] = min;
		sequel(vec, sum, min, level - 1);
	}
}

void partition(vector<vector<int8> > & array, size_t lim, int length, int sum,
		bool bu) {
	if (length == 0 || lim == 0) {
		return;
	}
	int8 min, max;
	if (bu) { // for propagators in the denomoinator
		min = 1;
		max = static_cast<int8>(sum - length + 1);
	} else { // for propagators in the numerator
		min = 0;
		max = static_cast<int8>(sum);
	}
	// build the first vector {max, min, ..., min}
	array[0][0] = max;
	for (int i = 1; i < length; ++i) {
		array[0][i] = min;
	}
	// build all other vectors
	for (size_t i = 1; i < lim; ++i) {
		array[i] = array[i - 1];
		sequel(array[i], static_cast<int8>(sum), min, length - 1);
	}
}

}
// namespace Reduze
