/*  job_computedifferentialequations.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_computedifferentialequations.h"
#include "functions.h"
#include "files.h"
#include "filedata.h"
#include "ginacutils.h"
#include "streamutils.h"
#include "kinematics.h"
#include "integralfamily.h"
#include "equation.h"
#include "yamlutils.h"
#include "sector.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// register job type at JobFactory
namespace {
JobProxy<ComputeDifferentialEquations> dummy;
}

bool ComputeDifferentialEquations::find_dependencies(
		const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {
	//find_dependencies_all_sectormappings(outothers, in, auxjobs);
	find_dependencies_reductions(outothers, in);
	out.push_back(output_filename_);
	return true;
}

void ComputeDifferentialEquations::print_manual_options(YAML::Emitter& os) const {
	using namespace YAML;
	os << Key << "differentiation_invariants" << Value
			<< differentiation_invariants_;
	os << Key << "auxiliary_invariants" << Value << auxiliary_invariants_;
}

void ComputeDifferentialEquations::read_manual_options(const YAML::Node& node) {
	if (output_filename_.empty())
		throw runtime_error("output file undefined");
	Kinematics *kin = Files::instance()->kinematics();
	Reduze::read<GiNaC::ex>(node["differentiation_invariants"],
			differentiation_invariants_, add_lst(kin->kinematic_invariants(),
					kin->external_momenta()));
	if (node.FindValue("auxiliary_invariants"))
		Reduze::read<GiNaC::ex>(node["auxiliary_invariants"],
				auxiliary_invariants_, add_lst(kin->kinematic_invariants(),
						kin->external_momenta()));
	size_t nmom = kin->independent_external_momenta().nops();
	size_t nreq = (nmom * (nmom + 1)) / 2;
	if (differentiation_invariants_.nops() + auxiliary_invariants_.nops()
			!= nreq)
		throw runtime_error("differentiation_invariants must contain "
				+ to_string(nreq) + " elements");
}

std::string ComputeDifferentialEquations::get_description() const {
	return string("compute differential equations");
}

/// returns (q * d/dk f) for two momenta q and k
/** k must enter only via ScalarProducts, not Propagators **/
static ex derivative_contracted(const ex& f, const symbol& q, const symbol& k) {
	ASSERT(!f.has(Propagator(wild(1), wild(2))));
	ASSERT(!f.has(Propagator(wild(1), wild(2), wild(3))));
	// search ScalarProducts in f and introduce abbreviations for them
	exset found;
	f.find(ScalarProduct(wild(1), wild(2)), found);
	exmap sp2tmp, tmp2sp;
	for (exset::const_iterator s = found.begin(); s != found.end(); ++s) {
		symbol tmpvar;
		sp2tmp[*s] = tmpvar;
		tmp2sp[tmpvar] = *s;
	}
	ex foftmp = f.subs(sp2tmp);
	if (foftmp.has(k))
		ERROR("can't take derivative of " << foftmp << " w.r.t. " << k);
	// take (d f)/(d pk) = (d f)/(d scalarproduct_i) (d scalarproduct_i)/(d pk)
	ex deriv = 0;
	exmap::const_iterator sv;
	for (sv = sp2tmp.begin(); sv != sp2tmp.end(); ++sv) {
		if (!is_a<ScalarProduct> (sv->first) || //
				!is_a<symbol> (sv->second))
			ERROR("failed to substitute scalar products");
		ScalarProduct sprod = ex_to<ScalarProduct> (sv->first);
		ex lhs = sprod.op(0);
		ex rhs = sprod.op(1);
		symbol var = ex_to<symbol> (sv->second);
		deriv += foftmp.diff(var) * (//
				/*      */ScalarProduct(q * lhs.diff(k), rhs)//
						+ ScalarProduct(lhs, q * rhs.diff(k))//
				);
	}
	deriv = deriv.subs(tmp2sp);
	return deriv;
}

/// compute pj dI/dpk where masses are functions of the momenta
// note: we can't use Propagator::derivative_contracted since Propagator
//       contains mass only linearly
static LinearCombination compute_pj_dI_dpk(const INT& i, const symbol& pj,
		const symbol& pk) {
	const IntegralFamily* ic = i.integralfamily();
	const exmap inv2sp = ic->kinematics()->find_rules_invariants_to_sp();
	ex f = i.get_integrand();
	exset found;
	f.find(Propagator(wild(1), wild(2)), found);
	f.find(Propagator(wild(1), wild(2), wild(3)), found);
	ex fexpl = f;
	for (exset::const_iterator s = found.begin(); s != found.end(); ++s) {
		ASSERT(is_a<Propagator>(*s));
		const Propagator& prop = ex_to<Propagator> (*s);
		fexpl = fexpl.subs(prop == prop.to_explicite_form().subs(inv2sp));
	}
	ex deriv = derivative_contracted(fexpl, pj, pk);
	const exmap& sp2inv = ic->kinematics()->rules_sp_to_invariants();
	const exmap& sp2prop = ic->rules_sp_to_prop();
	deriv = deriv.expand().subs(sp2inv).subs(sp2prop).expand();
	LinearCombination lc;
	propagator_products_to_integrals(deriv, lc, ic);
	return lc;
}

void ComputeDifferentialEquations::run_serial() {
	LOG("Generating differential equations for integrals in " << integrals_filename_);
	OutFileLinearCombinations out(output_filename_.c_str());

	LOG("Reading integrals to compute DEQ for from " + integrals_filename_);
	std::list<INT> integrals;
	InFileINTs in(integrals_filename_);
	while (in.good())
		integrals.push_back(InFileINTs::get(in));
	LOG("  found " << integrals.size() << " integrals");

	list<INT>::const_iterator i;
	for (i = integrals.begin(); i != integrals.end(); ++i) {
		LOG("\nProcessing " << *i << " from sector " << i->get_sector());

		// basic idea of the algorithm:
		//   compute: p_j (d I)/(d p_k) = p_j (d s_l)/(d p_k) (d I)/(d s_l),
		//            where s_l are functions of kinematic invariants,
		//   solve for (d I)/(d s_l) (over-determined system),
		//   reduce the rhs with available reductions, e.g. IBPs

		const IntegralFamily* ic = i->integralfamily();
		vector<symbol> p;
		const lst& pin = ic->kinematics()->independent_external_momenta();
		p.reserve(pin.nops());
		for (lst::const_iterator pi = pin.begin(); pi != pin.end(); ++pi) {
			if (!is_a<symbol> (*pi))
				ERROR("external momentum is no symbol: " << *pi);
			p.push_back(ex_to<symbol> (*pi));
		}

		// prepare differentiation invariants
		LOG("Differentiation invariants specified as:\n " << differentiation_invariants_);
		LOG("Auxiliary invariants specified as:\n " << auxiliary_invariants_);
		lst all_invariants = add_lst(differentiation_invariants_,
				auxiliary_invariants_);
		lst s = all_invariants;
		if (!freeof(s, ic->kinematics()->kinematic_invariants())) {
			const exmap inv2sp = ic->kinematics()->find_rules_invariants_to_sp();
			lst sold;
			swap(s, sold);
			for (lst::const_iterator sl = sold.begin(); sl != sold.end(); ++sl)
				s.append(sl->subs(inv2sp));
			if (!freeof(s, ic->kinematics()->kinematic_invariants()))
				ERROR("failed to replace all invariants by scalar products in: " << s
						<< "\ninvariants should only use left hand sides in: " << inv2sp);
		}
		lst sold;
		swap(s, sold);
		for (lst::const_iterator sl = sold.begin(); sl != sold.end(); ++sl)
			s.append(sl->subs(ic->kinematics()->rule_momentum_conservation()));
		LOG("All off-shell invariants as functions of momenta (please check!):\n " << s);

		// compute p_j (d s_l)/(d p_k)
		// row index: build from j and k according to j*size(p) + k, column index: l
		LOG("Differentiating invariants w.r.t. momenta...");
		const exmap& sp2inv = ic->kinematics()->rules_sp_to_invariants();
		matrix pjdsldpk(p.size() * p.size(), s.nops());
		for (size_t j = 0; j < p.size(); ++j)
			for (size_t l = 0; l < s.nops(); ++l)
				for (size_t k = 0; k < p.size(); ++k) {
					ex deriv = derivative_contracted(s.op(l), p[j], p[k]);
					deriv = deriv.expand().subs(sp2inv);
					LOGX(" " << p[j] << " d/d"<< p[k] << " " << s.op(l) << " = " << deriv);
					pjdsldpk(j * p.size() + k, l) = deriv;
				}

		// solve system for dI/dsl
		LOG("Solving for derivatives w.r.t. invariants...");
		lst dIdsl_syms, pjdIdpk_syms; // symbols for system solving
		for (size_t j = 0; j < p.size(); ++j)
			for (size_t k = 0; k < p.size(); ++k) {
				string pj_str = to_string(p[j]);
				string pk_str = to_string(p[k]);
				pjdIdpk_syms.append(symbol(pj_str + "dId" + pk_str));
			}
		for (size_t l = 0; l < s.nops(); ++l) {
			stringstream ss;
			i->to_mma_stream(ss);
			string int_str = to_safe_variable_name(ss.str());
			const ex& sl = all_invariants[l];
			string var_str = to_safe_variable_name(to_string(sl));
			dIdsl_syms.append(symbol("d" + int_str + "d" + var_str));
		}
		lst eqns;
		for (size_t j = 0; j < p.size(); ++j) {
			for (size_t k = 0; k < p.size(); ++k) {
				ex expr = 0;
				for (size_t l = 0; l < s.nops(); ++l)
					expr += pjdsldpk(j * p.size() + k, l) * dIdsl_syms.op(l);
				eqns.append(pjdIdpk_syms[j * p.size() + k] == expr);
			}
		}
		unsigned opts = solve_algo::gauss; // avoid division problems with sqrt
		LOGX(" all eqs: " << eqns);
		lst sol = ex_to<lst> (lsolve_overdetermined(eqns, dIdsl_syms, opts));
		LOGX(" found solutions:  " << sol);
		// assert non-ambigious solutions
		lst::const_iterator solit, symit;
		for (solit = sol.begin(); solit != sol.end(); ++solit)
			for (symit = dIdsl_syms.begin(); symit != dIdsl_syms.end(); ++symit)
				if (solit->op(1).has(*symit))
					ERROR("can't solve for requested derivatives (underdetermined system)");

		// actually compute p_j (d I)/(d p_k)
		LOG("Differentiating integral w.r.t. momenta");
		vector<LinearCombination> pjdIdpk(p.size() * p.size());
		for (size_t j = 0; j < p.size(); ++j)
			for (size_t k = 0; k < p.size(); ++k) {
				LinearCombination deriv = compute_pj_dI_dpk(*i, p[j], p[k]);
				pjdIdpk[j * p.size() + k] = deriv;
				stringstream ss;
				i->to_mma_stream(ss);
				deriv.to_mma_stream(ss);
				LOGX(" " << p[j] << " d/d" << p[k] << " " << ss.str());
			}

		// insert the integral expressions
		LOG("Inserting linear combinations of integrals");
		int ntodo = differentiation_invariants_.nops(); // discard aux derivs
		for (lst::const_iterator s = sol.begin(); ntodo > 0 && s != sol.end(); ++s, --ntodo) {
			ex sym = s->op(0);
			ex todo = s->op(1).expand();
			//LOGX("trying to convert: " << todo);
			LinearCombination dIds;
			for (size_t j = 0; j < p.size(); ++j) {
				for (size_t k = 0; k < p.size(); ++k) {
					ex sym = pjdIdpk_syms[j * p.size() + k];
					ex coeff = todo.coeff(sym, 1);
					//LOGX("processing: " << sym << " with coeff: " << coeff);
					const LinearCombination& d = pjdIdpk[j * p.size() + k];
					LinearCombination::const_iterator t;
					for (t = d.begin(); t != d.end(); ++t)
						dIds.insert(t->first, t->second * coeff, true /*norm*/);
					todo = todo - coeff * sym;
				}
			}
			todo = todo.normal().expand();
			if (!todo.is_zero())
				ERROR("encountered non-linear or inhomogenous terms w.r.t. pjdIdpk:\n" << todo);
			dIds.set_name(to_string(sym));
			out << dIds;
		}
	}
	out.finalize();
}

} // namespace Reduze

