/*  job_finddiagramshifts.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_finddiagramshifts.h"
#include "functions.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "files.h"
#include "diagram.h"
#include "graph.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "int.h"

using namespace std;

namespace Reduze {

typedef std::pair<GraphLabel, GraphRelabeling> LabelPair;

// register job type at JobFactory
namespace {
JobProxy<FindDiagramShifts> dummy;
}

size_t FindDiagramShifts::get_graphs_of_sectors(std::map<int, std::map<int,
		std::set<SectorGL> > >& secbytbyloop, int l_min, int l_max) const {
	LOG("Finding sectors and their graphs to match with diagrams:");
	Files* files = Files::instance();
	list<IntegralFamily*> ics = files->integralfamilies();
	list<IntegralFamily*>::const_iterator ic;
	size_t num_secs = 0;
	// option "construct_minimal_graphs" set to false
	// the user must rely on "minimize_graphs_by_twists", otherwise also the
	// diagrams from QGRAF must be reconstructed with this option
	bool construct_minimal_graphs = false;
	for (ic = ics.begin(); ic != ics.end(); ++ic) {
		int loop = (*ic)->loop_momenta().nops();
		if ((*ic)->is_crossed() || loop < l_min || loop > l_max)
			continue;
		string name = (*ic)->name();
		const SectorMappings* m = files->sectormappings(name);
		set<SectorGL> targets;
		LOG("Finding sectors of integral family " << name << ":");
		m->find_shift_targets_with_graph(targets, minimize_graphs_by_twists_,
				construct_minimal_graphs);
		num_secs += targets.size();
		map<int, set<SectorGL> >& sectorgraphs = secbytbyloop[loop];
		set<SectorGL>::iterator s; // \todo avoid copy
		for (s = targets.begin(); s != targets.end();) {
			sectorgraphs[s->sector_.t()].insert(*s);
			targets.erase(s++);
		}

	}
	return num_secs;
}

size_t FindDiagramShifts::get_graphs_of_diagrams(std::map<int, std::map<int,
		std::list<std::pair<Graph, Diagram> > > >& diasbytbyloop, int& l_min,
		int& l_max) const {
	LOG("Loading diagrams and preparing their graphs to match with sectors:");

	list<Diagram> alldias;
	read_diagrams(alldias, qgraf_filename_, names_);
	size_t num_dias = alldias.size();
	l_min = 0;
	l_max = -1;
	if (!alldias.empty())
		l_min = alldias.begin()->num_loops();

	while (!alldias.empty()) {
		Diagram& dia = *alldias.begin();
		Graph graph = dia.get_Graph();
		graph.contract_multiple_internal_propagators();
		graph.disconnect_vacuum_components();
		graph.contract_bridges();
		if (minimize_graphs_by_twists_)
			graph.minimize_by_twists();

		int loops = dia.num_loops();
		l_min = std::min(l_min, loops);
		l_max = std::max(l_max, loops);
		map<int, list<pair<Graph, Diagram> > >& diasbyt = diasbytbyloop[loops];
		list<pair<Graph, Diagram> >& graphlist =
				diasbyt[graph.num_internal_edges()];
		VERIFY(graph.get_propagators_with_loop_momenta().size()
				== graph.num_internal_edges()); // \todo remove this test
		graphlist.push_back(make_pair(graph, dia));
		alldias.erase(alldias.begin());
	}
	LOG("Loaded " << num_dias << " diagrams.");
	return num_dias;
}

void FindDiagramShifts::run_serial() {
	using GiNaC::exmap;
	using GiNaC::lst;
	using GiNaC::subs_options;

	Files* files = Files::instance();

	string diag_tmp_filename = output_filename_ + ".tmp";
	ofstream os_diag(diag_tmp_filename.c_str());
	if (!os_diag)
		ABORT("Can't open file " << diag_tmp_filename);

	bool write_info_file = !info_filename_form_.empty();
	string info_tmp_filename = info_filename_form_ + ".tmp";
	ofstream os_info;
	if (write_info_file)
		os_info.open(info_tmp_filename.c_str());

	// read in the diagrams

	int l_min, l_max;
	map<int, map<int, list<pair<Graph, Diagram> > > > diasbytbyloop;
	size_t num_dias = get_graphs_of_diagrams(diasbytbyloop, l_min, l_max);

	// read in the graphs to which the diagrams can be matched

	map<int, map<int, set<SectorGL> > > secbytbyloop;
	get_graphs_of_sectors(secbytbyloop, l_min, l_max);

	// match diagrams, for-loop over loop and t

	size_t num_matched = 0, num_unmatched = 0;
	set<Sector> secs; // direct target sectors
	set<Sector> minsecs; // minimal target sectors (use crossings, no trivial zeros)
	int col = 0;

	LOGN("\nMatching diagrams:");
	list<Diagram> unmatched;
	map<int, map<int, list<pair<Graph, Diagram> > > >::iterator it1;
	for (it1 = diasbytbyloop.begin(); it1 != diasbytbyloop.end(); ++it1) {
		int loop = it1->first;
		map<int, list<pair<Graph, Diagram> > >& diasbyt = it1->second;
		map<int, set<SectorGL> >& secbyt = secbytbyloop[loop];

		map<int, list<pair<Graph, Diagram> > >::iterator it2;
		for (it2 = diasbyt.begin(); it2 != diasbyt.end(); ++it2) {
			int t = it2->first;
			list<pair<Graph, Diagram> >& dias = it2->second;
			set<SectorGL>& sectors = secbyt[t];

			list<pair<Graph, Diagram> >::iterator g;
			for (g = dias.begin(); g != dias.end(); ++g) {
				Diagram& dia = g->second;
				const Graph& graph = g->first;
				LOGX("\nAnalyzing diagram " << dia.name());
				if ((col++) % 60 == 0)
					LOG("");

				// case where Diagram is trivially zero not handled
				//bool is_trivially_zero = independent_propagators(
				//		graph.get_propagators_with_loop_momenta(),
				//		graph.loop_momenta()).first.is_zero();
				//if (is_trivially_zero) {
				//	WARNING("Diagram " << dia.name() << " is trivially zero");
				//	++num_unmatched;
				//	continue;
				//}

				LabelPair label = graph.find_canonical_graph_label();
				set<SectorGL>::const_iterator s;
				for (s = sectors.begin(); s != sectors.end(); ++s) {
					if (label.first != s->labels_.first)
						continue; // not isomorphic
					exmap shift;
					const Kinematics* kin =
							s->sector_.integralfamily()->kinematics();
					Crossing crossing(kin);
					list<map<int, int> > dummysym;
					GiNaC::ex det = Graph::find_shift(graph, label, s->graph_,
							s->labels_, shift, crossing, dummysym);
					if (!det.is_equal(1))
						continue; // det not 1

					// graph has same propagators as sector once shift and crossing are applied to graph
					// if only the shift is applied to the graph then the graph matches to the sector
					// where the inverse crossing has been applied

					const IntegralFamily* f = s->sector_.integralfamily();
					string name_inv_cf =
							crossing.inverse().name_for_crossed_family(f);
					const IntegralFamily* inv_cf = files->integralfamily(
							name_inv_cf);
					Sector matched(inv_cf, s->sector_.id());

					if (!SectorMappings::is_zero(s->sector_)) {
						LOGN("x");
						minsecs.insert(s->sector_); // uncrossed
					} else {
						LOGN("o");
					}

					LOGX("  => matched diagram " << dia.name() << " to sector " << matched);
					LOGX("     via shift: " << shift);

					// write to info file
					os_info << "id DiaMatch(" << dia.name() << ") = " //
							<< "Sector(" << name_inv_cf//
							<< ", " << matched.t()//
							<< ", " << matched.id() << ")"//
							<< " * Shift(";
					for (exmap::const_iterator i = shift.begin(); i
							!= shift.end();) {
						os_info << i->first << ", " << i->second << ", []";
						if (++i != shift.end())
							os_info << ", ";
					}
					os_info << ");\n";

					dia.substitute_momenta(shift, subs_options::no_pattern);
					dia.set_loop_momenta(s->graph_.loop_momenta());
					dia.set_sector(matched);
					++num_matched;
					secs.insert(matched);
					break;
				} // sectors
				if (s == sectors.end()) {
					++num_unmatched;
					unmatched.push_back(dia);
					LOGN("n");
					LOGX("  no match for diagram " << dia.name());
				}
			} // graphs
		} // t
	} // loop
	LOG("");

	if (write_info_file) {
		os_info.close();
		rename(info_tmp_filename, info_filename_form_);
	}

	VERIFY(num_matched + num_unmatched == num_dias);

	// write out the new diagram file
	LOG("\n\nWriting new diagram file: " << output_filename_);
	for (it1 = diasbytbyloop.begin(); it1 != diasbytbyloop.end(); ++it1) {
		map<int, list<pair<Graph, Diagram> > >::iterator it2;
		for (it2 = it1->second.begin(); it2 != it1->second.end(); ++it2) {
			list<pair<Graph, Diagram> >::iterator g;
			for (g = it2->second.begin(); g != it2->second.end(); ++g) {
				YAML::Emitter ye;
				ye << YAML::BeginMap;
				ye << YAML::Key << "diagram" << YAML::Value << g->second;
				ye << YAML::EndMap;
				os_diag << ye.c_str() << endl;
			}
			//for (g = it2->second.begin(); g != it2->second.end(); ++g)
			//	g->second.write_stream(os_diag);
		}
	}
	os_diag.close();
	rename(diag_tmp_filename, output_filename_);

	LOG("\nUnmatched diagrams: " << num_unmatched);
	for (list<Diagram>::const_iterator d = unmatched.begin(); d
			!= unmatched.end(); ++d)
		LOGX("  " << d->name());

	LOG("\nMatched diagrams: " << num_matched);

	map<int, list<string> >::const_iterator it;
	list<string>::const_iterator n;
	LOGN("\nSectors: " << secs.size());
	//int col = 0;
	int last_t = -1;
	const IntegralFamily* last_ic = 0;
	for (set<Sector>::const_iterator s = secs.begin(); s != secs.end(); ++s) {
		if (last_ic == 0 || *s->integralfamily() != *last_ic) {
			LOGN("\nfamily " << s->integralfamily()->name() << ":");
			last_ic = s->integralfamily();
			last_t = -1;
		}
		if (s->t() != last_t)
			LOGN("\nt = " << s->t() << ": ");
		last_t = s->t();
		LOGN(s->id() << " ");
	}
	LOG("");

	LOG("Determining compact recursive sector selection");
	SectorSelection sectree =
			SectorSelection::find_compact_recursive_selection(minsecs, true);
	YAML::Emitter ye;
	ye << YAML::BeginMap << YAML::Key << "sector_selection" << YAML::Value
			<< sectree << YAML::EndMap;
	LOG("\nThis selection of sectors should be reduced for the matched diagrams:");
	LOG(ye.c_str());
}

bool FindDiagramShifts::find_dependencies(const set<string>& outothers, list<
		string>& in, list<string>& out, list<Job*>& auxjobs) {
	//find_dependencies_all_sectormappings(outothers, in, auxjobs);
	in.push_back(qgraf_filename_);
	out.push_back(output_filename_);
	if (!info_filename_form_.empty())
		out.push_back(info_filename_form_);
	if (Files::instance()->integralfamilies().empty())
		ABORT("Can't match any diagrams since no integral families defined");
	return true;
}

std::string FindDiagramShifts::get_description() const {
	return "find diagram shifts for " + short_filename(qgraf_filename_);
}

}
