/*  job_printsectorinfo.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_printsectorinfo.h"
#include "functions.h"
#include "files.h"
#include "sector.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "int.h"
#include "graph.h"
#include "ginacutils.h"

using namespace std;

namespace Reduze {

// register job type at JobFactory
namespace {
JobProxy<PrintSectorInfo> dummy;
}

void PrintSectorInfo::run_serial() {
	const set<Sector> sectors = sector_selection_.find_sectors();
	LOG("\nAnalyzing " << sectors.size() << " sector(s)\n");
	set<Sector>::const_iterator s;
	size_t w_sec(17), w_sub_dir(7), w_sub_all(6), w_prop(40);
	for (s = sectors.begin(); s != sectors.end(); ++s) {
		w_sec = max(w_sec, to_string(*s).size() + 2);
		w_prop = max(w_prop,
				to_string(INT(s->integralfamily(), s->id())).size() + 1);
		w_sub_dir = max(w_sub_dir, to_string(s->t()).size() + 1);
		w_sub_all = max(w_sub_all, to_string(s->id()).size() + 2);
	}

	// print table
	if (true) {
		LOG(setw(w_sec) << left << "sector"
				<< setw(w_sub_dir + w_sub_all) << "subsectors"
				<< setw(w_prop) << "corner integral");
		//LOG(setw(w_sec + w_sub_dir + w_sub_all + w_prop)
		//		<< setfill('.') << "" << setfill(' '));
		LOG(setw(w_sec) << left << "(fam:t:id)"
				<< setw(w_sub_dir) << "direct"
				<< setw(w_sub_all) << "all"
				<< setw(w_prop) << "fam t id  r s  propagator_exponents");
		LOG(setw(w_sec + w_sub_dir + w_sub_all + w_prop)
				<< setfill('.') << "" << setfill(' '));
		for (s = sectors.begin(); s != sectors.end(); ++s)
			LOG(setw(w_sec) << to_string(*s)
					<< setw(w_sub_dir) << SectorSelection::find_prerequisite_sectors(*s, false).size()
					<< setw(w_sub_all) << SectorSelection::find_prerequisite_sectors(*s, true).size()
					<< setw(w_prop) << to_string(INT(*s)));
		LOG(setw(w_sec + w_sub_dir + w_sub_all + w_prop)
				<< setfill('.') << "" << setfill(' '));
	}

	// subsectors
	if (true) {
		for (s = sectors.begin(); s != sectors.end(); ++s) {
			LOG("\nSector " << *s);
			stringstream ssd, ssa;
			set<Sector> a =
					SectorSelection::find_prerequisite_sectors(*s, true);
			copy(a.begin(), a.end(), ostream_iterator<Sector> (ssa, " "));
			LOG("  all prerequisite sectors (subsectors): " << ssa.str());
		}
	}

	if (!feynman_diagram_dot_file_.empty()) {
		string dotfilename = feynman_diagram_dot_file_;
		string dotfiletmp = dotfilename + ".tmp";
		string psfilename = dotfilename + ".ps";
		LOG("\nFind Feynman diagrams (unphysical and known zero sectors excluded)");
		ProgressBar pbar(0, "find diagrams", sectors.size());
		pbar.start();
		ofstream of(dotfiletmp.c_str());
		int num_dias(0);
		set<Sector> unphysical, zeros;
		// bottom up!
		for (s = sectors.begin(); s != sectors.end(); ++s) {
			pbar.print();
			if (true) {
				set<Sector> subsecs =
						SectorSelection::find_prerequisite_sectors(*s, false);
				set<Sector>::const_iterator sub, found = unphysical.end();
				for (sub = subsecs.begin(); sub != subsecs.end(); ++sub) {
					if ((found = unphysical.find(*sub)) != unphysical.end())
						break;
				}
				if (found != unphysical.end()) {
					LOGX("No graph for sector "
							<< s->get_safe_string_for_filename()
							<< "\n  has unphysical subsector "
							<< found->get_safe_string_for_filename());
					unphysical.insert(*s);
					continue;
				}
			}
			Graph g;
			bool u_poly_non_zero;
			if (!s->search_graph(g, u_poly_non_zero, construct_minimal_graphs_)) {
				LOGX("No graph for sector "
						<< s->get_safe_string_for_filename());
				if (u_poly_non_zero) {
					unphysical.insert(*s);
					LOGX("  is not trivially zero.");
				} else {
					LOGX("  is trivially zero.");
					zeros.insert(*s);
				}
				continue;
			}
			++num_dias;
			if (minimize_graphs_by_twists_)
				g.minimize_by_twists();
			g.print_dot(of);
			of << "\n";
			LOGX("Found graph for sector " << s->get_safe_string_for_filename());
		}
		pbar.end();
		of.close();
		rename(dotfiletmp, dotfilename);
		LOG("Number of printed Feynamn diagrams: " << num_dias);
		LOG("Number of omitted Feynamn diagrams: (U-poly not zero, U-poly zero) (" << unphysical.size() << ", " << zeros.size() << ")");
		LOG("To produce a postscript file run the following command:");
		LOG("neato -Tps " << dotfilename << " -o " << psfilename);
	}

	//	string dotfilename = Files::instance()->get_project_directory()
	//			+ "sectors.dot";
	//	string psfilename = Files::instance()->get_project_directory()
	//			+ "sectors.ps";
	// \todo possible to print only physical sectors
	if (!sector_tree_dot_file_.empty()) {
		string dotfilename = sector_tree_dot_file_;
		string psfilename = dotfilename + ".ps";
		LOG("\nGenerating relationship graph file: " << dotfilename);
		SectorMappings::print_sectortree_dot(sectors, dotfilename);
		LOG("To produce a postscript file run the following command:");
		LOG("dot -Tps " << dotfilename << " -o " << psfilename);
		LOG("");
	}
}

bool PrintSectorInfo::find_dependencies(const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {
	//find_dependencies_all_sectormappings(outothers, in, auxjobs);
	if (!sector_tree_dot_file_.empty())
		out.push_back(sector_tree_dot_file_);
	if (!feynman_diagram_dot_file_.empty())
		out.push_back(feynman_diagram_dot_file_);
	return true;
}

void PrintSectorInfo::print_manual_options(YAML::Emitter& os) const {
	os << YAML::Key << "sector_selection" << YAML::Value << sector_selection_;
}

void PrintSectorInfo::read_manual_options(const YAML::Node& node) {
	node["sector_selection"] >> sector_selection_;
}

void PrintSectorInfo::init() {
	if (!sector_tree_dot_file_.empty() && sector_tree_dot_file_
			== feynman_diagram_dot_file_)
		ABORT("sector_tree_dot_file and feynman_diagram_dot_file are identical: "
				<< sector_tree_dot_file_);
}

std::string PrintSectorInfo::get_description() const {
	return string("print sector info");
}

}
