/*  job_setupsectormappings.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_setupsectormappings.h"
#include "functions.h"
#include "sectormappings.h"
#include "files.h"
#include "graph.h"
#include "kinematics.h"

using namespace std;

namespace Reduze {

// register job type
namespace {
JobProxy<SetupSectorMappings> dummy;
}

SetupSectorMappings::SetupSectorMappings() :
	find_sector_relations_(true), find_sector_symmetries_(true),
			find_zero_sectors_(true), find_graphless_zero_sectors_(false),
			minimize_graphs_by_twists_(true), construct_minimal_graphs_(false),
			minimize_target_crossings_(true), allow_general_crossings_(true) {
	add_auto_options();
}

void SetupSectorMappings::make_graph_directories(const IntegralFamily* ic,
		std::string& all_dir, std::string& shift_targets_dir) const {
	string graph_dir = Files::instance()->get_graph_directory(ic);
	remove_directory_with_files(graph_dir);
	make_directory(graph_dir);
	all_dir = graph_dir + "all/";
	shift_targets_dir = graph_dir + "shift_targets/";
	make_directory(all_dir);
	make_directory(shift_targets_dir);
}

void SetupSectorMappings::find_crossed_relations(const IntegralFamily* ic,
		const std::set<SectorGL>& sectors) {
	// find crossed counter parts for sector mappings (one per equiv. class)
	list<SectorMappings*> ms;
	// it is enough to consider minimal crossing representatives as sources
	set<Crossing> xs = ic->kinematics()->equivalent_crossing_classes();
	for (set<Crossing>::const_iterator x = xs.begin(); x != xs.end(); ++x)
		if (!x->is_equivalent_to_identity()) {
			string xfname = x->name_for_crossed_family(ic);
			ms.push_back(Files::instance()->sectormappings(xfname));
		}
	// shift crossed versions of target sectors to uncrossed ones
	ProgressBar pbar(2, "crossed relations:", sectors.size());
	pbar.start();
	for (set<SectorGL>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		ASSERT(*ic == *(s->sector_.integralfamily()));
		for (list<SectorMappings*>::iterator m = ms.begin(); m != ms.end(); ++m) {
			// construct crossed sector and its graph
			Sector xsec((*m)->integralfamily(), s->sector_.id());
			map<Sector, Graph> gs;
			set<Sector> xsecset;
			xsecset.insert(xsec);
			SectorMappings::get_graphs(gs, xsecset, minimize_graphs_by_twists_,
					construct_minimal_graphs_, false);
			VERIFY(gs.find(xsec) != gs.end());
			Graph g = gs.find(xsec)->second;
			// find shift
			set<SectorGL> from, to;
			from.insert(SectorGL(xsec, g, g.find_canonical_graph_label()));
			to.insert(*s);
			map<SectorGL, list<SectorGL> > dummy;
			(*m)->find_sector_shifts(to, from, dummy, true, false, true, true,
					cached_node_perm_, false);
			map<int, Sector>::const_iterator rel = //
					(*m)->sector_relations().find(s->sector_.id());
			if (rel != (*m)->sector_relations().end()) {
				const Sector& target = rel->second;
				if (Crossing::to_minimal_crossing(target) == //
						Crossing::to_minimal_crossing(xsec)) {
					// delete shift, is covered by crossing equivalences
					(*m)->remove_shift(s->sector_.id());
				} else {
					LOGX(xsec << "=>" << s->sector_ << ": " << target);
				}
			}
		}
		pbar.print();
	}
	pbar.end();
}

void SetupSectorMappings::setup(SectorMappings* sm,
		std::set<SectorGL>& shift_targets) {

	const IntegralFamily* ic = sm->integralfamily();
	const Kinematics* kin = ic->kinematics();

	// the sectors with graphs of this family which are not shifted to any sector of 'shift_targets'
	std::set<SectorGL> local_shift_targets;

	LOG("Setup sector mappings for " << ic->name());
	if (!shift_targets.empty())
		ASSERT(*(shift_targets.rbegin()->sector_.integralfamily()) < *ic);

	// Log crossings

	LOGX("Ordered crossings:");
	const list<Crossing>& oc = ic->kinematics()->ordered_crossings();
	for (list<Crossing>::const_iterator c = oc.begin(); c != oc.end(); ++c) {
		const map<Crossing, Crossing>& xmap =
				ic->kinematics()->equivalent_crossings();
		VERIFY(xmap.find(*c) != xmap.end());
		YAML::Emitter ye;
		ye << *c;
		LOGX ("  " << ye.c_str() << "\n"
				<< "is equivalent to '" << xmap.find(*c)->second.name() << "'");
	}
	LOGX("");

	// Log some translations

	YAML::Emitter e;
	ic->print(e);
	LOGX(e.c_str());
	LOGX("");
	LOGX("Rules scalar products to invariants:\n" << ic->kinematics()->rules_sp_to_invariants());
	try {
		LOGX("Rules invariants to scalar products:\n" << ic->kinematics()->find_rules_invariants_to_sp());
	} catch (exception& e) {
		WARNING("failed to invert scalar product rules, some features might not be available");
	}
	LOGX("Rules scalar products to propagators:\n" << ic->rules_sp_to_prop());
	LOGX("");

	const int max_sec_id = ic->max_sector_id();
	LOG("Maximal sector id is " << max_sec_id << ".");
	LOG(max_sec_id + 1 << " sectors occur in total");

	// sector tree (using global permutation symmetry)

	LOG("Finding non-equivalent sectors using permutation symmetries:");
	bool have_perm_syms = !ic->permutation_symmetries().empty();
	set<Sector> sectors;
	set<int> sector_ids;
	for (int i = 0; i <= max_sec_id; ++i) {
		int id = have_perm_syms ? ic->get_equivalent(i) : i;
		sectors.insert(Sector(ic, id));
		sector_ids.insert(id);
	}
	// print non-equivalent sectors to log-file
	LOGX("\nNon-equivalent sectors:");
	LOGX(ic->t_tree_nice_string(sector_ids));
	LOG(sector_ids.size() << " non-equivalent sectors were found");

	// find graphs of sectors and add sectors with vanishing jacobian determinant
	// of the loop momenta matrix to zero_sectors_ and set the sector_without_graph_
	map<Sector, Graph> graph_by_sector;
	bool verbose = true;
	sm->find_graphs(graph_by_sector, sectors, minimize_graphs_by_twists_,
			construct_minimal_graphs_, verbose);

	// print graphs as dot files in standard directory ./graphs/family_name/
	string all_dir, shift_targets_dir;
	make_graph_directories(ic, all_dir, shift_targets_dir);
	LOG("Printing all graphs from sectors in " << all_dir << ":");
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s)
		s->second.print_dot(all_dir + s->first.get_safe_string_for_filename(
				true) + ".dot");

	LOGX("\nSectors without a graph:");
	LOGX(ic->t_tree_nice_string(sm->sectors_without_graph()));

	// canonical labeling

	LOG("Finding canonical labels:");
	ProgressBar pbar(2, "canonical labels:", graph_by_sector.size());
	pbar.start();
	set<SectorGL> local_sectors;
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s) {
		local_sectors.insert(SectorGL(s->first, s->second,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();

	// list of sectors by their shift target
	// all sectors are from this family, the target itself is not included in the list
	map<SectorGL, list<SectorGL> > local_targets_by_equiv;

	// try to shift the 'local_sectors' to the shift_targets or to them self
	// updates the 'shift_targets'
	// sets the 'local_shift_targets'
	if (find_sector_relations_) {
		sm->find_sector_shifts(shift_targets, local_sectors,
				local_targets_by_equiv, minimize_target_crossings_,
				find_zero_sectors_, true, allow_general_crossings_,
				cached_node_perm_, true);
	} else {
		// move local sectors to local_shift_targets
		map<SectorGL, list<SectorGL> >::iterator in =
				local_targets_by_equiv.end();
		list<SectorGL> empty;
		for (set<SectorGL>::iterator s = local_sectors.begin(); s
				!= local_sectors.end();) {
			in = local_targets_by_equiv.insert(in, make_pair(*s, empty));
			local_sectors.erase(s++);
		}
	}
	LOG("Found " << local_targets_by_equiv.size() << " shift target sectors.");
	// print the number of different graphs in the shift targets
	set<GraphLabel> different_graphs;
	for (map<SectorGL, list<SectorGL> >::const_iterator s =
			local_targets_by_equiv.begin(); s != local_targets_by_equiv.end(); ++s)
		different_graphs.insert(s->first.labels_.first);
	LOG("  shift targets contain " << different_graphs.size() << " different graph(s)");

	// zero test for the local_targets

	const bool allow_long_zero_test = (kin->external_momenta().nops() != 0);
	if (find_zero_sectors_) {
		LOGN("Finding zero sectors of sectors with a graph:");
		map<Sector, list<Sector> > zeros;
		map<Sector, list<Sector> >::iterator in = zeros.end();
		map<SectorGL, list<SectorGL> >::const_iterator s;
		list<SectorGL>::const_iterator t;
		s = local_targets_by_equiv.begin();
		for (; s != local_targets_by_equiv.end(); ++s) {
			in = zeros.insert(in, make_pair(s->first.sector_, list<Sector> ()));
			for (t = s->second.begin(); t != s->second.end(); ++t)
				in->second.push_back(t->sector_);
		}
		sm->find_zero_sectors(zeros, allow_long_zero_test);
		LOG("Found " << sm->zero_sectors().size() << " zero sectors (isomorphics included).");
	}
	if (find_zero_sectors_ && find_graphless_zero_sectors_) {
		LOGN("Finding zero sectors for sectors without a graph:");
		map<Sector, list<Sector> > zeros;
		map<Sector, list<Sector> >::iterator in = zeros.end();
		int num_phys_zeros = sm->zero_sectors().size();
		set<int>::const_iterator s = sm->sectors_without_graph_.begin();
		for (; s != sm->sectors_without_graph_.end(); ++s)
			in = zeros.insert(in, make_pair(Sector(ic, *s), list<Sector> ()));
		sm->find_zero_sectors(zeros, allow_long_zero_test);
		int num_zeros = sm->zero_sectors().size();
		LOG("Found additional " << num_zeros - num_phys_zeros << " zero sectors.");
	}

	set<int> non_zero_shift_targets;
	set<GraphLabel> non_zero_different_graphs;
	for (map<SectorGL, list<SectorGL> >::const_iterator s =
			local_targets_by_equiv.begin(); s != local_targets_by_equiv.end(); ++s) {
		if (sm->zero_sectors().find(s->first.sector_.id())
				== sm->zero_sectors().end()) {
			non_zero_shift_targets.insert(s->first.sector_.id());
			non_zero_different_graphs.insert(s->first.labels_.first);
		}
		local_shift_targets.insert(s->first);
	}
	LOGX("\nNon-vanishing shift targets:");
	LOGX(ic->t_tree_nice_string(non_zero_shift_targets));
	LOG("Found " << non_zero_shift_targets.size() << " non-vanishing shift target sectors");
	LOG("  non-vanishing shift targets contain " << non_zero_different_graphs.size() << " different graph(s)");

	// symmetry shifts

	if (find_sector_symmetries_) {
		LOG("Finding symmetry shift relations:");
		sm->find_symmetry_shifts(local_shift_targets, cached_node_perm_);
	}
	//LOG("");

	// pure crossing relations: crossed to uncrossed sector
	// setup of uncrossed mappings done, it is safe now to construct
	// crossed mappings based on them (implicitely via Files::sectormappings())
	// and enrich them
	if (allow_general_crossings_) {
		LOG("Finding crossed sector relations for non-zero shift targets");
		find_crossed_relations(ic, local_shift_targets);
	}

	LOG("Printing graphs of shift targets in " << shift_targets_dir << ":");
	for (set<SectorGL>::const_iterator s = local_shift_targets.begin(); s
			!= local_shift_targets.end(); ++s)
		s->graph_.print_dot(shift_targets_dir
				+ s->sector_.get_safe_string_for_filename(true) + ".dot");

	string dot_file_prefix = all_dir + safe_filename(ic->name());
	string dot_file_prefix2 = shift_targets_dir + safe_filename(ic->name());
	string ps_file = safe_filename(ic->name()) + ".ps";
	string ps_file2 = safe_filename(ic->name()) + "_shift_targets.ps";
	LOG("To generate a postscript file from the graphs in the dot format type e.g.");
	LOG("  dot -Tps -o " << ps_file << " " << dot_file_prefix << "_*.dot");
	LOG("  dot -Tps -o " << ps_file2 << " " << dot_file_prefix2 << "_*.dot");

	LOG("Done setup sector mappings for " << ic->name());
}

void SetupSectorMappings::run_serial() {

	// sort families by loop

	Files* files = Files::instance();
	map<int, list<IntegralFamily*> > families_by_loop;
	list<IntegralFamily*> ifs = files->integralfamilies();
	for (list<IntegralFamily*>::const_iterator f = ifs.begin(); f != ifs.end(); ++f)
		if (!(*f)->is_crossed())
			families_by_loop[(*f)->loop_momenta().nops()].push_back(*f);

	// print what will be done

	map<int, list<IntegralFamily*> >::const_reverse_iterator fbl; // heavy work first
	LOG("Setup sector mappings for the following integral families requested:");
	for (fbl = families_by_loop.rbegin(); fbl != families_by_loop.rend(); ++fbl) {
		list<IntegralFamily*>::const_iterator fam;
		LOG("  " << fbl->first << "-loop:");
		for (fam = fbl->second.begin(); fam != fbl->second.end(); ++fam) {
			LOG("    " << (*fam)->name());
		}
	}
	LOG("");

	// setup the sector mappings for each class of families with the same number of loop momenta

	for (fbl = families_by_loop.rbegin(); fbl != families_by_loop.rend(); ++fbl) {
		// cached shift target sectors
		set<SectorGL> shift_targets;
		// cached node permutations by sector
		map<Sector, list<map<int, int> > > cached_node_perm;
		bool lower_family_changed = false;
		list<IntegralFamily*>::const_iterator fam;
		for (fam = fbl->second.begin(); fam != fbl->second.end(); ++fam) {
			string sm = files->get_filename_sectormappings((*fam)->name());
			bool sm_already_setup = is_readable_file(sm)
					&& !files->is_empty_sectormappings_file((*fam)->name());
			if (sm_already_setup && is_conditional()) {
				// case of where a family is already setup: we must find the shift target sectors of this family
				SectorMappings* mp = files->sectormappings((*fam)->name());
				LOG("Integral family " << (*fam)->name() << " already setup.");
				// case where lower family has been changed not handled
				if (lower_family_changed)
					WARNING("Lower integral family has been changed."
							<< " This might invalidate the already existent setup"
							<< " of integral family " << (*fam)->name() << "\n");
				if (find_sector_relations_)
					mp->find_shift_targets_with_graph(shift_targets,
							minimize_graphs_by_twists_,
							construct_minimal_graphs_);
			} else {
				// clear and create empty sector mappings and set it up
				files->clear_sectormappings((*fam)->name());
				files->save(SectorMappings(*fam));
				SectorMappings* m = files->sectormappings((*fam)->name());
				setup(m, shift_targets);
				LOG("Saving sector mappings for " << (*fam)->name());
				files->save(*m);
				lower_family_changed = true;
				LOG("");
			}
		} // t
	} // loop
}

bool SetupSectorMappings::find_dependencies(const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {
	Files* files = Files::instance();
	list<IntegralFamily*> ics = files->integralfamilies();
	list<IntegralFamily*>::const_iterator ic;
	for (ic = ics.begin(); ic != ics.end(); ++ic)
		if (!(*ic)->is_crossed())
			out.push_back(files->get_filename_sectormappings((*ic)->name()));
	return true;
}

void SetupSectorMappings::init() {
}

std::string SetupSectorMappings::get_description() const {
	return string("setup sector mappings");
}

}
