/*  job_setupsectormappingsalt.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_setupsectormappingsalt.h"

#include "kinematics.h"
#include "sectormappings.h"
#include "functions.h"
#include "files.h"
#include "ginacutils.h"
#include "graph.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// register job type
namespace {
JobProxy<SetupSectorMappingsAlt> dummy;
}

bool SetupSectorMappingsAlt::find_dependencies(const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {

	Files* files = Files::instance();
	list<IntegralFamily*> ics = files->integralfamilies();
	list<IntegralFamily*>::const_iterator ic;
	for (ic = ics.begin(); ic != ics.end(); ++ic)
		if (!(*ic)->is_crossed())
			out.push_back(files->get_filename_sectormappings((*ic)->name()));
	return true;
}

void SetupSectorMappingsAlt::init() {
}

std::string SetupSectorMappingsAlt::get_description() const {
	return string("setup sector mappings (alt)");
}

class LessExmap {
public:
	bool operator()(const GiNaC::exmap& a, const GiNaC::exmap& b) const {
		if (a.size() != b.size())
			return a.size() < b.size();
		exmap::const_iterator i, j;
		size_t arhs = 0;
		size_t brhs = 0;
		for (i = a.begin(); i != a.end(); ++i)
			arhs += i->second.nops();
		for (j = a.begin(); j != a.end(); ++j)
			brhs += j->second.nops();
		if (arhs != brhs)
			return arhs < brhs;
		for (i = a.begin(), j = b.begin(); i != a.end() && j != b.end(); ++i, ++j)
			if (!i->first.is_equal(j->first))
				return i->first.compare(j->first) < 0;
			else if (!i->second.is_equal(j->second))
				return i->second.compare(j->second) < 0;
		return false;
	}
};

// propagators are given in form of pair<momentum, squaredmass>
// shift of loop and crossing of ext. momenta is simultaneous trafo
bool find_shifts(//
		const vector<pair<ex, ex> >& srcprops,//
		const lst& srcmoms, // src loop momenta (alt symbols)
		const vector<pair<ex, ex> >& destprops,//
		const lst& destmoms, // dest loop momenta
		list<lst>& shifts, bool stop_on_first_match) {
	stack<int> matched; // assigned dest propagators
	stack<lst> sysmatched; // eqs for assigned dest props
	stack<bool> signmatched; // minus signs for momentum assignments
	list<int> unmatched; // ordered list of unassigned dest propagators

	size_t t = srcprops.size();
	ASSERT(srcprops.size() == t && destprops.size() == t);
	for (size_t i = 0; i < t; ++i)
		unmatched.push_back(i);

	list<int>::iterator totry = unmatched.begin(); // next dest propagator to try
	bool sign = false; // minus sign for dest momentum ?
	while (totry != unmatched.end() || !matched.empty()) {
		if (totry != unmatched.end()) {
			// try to match one specific propagator
			const pair<ex, ex>& sp = srcprops[matched.size()];
			const pair<ex, ex>& dp = destprops[*totry];
			lst sys;
			if (sp.second.is_equal(dp.second)) { // mass^2 match
				// now try the momenta
				sys = (sysmatched.empty() ? lst() : sysmatched.top());
				sys.append(sp.first == (sign ? -dp.first : dp.first));
				if (matched.size() + 1 >= srcmoms.nops() || //
						unmatched.size() == 1 /* last prop to match ? */) {
					//	LOG("Solving");
					sys = ex_to<lst> (lsolve(sys, srcmoms));
				}
			}
			if (sys.nops() > 0) { // match succeeded ? then go one level deeper
				matched.push(*totry);
				unmatched.erase(totry);
				totry = unmatched.begin();
				sysmatched.push(sys);
				signmatched.push(sign);
				sign = false;
			} else { // match failed, try next
				if (!(sign = !sign))
					++totry;
			}
			if (unmatched.empty()) { // full match of all propagators ?
				ex det = abs(jacobi_determinant(sys, destmoms).expand());
				if (det == 1) {
					shifts.push_back(sys);
					if (stop_on_first_match)
						return true;
				}
			}
		} else if (!matched.empty()) {
			// next try at one level up, discard last matched
			int last = matched.top();
			totry = lower_bound(unmatched.begin(), unmatched.end(), last);
			totry = unmatched.insert(totry, last);
			if (!(sign = !signmatched.top()))
				++totry;
			matched.pop();
			sysmatched.pop();
			signmatched.pop();
		} else {
			// all tries done
			return false;
		}
	}
	return false; // never reached, suppresses compiler warning
}

// will put simplest shift to front
bool find_shifts(const Sector& src,//
		set<Sector>::const_iterator destbegin,//
		set<Sector>::const_iterator destend,//
		const list<Crossing>& crossings, /*crossings equivalent to identity*/
		list<pair<Sector, exmap> >& shifts, bool stop_on_first_match) {
	size_t t = src.t();
	vector<Propagator> srcprops = src.get_propagators();
	ASSERT(!crossings.empty()); // at least the identity
	ASSERT(srcprops.size() == t);
	// temp names for src loop momenta in case they coincide with those from ic
	const Kinematics* kin = Files::instance()->kinematics();
	const lst& extmoms = kin->independent_external_momenta();
	lst srcloopmoms = src.integralfamily()->loop_momenta();
	lst tmploopmoms, tmpextmoms; // temp symbols for srcloopmoms, extmoms
	exmap src2tmp, tmp2src;
	provide_alternative_symbols(srcloopmoms, tmploopmoms, src2tmp, tmp2src);
	provide_alternative_symbols(extmoms, tmpextmoms, src2tmp, tmp2src);

	for (set<Sector>::const_iterator dest = destbegin; dest != destend; ++dest) {
		//LOGX("checking " << src << " -> " << *dest);
		lst destloopmoms = dest->integralfamily()->loop_momenta();
		vector<Propagator> destprops = dest->get_propagators();
		ASSERT(srcloopmoms.nops() == destloopmoms.nops())
		ASSERT(destprops.size() == t);

		// quick return if masses can't be matched (not necessary but useful)
		vector<ex> smass2(t), dmass2(t);
		for (size_t i = 0; i < t; ++i) {
			smass2[i] = srcprops[i].squaredmass();
			dmass2[i] = destprops[i].squaredmass();
		}
		sort(smass2.begin(), smass2.end(), ex_is_less());
		sort(dmass2.begin(), dmass2.end(), ex_is_less());
		if (!equal(smass2.begin(), smass2.end(), dmass2.begin()))
			continue;

		list<Crossing>::const_iterator c;
		for (c = crossings.begin(); c != crossings.end(); ++c) {
			//ASSERT(c->is_equivalent(Crossing(extmoms, invars))); // equiv to id
			ASSERT(c->is_equivalent_to_identity());
			lst eqns;
			exmap subst;
			for (lst::const_iterator p = extmoms.begin(); p != extmoms.end(); ++p) {
				ex lhs = p->subs(src2tmp);
				ex rhs = p->subs(c->rules_momenta());
				eqns.append(lhs == rhs);
				subst[lhs] = rhs;
			}
			vector<pair<ex, ex> > sps(t), dps(t);
			try {
				for (size_t i = 0; i < t; ++i) {
					ex smom = srcprops[i].momentum();
					ex dmom = destprops[i].momentum();
					smom = smom.subs(src2tmp).subs(subst);
					sps[i] = make_pair(smom, srcprops[i].squaredmass());
					dps[i] = make_pair(dmom, destprops[i].squaredmass());
				}
			} catch (exception& e) {
				//WARNING("can't match non-standard propagators in " << src << " or " << *dest);
				break;
			}
			list<lst> matches;
			find_shifts(sps, tmploopmoms, dps, destloopmoms, matches,
					stop_on_first_match);
			for (list<lst>::const_iterator m = matches.begin(); m
					!= matches.end(); ++m) {
				// note: shift might transform external momenta
				// undo with: Crossing invcross(kin, c->permutation().inverse());
				exmap mmap = equations_to_substitutions(add_lst(*m, eqns),
						tmp2src, true);
				// insert to front if better than previous best
				if (!shifts.empty() && !(*dest > shifts.front().first)//
						&& (*dest < shifts.front().first || LessExmap()(mmap,
								shifts.front().second)))
					shifts.push_front(make_pair(*dest, mmap));
				else
					shifts.push_back(make_pair(*dest, mmap));
			}
			if (stop_on_first_match && !matches.empty())
				return true;
		}
	}
	return false;
}

/*
 // NOTE: take care of this before using: may find non-minimal target sectors
 // will put simplest shift to front
 bool find_shifts_via_subsectors(const Sector& src,
 list<pair<Sector, exmap> >& shifts) {
 const IntegralFamily* ic = src.integralfamily();
 string name = ic->name();
 const SectorMappings* m = Files::instance()->sectormappings(name);
 set<int> sub = ic->get_immediate_subsector_equivalents(src.get_id());
 bool found = false;
 for (set<int>::const_iterator s = sub.begin(); s != sub.end(); ++s) {
 try {
 Sector sub(ic, *s);
 Sector subdest = SectorMappings::get_shifted_sector(sub);
 if (subdest == sub) // no shift
 continue;
 const IntegralFamily* destic = subdest.integralfamily();
 INT integral(ic, *s);
 exmap shift = m->find_shift_for_sector_relation(integral);
 ex integrand = INT(src).get_integrand().subs(shift);
 INT i(destic);
 ex c;
 //LOGX("testing shift " << sub << " -> " << subdest);
 if (destic->match_integrand(integrand, i, c) //
 && i.get_sector() < src) {
 // note: might be a sector which is reducible by a shift
 if (!shifts.empty() && shifts.front().first < i.get_sector())
 shifts.push_back(make_pair(i.get_sector(), shift));
 else
 shifts.push_front(make_pair(i.get_sector(), shift));
 found = true;
 }
 } catch (exception&) {
 }
 }
 return found;
 }
 */

//void SetupSectorMappingsAlt::find_target_sectors(std::set<Sector>& secs) const {
//	set<Sector> sectors;
//	Files* files = Files::instance();
//	list<IntegralFamily*> fams = files->integralfamilies();
//	for (list<IntegralFamily*>::iterator f = fams.begin(); f != fams.end(); ++f)
//		files->sectormappings((*f)->name())->find_shift_targets(sectors);
//}

void SetupSectorMappingsAlt::run_serial() {
	// validate and post process the input
	source_set_ = source_sectors_.find_sectors();
	set<Sector>& secs = source_set_;
	for (set<Sector>::iterator s = secs.begin(); s != secs.end();) {
		if (s->integralfamily()->is_crossed()) {
			LOG("ignoring sector " << *s << " from crossed integral family");
			secs.erase(s++);
		} else {
			source_families_.insert(s->integralfamily()->name());
			++s;
		}
	}

	set<Sector>::const_iterator s;
	if (find_zero_sectors_) {
		LOG("\nFinding zero sectors:");
		LOG("  analyzing " << source_set_.size() << " sectors from "
				<< source_families_.size() << " families");
		map<pair<int, string> , map<Sector, list<Sector> > > sec_by_fam;
		for (set<Sector>::iterator s = source_set_.begin(); s
				!= source_set_.end(); ++s) {
			const IntegralFamily* f = s->integralfamily();
			sec_by_fam[make_pair(f->id(), f->name())][*s] = //
					list<Sector> ();
		}
		// Include missing zero sectors in the sector mappings files which have
		// been omitted by SectorSelection::find_sectors().
		// This might also include zero sectors which are not sectors of the
		// user-defined selection, but who cares?
		map<pair<int, string> , map<Sector, list<Sector> > >::iterator sf;
		for (sf = sec_by_fam.begin(); sf != sec_by_fam.end(); ++sf) {
			const string& n = sf->first.second; // name
			LOG("\n  integral family '" << n << "':");
			const IntegralFamily* f = Files::instance()->integralfamily(n);
			SectorMappings* m = Files::instance()->sectormappings(n);
			LOG("   add trivial zero sectors with t < " << f->loop_momenta().nops());
			LOGN("   finding zero sectors with t >= " << f->loop_momenta().nops() << ":");
			set<int> trivial_zeros = f->get_trivial_zero_sector_equivalents();
			m->insert_zero_sectors(trivial_zeros);
			bool allow_longtest = f->kinematics()->external_momenta().nops()
					!= 0;
			m->find_zero_sectors(sf->second, allow_longtest);
			Files::instance()->save(*m);
		}
	}
	if (!find_sector_relations_ && !find_sector_symmetries_)
		return;

	map<int, set<Sector> > src_by_t;
	LOG("\nFinding source sectors");
	for (set<Sector>::iterator s = source_set_.begin(); s != source_set_.end();)
		if (!SectorMappings::is_zero(*s)) {
			src_by_t[s->t()].insert(*s);
			s++;
		} else {
			source_set_.erase(s++);
		}
	LOG("  found " << source_set_.size() << " source sectors");
	set<Sector> shift_targets;

	list<Crossing> cross, idcross, idequivcross;
	if (find_sector_relations_ || find_sector_symmetries_) {
		const Kinematics* kin = Files::instance()->kinematics();
		Crossing identity(kin);
		cross = kin->ordered_crossings();
		cross.push_front(identity);
		for (list<Crossing>::const_iterator c = cross.begin(); c != cross.end(); ++c)
			if (c->is_equivalent_to_identity())
				idequivcross.push_back(*c);
		idcross.push_front(identity);
	}
	map<int, set<Sector> > dest_by_t;
	if (find_sector_relations_) {
		LOG("\nFinding target sectors");
		set<Sector> dest_set;
		Files* files = Files::instance();
		list<IntegralFamily*> fams = files->integralfamilies();
		for (list<IntegralFamily*>::iterator f = fams.begin(); f != fams.end(); ++f) {
			if ((*f)->is_crossed())
				continue;
			set<int> dest_int_set;
			files->sectormappings((*f)->name())->find_shift_targets(
					dest_int_set);
			for (set<int>::const_iterator s = dest_int_set.begin(); s
					!= dest_int_set.end(); ++s)
				dest_set.insert(Sector(*f, *s));
		}
		LOG("  found " << dest_set.size() << " target sectors (without crossings)");
		for (set<Sector>::iterator s = dest_set.begin(); s != dest_set.end(); ++s)
			for (list<Crossing>::iterator c = cross.begin(); c != cross.end(); ++c)
				dest_by_t[s->t()].insert(c->transform(*s));
		LOG("  using " << idcross.size() << " crossings (equivalent to identity)");
	}

	LOG("\nFinding shifts with det(Jacobian) = 1:");
	map<int, set<Sector> >::iterator sbt;
	for (sbt = src_by_t.begin(); sbt != src_by_t.end(); ++sbt) {
		int t = sbt->first;
		set<Sector>& src_secs = sbt->second;
		const set<Sector>& dest_secs = dest_by_t[t];
		LOG("\nt = " << t);
		set<Sector>::iterator src, dest;
		for (src = src_secs.begin(); src != src_secs.end();) {
			LOGN("  " << *src << " ");
			list<pair<Sector, exmap> > shifts;
			if (find_sector_relations_) {
				find_shifts(*src, // from sec
						dest_secs.begin(), dest_secs.lower_bound(*src), // to secs
						idcross, shifts, true);
			}
			if (!shifts.empty()) { // found a shift to lower sector for src
				const Sector& dest = shifts.front().first;
				const exmap& shift = shifts.front().second;
				LOGN("shifted to " << dest);
				LOGNX("  using: " << shift << "  ");
				LOG("");
				SectorMappings::insert_shift(*src, dest, shift);
				shift_targets.insert(dest);
				src_secs.erase(src++); // don't use it as target sector
			} else { // no shift to lower sector for src
				shift_targets.insert(*src);
				if (find_sector_relations_)
					LOGN("no shift to lower sector");
				string name = src->integralfamily()->name();
				SectorMappings* m = Files::instance()->sectormappings(name);
				const map<int, list<exmap> >& s = m->sector_symmetries();
				if (find_sector_symmetries_ && s.find(src->id()) == s.end()) {
					set<Sector>::iterator destend = src;
					++destend;
					find_shifts(*src, src, destend, idequivcross, shifts, false);
					list<pair<Sector, exmap> >::const_iterator symm;
					list<exmap> rawshifts;
					for (symm = shifts.begin(); symm != shifts.end(); ++symm) {
						ASSERT(symm->first == *src);
						const exmap& shift = symm->second;
						lst moms = src->integralfamily()->get_all_momenta();
						if (!moms.subs(shift).is_equal(moms))
							rawshifts.push_back(shift);
					}
					LOGN(", " << rawshifts.size() << " shifts onto itself");
					if (!rawshifts.empty())
						SectorMappings::insert_symmetry_shifts(*src, rawshifts);
				}
				LOG("");
				++src;
			}
		} // source sector loop
		set<string>::const_iterator name = source_families_.begin();
		for (; name != source_families_.end(); ++name) {
			LOG("Saving updated sector mappings for " << *name);
			const SectorMappings * m = Files::instance()->sectormappings(*name);
			Files::instance()->save(*m);
		}
		LOG("");
	} // t loop

	LOGX("Found " << shift_targets.size() << " shift target sectors:");
	for (set<Sector>::const_iterator it = shift_targets.begin(); it
			!= shift_targets.end(); ++it)
		LOGX(*it);
}

}

