/*  kinematics.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "kinematics.h"
#include "propagator.h" // ScalarProduct
#include "functions.h"
#include "ginacutils.h"
#include "yamlutils.h"

#include "files.h"

using namespace GiNaC;
using namespace std;

namespace Reduze {

// register types
namespace {
YAMLProxy<Kinematics> dummy1;
YAMLProxy<Paths> dummy2;
}

Kinematics::Kinematics() :
	symbol_to_replace_by_one_("undefined"), generate_crossings_(true) {
	dimension_ = symbol("d");
	loop_momentum_prefix_ = "k";
	// rule_dimension_to_epsilon_[dimension_] = 4 - 2 * epsilon_dimension_;
}

Kinematics::~Kinematics() {
}

GiNaC::lst Kinematics::get_external_momenta_and_kinematics() const {
	return add_lst(external_momenta(), kinematic_invariants());
}

const std::string& Kinematics::loop_momentum_prefix() const {
	return loop_momentum_prefix_;
}

const Crossing& Kinematics::crossing(const std::string& name) const {
	map<string, Crossing>::const_iterator c;
	c = crossings_by_name_.find(name);
	if (c == crossings_by_name_.end())
		ABORT("Unknown Crossing: '" << name << "'");
	return c->second;
}

int Kinematics::find_mass_dimension(const GiNaC::ex& expr) const {
	try {
		ex tmp = expr.expand();
		if (is_a<numeric> (tmp)) {
			return 0;
		} else if (is_a<symbol> (tmp)) {
			symbol invariant = ex_to<symbol> (tmp);
			map<GiNaC::symbol, int>::const_iterator i = mass_dimension_.find(
					invariant);
			if (i == mass_dimension_.end())
				throw runtime_error("failed");
			return i->second;
		} else if (is_a<mul> (tmp)) {
			int d = 0;
			for (size_t i = 0; i < tmp.nops(); ++i)
				d += find_mass_dimension(tmp.op(i));
			return d;
		} else if (is_a<power> (tmp)) {
			if (!is_a<numeric> (tmp.op(1))
					|| !ex_to<numeric> (tmp.op(1)).is_integer())
				throw runtime_error("failed");
			return find_mass_dimension(tmp.op(0))
					* ex_to<numeric> (tmp.op(1)).to_int();
		} else if (is_a<add> (tmp)) {
			VERIFY(tmp.nops() > 0);
			int d = find_mass_dimension(tmp.op(0));
			for (size_t i = 1; i < tmp.nops(); ++i)
				if (find_mass_dimension(tmp.op(i)) != d)
					throw runtime_error("failed");
			return d;
		} else {
			throw runtime_error("failed");
		}
	} catch (exception& e) {
		stringstream ss;
		ss << "can't determine mass dimension of " << expr;
		throw runtime_error(ss.str());
	}
}

void Kinematics::read(const YAML::Node& node) {
	verify_yaml_spec(node);
	string s1, s2;
	int i;

	const YAML::Node& inmom_node = node["incoming_momenta"];
	for (YAML::Iterator it = inmom_node.begin(); it != inmom_node.end(); ++it) {
		(*it) >> s1;
		symbol momentum(s1);
		incoming_momenta_.append(momentum);
		external_momenta_.append(momentum);
	}
	assert_unique_symbols_and_valid_names(incoming_momenta_);

	const YAML::Node& outmom_node = node["outgoing_momenta"];
	for (YAML::Iterator it = outmom_node.begin(); it != outmom_node.end(); ++it) {
		(*it) >> s1;
		symbol momentum(s1);
		outgoing_momenta_.append(momentum);
		external_momenta_.append(momentum);
	}
	assert_unique_symbols_and_valid_names(outgoing_momenta_);
	assert_unique_symbols_and_valid_names(external_momenta_);

	if (external_momenta_.nops() > 0) {
		node["momentum_conservation"][0] >> s1;
		node["momentum_conservation"][1] >> s2;
		ex lhs(s1, external_momenta_);
		ex rhs(s2, external_momenta_);
		if (!is_a<symbol> (lhs)) {
			stringstream ss;
			ss << lhs << " is not one of " << external_momenta_;
			throw runtime_error(ss.str());
		}
		if (lhs.has(rhs))
			throw runtime_error("Momentum on left hand side appears"
				" on right hand side in momentum conservation.");
		rule_momentum_conservation_[lhs] = rhs;
		lst::const_iterator p;
		for (p = external_momenta_.begin(); p != external_momenta_.end(); ++p)
			if (*p != lhs)
				independent_external_momenta_.append(*p);

		// check momentum conservation
		ex sum = 0;
		for (p = incoming_momenta_.begin(); p != incoming_momenta_.end(); ++p)
			sum += *p;
		for (p = outgoing_momenta_.begin(); p != outgoing_momenta_.end(); ++p)
			sum -= *p;
		sum = sum.subs(rule_momentum_conservation_);
		if (!sum.expand().is_zero())
			throw runtime_error(
					"Found violation of momentum conservation for external momenta.");
	}

	if (node.FindValue("loop_momentum_prefix") != 0) {
		node["loop_momentum_prefix"] >> loop_momentum_prefix_;
	}

	const YAML::Node& inv_node = node["kinematic_invariants"];
	for (YAML::Iterator it = inv_node.begin(); it != inv_node.end(); ++it) {
		(*it)[0] >> s1;
		(*it)[1] >> i;
		symbol invariant(s1);
		kinematic_invariants_.append(invariant);
		mass_dimension_[invariant] = i;
	}
	try {
		const YAML::Node& spr_node = node["scalarproduct_rules"];
		lst eqns;
		for (YAML::Iterator it = spr_node.begin(); it != spr_node.end(); ++it) {
			//  it -> [[p1+p2, p1+p2], s]
			const YAML::Node& sp = (*it)[0];
			if (sp.size() != 2)
				throw runtime_error("Scalar product is no list of 2 elements");
			sp[0] >> s1;
			sp[1] >> s2;
			ex pl(s1, external_momenta_);
			ex pr(s2, external_momenta_);
			(*it)[1] >> s1;
			ex inv(s1, kinematic_invariants_);
			const exmap& r = rule_momentum_conservation_;
			eqns.append(ScalarProduct(pl, pr).subs(r).expand() == inv);
		}
		rules_sp_to_invariants_ = //
				lsolve_pattern(eqns, ScalarProduct(wild(1), wild(2)));
		//k.rules_invariants_to_sp_ = //
		//		lsolve_unique(eqns, k.kinematic_invariants_);
		// does not work for nonlinear relations, e.g. m instead of m^2
		for (exmap::const_iterator mit = rules_sp_to_invariants_.begin(); mit
				!= rules_sp_to_invariants_.end(); ++mit) {
			momenta_to_invariants_.add(mit->first.op(0), mit->first.op(1),
					mit->second);
			LOGX("  added rule: " << mit->first.op(0) << " * " << mit->first.op(1)
					<< " --> " << mit->second);
		}
	} catch (exception& e) {
		throw runtime_error(string("scalarproduct_rules: ") + e.what());
	}

	if (node.FindValue("dimension") != 0) {
		node["dimension"] >> s1;
		dimension_ = symbol(s1);
	}
	if (node.FindValue("symbol_to_replace_by_one") != 0) {
		symbol sym;
		node["symbol_to_replace_by_one"] >> s1;
		if (s1 == "undefined") {
			symbol_to_replace_by_one_ = GiNaC::symbol("undefined");
		} else {
			ex expr(s1, kinematic_invariants_);
			if (!is_a<symbol> (expr))
				ABORT("in symbol_to_replace_by_one: " << expr << " is not a plain symbol");
			symbol_to_replace_by_one_ = ex_to<symbol> (expr);
		}
	}

	all_symbols_ = add_lst(kinematic_invariants_, external_momenta_);
	all_symbols_.append(dimension_);
	assert_unique_symbols_and_valid_names(all_symbols_);
	kinematic_invariants_and_dimension_ = kinematic_invariants_;
	kinematic_invariants_and_dimension_.append(dimension_);

	if (node.FindValue("generate_crossings") != 0)
		node["generate_crossings"] >> generate_crossings_;
}

void Kinematics::print(YAML::Emitter& os) const {
	using namespace YAML;
	os << BeginMap;
	os << Key << "incoming_momenta" << Flow << Value << incoming_momenta_;
	os << Key << "outgoing_momenta" << Flow << Value << outgoing_momenta_;
	os << Key << "momentum_conservation" << Value << Flow
			<< rule_momentum_conservation_;
	os << Key << "kinematic_invariants" << Value;
	os << BeginSeq;
	for (lst::const_iterator it = kinematic_invariants_.begin(); it
			!= kinematic_invariants_.end(); ++it) {
		ASSERT(is_a<symbol>(*it));
		ASSERT(mass_dimension_.find(ex_to<symbol>(*it)) != mass_dimension_.end());
		os << Flow << BeginSeq << *it << mass_dimension_.find(ex_to<symbol> (
				*it))->second << EndSeq;
	}
	os << EndSeq;
	os << Key << "scalarproduct_rules" << Value << BeginSeq;
	GiNaC::exmap::const_iterator r;
	for (r = rules_sp_to_invariants_.begin(); r
			!= rules_sp_to_invariants_.end(); ++r)
		os << Flow << BeginSeq << BeginSeq << r->first.op(0) << r->first.op(1)
				<< EndSeq << r->second << EndSeq;
	os << EndSeq;

	os << Key << "dimension" << Value << dimension_;
	os << Key << "symbol_to_replace_by_one" << Value
			<< symbol_to_replace_by_one_;
	os << Key << "loop_momentum_prefix" << Value << loop_momentum_prefix_;
	os << Key << "generate_crossings" << Value << generate_crossings_;
	os << EndMap;
}

namespace {

/// finds map of crossing to lowest equivalent crossing of all ordered crossing of the kinematics
/** equivalent means the same transformation of kinematical invariants
 ** note: identity is automatically inserted at front of list */

std::map<Crossing, Crossing> find_equivalent_crossings(const Kinematics* kin,
		set<Crossing>& classes) {
	const lst invs = kin->kinematic_invariants();
	list<Crossing> all(kin->ordered_crossings());
	Crossing identity(kin);
	all.push_front(identity);
	map<Crossing, Crossing> result;
	list<Crossing>::const_reverse_iterator cfrom;
	for (cfrom = all.rbegin(); cfrom != all.rend(); ++cfrom) {
		ex frominvs = invs.subs(cfrom->rules_invariants());
		list<Crossing>::const_iterator cto;
		for (cto = all.begin(); cto != --cfrom.base(); ++cto) {
			ex toinvs = invs.subs(cto->rules_invariants());
			ex diff = (frominvs - toinvs).expand();
			if (diff.is_zero())
				break;
		}
		result.insert(make_pair(*cfrom, *cto));
		classes.insert(*cto);
		string fn = cfrom->name(), tn = cto->name();
		LOGX("Crossing equivalence: '" << fn << "' -> '" << tn << "'");
	}
	return result;
}

}

void Kinematics::set_crossings() {
	if (!generate_crossings_)
		return;

	const lst& em = external_momenta_;
	const exmap& rmc = rule_momentum_conservation_;
	const exmap& sp2inv = rules_sp_to_invariants();

	// find all legs with the same mass

	LOGX("find crossings");
	set<int> alllegs;
	map<ex, set<int> , ex_is_less> legs_by_mass;
	for (unsigned i = 0; i < em.nops(); ++i) {
		int leg = i + 1;
		ASSERT(is_a<symbol>(em.op(i)));
		ex mom = em.op(i).subs(rmc);
		ex mass = ScalarProduct(mom, mom).expand().subs(sp2inv).expand();
		LOGX("external leg " << leg << " has momenta " << mom << " and mass " << mass);
		legs_by_mass[mass].insert(leg);
		alllegs.insert(leg);
	}

	// get all permutations of the legs with the same mass

	PermutationSet pset;
	map<ex, set<int> , ex_is_less>::const_iterator m;
	for (m = legs_by_mass.begin(); m != legs_by_mass.end(); ++m) {
		vector<int> next(m->second.begin(), m->second.end());
		vector<int> from(next);
		do {
			pset.insert(Permutation(from, next));
		} while (next_permutation(next.begin(), next.end()));
	}

	// complete the set of permutations

	pset.complete_permutations();
	pset.insert(Permutation()); // otherwise unity not included
	const set<Permutation>& permutations = pset.pset();

	// set up the corresponding crossing
	for (set<Permutation>::const_reverse_iterator p = permutations.rbegin();
			p != permutations.rend(); ++p) {
		try {
			Crossing cr(this, *p);
			crossings_by_name_.insert(make_pair(cr.name(), cr));
			if (!cr.is_identity())
				ordered_crossings_.push_back(cr);
		} catch (exception& e) {
			ERROR(e.what());
		}
	}
	ordered_crossings_.sort();
	// determine equivalent crossings, need access to ordered_crossings_ via this
	equivalent_crossings_ = //
			find_equivalent_crossings(this, equivalent_crossing_classes_);
	LOGX("Number of crossings (identity included): " << ordered_crossings_.size() + 1);
}

GiNaC::exmap Kinematics::find_rules_invariants_to_sp() const {
	const exmap& sp2inv = rules_sp_to_invariants_;
	// find d.o.f. for scalar products (some invariants might be masses
	// on which the scalar products don't depend)
	lst rhs;
	for (exmap::const_iterator e = sp2inv.begin(); e != sp2inv.end(); ++e) {
		// this code could be used if we wanted the method to check correct
		// wrong mass dimensions of scalar product rules:
		// (inactive to allow numerical rules or intentional "misuses")
		/*
		if (!e->second.is_zero()) {
			try {
				int d = find_mass_dimension(e->second);
				if (d != 2)
					throw runtime_error(
							"invalid mass dimension " + to_string(d));
			} catch (exception& exc) {
				WARNING(
						"encountered potential problem with mass dimension of rule:\n"
								+ to_string(e->first) + " -> "
								+ to_string(e->second) + "\n" + exc.what());
			}
		}
		*/
		rhs.append(e->second);
	}
	exset indep; // independent d.o.f. for scalar products
	gather_objects<symbol> (rhs, indep);
	// hack to treat mass dim 1 symbols
	exmap elim_dimone;
	exmap elim_helpers;
	lst indets;
	for (exset::const_iterator i = indep.begin(); i != indep.end(); ++i) {
		try {
			if (!is_a<symbol> (*i))
				throw string("can't find mass dimension of ") + to_string(*i);
			map<symbol, int, ex_is_less>::const_iterator d = //
					mass_dimension_.find(ex_to<symbol> (*i));
			if (d == mass_dimension_.end())
				throw string("can't find mass dimension of ") + to_string(*i);
			if (d->second == 2) {
				indets.append(d->first);
			} else if (d->second == 1) {
				symbol squaredsymbol(d->first.get_name() + "squared");
				elim_helpers[squaredsymbol] = pow(d->first, 2);
				elim_dimone[pow(d->first, 2)] = squaredsymbol;
				indets.append(squaredsymbol);
			} else {
				throw "unsupported mass dimension " + to_string(d->second)
						+ " of invariant " + to_string(d->first);
			}
		} catch (string& s) {
			throw runtime_error("can't invert scalar product rules: " + s);
		}
	}
	exmap rulesraw;
	ex eqns = to_equations(sp2inv).subs(elim_dimone);
	ex sols = lsolve_overdetermined(eqns, indets);
	rulesraw = lsolve_result_to_rules(sols, indets);
	if (rulesraw.size() != indets.nops())
		throw runtime_error("can't invert scalar products rules: num mismatch");
	exmap rules;
	for (exmap::const_iterator s = rulesraw.begin(); s != rulesraw.end(); ++s) {
		ex inv = s->first.subs(elim_helpers);
		rules[inv] = s->second.subs(elim_helpers);
		if (!(inv - s->first).is_zero()) // cover mt^(-2) case
			rules[pow(inv, -1)] = pow(s->second.subs(elim_helpers), -1);
	}
	//LOGX("found inverted rules: " << rules);
	return rules;
}

} // namespace Reduze

