/*  kinematics.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#ifndef KINEMATICS_H_
#define KINEMATICS_H_

#include <ginac/ginac.h>
#include "yamlconfigurable.h"
#include "crossing.h"

namespace Reduze {

class Crossing;

/// External kinematics of the process and dimension
class Kinematics: public YAMLConfigurable {
public:
	static YAMLSpec yaml_spec() {
		YAMLSpec s;
		s.set_keyword("kinematics");
		s.set_short_description("Declaration of the kinematics.");
		s.set_long_description(s.short_description());
		s.add_option("incoming_momenta", true, "sequence of strings", ""//
					"The incoming external momenta.");
		s.add_option("outgoing_momenta", true, "sequence of strings", ""//
					"The outgoing external momenta.");
		s.add_option("momentum_conservation", false, "two-element sequence", ""//
					"The external momentum (first entry) that has to be"
					" replaced by a combination of the other external"
					" momenta (second entry).");
		s.add_option("kinematic_invariants", true, "sequence of strings", ""//
					"Each invariant must be given as a sequence of"
					" a symbol and its mass dimension.");
		s.add_option("scalarproduct_rules", true, "sequence", ""//
					"Rules how to replace scalar products of independent"
					" external momenta by kinematic invariants."
					" The first entry of such a rule is a two-element sequence"
					" (scalar product of combination of independent"
					" external momenta) and the second entry is a combination"
					" of kinematic invariants");
		s.add_option("dimension", false, "string", ""//
					"A symbol for the dimension.");
		s.add_option("symbol_to_replace_by_one", false, "string", ""//
					"A kinematic invariant that will be set to one.");
		s.add_option("loop_momentum_prefix", false, "string", ""//
					"A symbol, e.g. \"k\", from which loop momenta"
					" \"k1\", \"k2\", ... are constructed."
					" They are only used for reading diagrams and"
					" not for defining integral families.");
		s.add_option("generate_crossings", false, "boolean", ""//
					"whether possible crossings of the external legs should be"
					" determined in order to generate crossed integral families."
					" Graph based methods might fail if set to false.");
		return s;
	}
	virtual YAMLSpec yaml_spec_link() const {
		return yaml_spec();
	}
	Kinematics();
	virtual ~Kinematics();

	/// determines all possible crossing of external legs by their leg permutations in the cycle representation
	/** this function must be called explicitly to get the crossings generated (see Files::kinematics()).
	 *  Nothing is done if generate_crossings_ is set to false.*/
	void set_crossings();

	virtual void read(const YAML::Node& node);
	virtual void print(YAML::Emitter& os) const;

	/// get the external momenta
	inline const GiNaC::lst& external_momenta() const;
	/// get the independent external momenta
	inline const GiNaC::lst& independent_external_momenta() const;
	/// get the incoming momenta (lin.comb. of ext.mom.)
	inline const GiNaC::lst& incoming_momenta() const;
	/// get the outgoing momenta (lin.comb. of ext.mom.)
	inline const GiNaC::lst& outgoing_momenta() const;
	/// get the kinematic invariants
	inline const GiNaC::lst& kinematic_invariants() const;
	/// access the kinematic invariants together with the dimension
	inline const GiNaC::lst& kinematic_invariants_and_dimension() const;
	/// get the dimension
	inline const GiNaC::symbol& dimension() const;
	/// external momenta, kinematic invariants, dimension
	inline const GiNaC::lst& all_symbols() const;

	/// get external momenta and the kinematic invariants
	GiNaC::lst get_external_momenta_and_kinematics() const;

	/// get the loop momentum prefix
	const std::string& loop_momentum_prefix() const;

	const GiNaC::scalar_products& momenta_to_invariants() const {
		return momenta_to_invariants_;
	}
	const GiNaC::exmap& rules_sp_to_invariants() const {
		return rules_sp_to_invariants_;
	}

	/// crossing by name
	const Crossing& crossing(const std::string& name) const;

	/// try to invert the scalar product to invariant rules
	/** Problems arise if non-linear dependencies between scalar products
	 * and invariants are specified, simple cases with e.g. m instead of m^2
	 * should be covered. AVOID using this function if possible ! **/
	GiNaC::exmap find_rules_invariants_to_sp() const;

	const GiNaC::exmap& rule_momentum_conservation() const {
		return rule_momentum_conservation_;
	}

	const GiNaC::symbol* symbol_to_replace_by_one() const {
		if (symbol_to_replace_by_one_.get_name() == "undefined")
			return 0;
		return &symbol_to_replace_by_one_;
	}

	GiNaC::exmap get_rule_symbol_to_replace_by_one() const {
		GiNaC::exmap rule;
		if (symbol_to_replace_by_one_.get_name() != "undefined")
			rule[symbol_to_replace_by_one_] = 1;
		return rule;
	}

	int find_mass_dimension(const GiNaC::ex&) const;

	/// returns all possible ordered crossing of external legs by their leg permutations (identity crossing NOT included)
	inline const std::list<Crossing>& ordered_crossings() const {
		return ordered_crossings_;
	}
	/// minimal equivalent crossings by name of crossing (identity included)
	inline const std::map<Crossing, Crossing>& equivalent_crossings() const {
		return equivalent_crossings_;
	}
	/// minimal crossing for each equivalence class
	inline const std::set<Crossing>& equivalent_crossing_classes() const {
		return equivalent_crossing_classes_;
	}

private:
	/// external momenta
	GiNaC::lst external_momenta_, independent_external_momenta_;
	/// incoming and outgoing momenta
	GiNaC::lst incoming_momenta_, outgoing_momenta_;
	/// kinematic invariants
	GiNaC::lst kinematic_invariants_;
	/// all symbols
	GiNaC::lst all_symbols_;
	/// dimension
	GiNaC::symbol dimension_;
	/// kinematic invariants and dimension
	GiNaC::lst kinematic_invariants_and_dimension_;
	/// symbol to replace by one
	GiNaC::symbol symbol_to_replace_by_one_;
	/// mass dimensions of kinematic invariants
	std::map<GiNaC::symbol, int, GiNaC::ex_is_less> mass_dimension_;
	/// map of scalar product ScalarProduct(pi,pj) of external momenta to kinematic invariants
	GiNaC::exmap rules_sp_to_invariants_;
	///  GiNaC::scalar_products of external momenta to kinematic invariants
	GiNaC::scalar_products momenta_to_invariants_;

	/// map of kinematic invariant to scalar products pi*pj of external momenta
	GiNaC::exmap rules_invariants_to_sp_;
	/// substitution of an external momentum employing global mom. conserv.
	GiNaC::exmap rule_momentum_conservation_;

	/// prefix for not explicitly specified loop momenta // \todo remove
	std::string loop_momentum_prefix_;

	// crossings of external legs

	/// whether crossings should be generated
	bool generate_crossings_;
	/// ordered crossings without the identity crossing
	std::list<Crossing> ordered_crossings_;
	/// minimal equivalent crossings by crossing (identity included)
	std::map<Crossing, Crossing> equivalent_crossings_;
	/// equivalence classes of crossings
	std::set<Crossing> equivalent_crossing_classes_;
	/// crossings by name
	std::map<std::string, Crossing> crossings_by_name_;
};

inline void operator>>(const YAML::Node& n, Kinematics& r) {
	r.read(n);
}
inline YAML::Emitter& operator<<(YAML::Emitter& ye, const Kinematics& r) {
	r.print(ye);
	return ye;
}

const GiNaC::lst& Kinematics::external_momenta() const {
	return external_momenta_;
}
const GiNaC::lst& Kinematics::independent_external_momenta() const {
	return independent_external_momenta_;
}
const GiNaC::lst& Kinematics::incoming_momenta() const {
	return incoming_momenta_;
}
const GiNaC::lst& Kinematics::outgoing_momenta() const {
	return outgoing_momenta_;
}
const GiNaC::lst& Kinematics::kinematic_invariants() const {
	return kinematic_invariants_;
}
const GiNaC::lst& Kinematics::kinematic_invariants_and_dimension() const {
	return kinematic_invariants_and_dimension_;
}
const GiNaC::symbol& Kinematics::dimension() const {
	return dimension_;
}
const GiNaC::lst& Kinematics::all_symbols() const {
	return all_symbols_;
}

} // namespace Reduze

#endif /* KINEMATICS_H_ */
