/*  sector.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "sector.h"
#include <string>
#include <fstream>
#include <exception>
#include "kinematics.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "functions.h"
#include "files.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "graph.h"
#include "int.h"

using namespace std;

namespace Reduze {

// register type
namespace {
YAMLProxy<Sector> dummy1;
YAMLProxy<SectorSelection> dummy2;
}

// private default constructor
Sector::Sector() :
	integralfamily_(0), id_(0), t_(0) {
	//integralfamily_ = new IntegralFamily();
}

Sector::Sector(const YAML::Node& node) {
	if (node.GetType() != YAML::CT_SEQUENCE || node.size() != 2)
		throw std::runtime_error("sector spec is not a two-element sequence "
				+ position_info(node));
	std::string family_name;
	node[0] >> family_name;
	try {
		integralfamily_ = Files::instance()->integralfamily(family_name);
	} catch (std::exception& e) {
		throw std::runtime_error("sector spec has unknown integral family "
				+ position_info(node[0]) + "\n" + e.what());
	}
	if (node[1].GetType() == YAML::CT_SCALAR) {
		node[1] >> id_;
	} else {
		throw std::runtime_error("unknown sector spec");
	}

	t_ = t_of_id(id_);
}

void Sector::read(const YAML::Node& node) {
	if (node.GetType() == YAML::CT_SCALAR) {
		int id;
		node >> id;
		*this = Sector(integralfamily_, id);
	} else {
		*this = Sector(node);
	}
}

Sector::Sector(const IntegralFamily* ic, int id) :
	integralfamily_(ic), id_(id), t_(Sector::t_of_id(id)) {
	ASSERT(integralfamily_ != 0);
}

Sector::Sector(const Sector& s1, const Sector& s2) {
	const IntegralFamily* ic1 = s1.integralfamily();
	const IntegralFamily* ic2 = s2.integralfamily();

	Files* f = Files::instance();
	const IntegralFamily* ic = f->integralfamily(ic1, ic2);

	int id1 = s1.id();
	int id2 = s2.id();
	int max = ic1->max_sector_id();

	integralfamily_ = ic;
	id_ = id1 + (max + 1) * id2;
	t_ = Sector::t_of_id(id_);
}

std::string Sector::get_safe_string_for_filename(bool fill_digits) const {
	ostringstream ss;
	if (!fill_digits) {
		ss << integralfamily_->name() << "_" << t_ << "_" << id_;
	} else {
		int maxid = integralfamily_->max_sector_id();
		ss << integralfamily_->name() << "_";
		ss << setfill('0') << setw(number_of_digits(t_of_id(maxid))) << t_;
		ss << "_";
		ss << setfill('0') << setw(number_of_digits(maxid)) << id_;
	}
	return safe_filename(ss.str());
}

bool Sector::is_obviously_zero() const {
	return (static_cast<size_t> (t_) < integralfamily_->loop_momenta().nops());
}

bool Sector::is_obviously_non_zero() const {
	return integralfamily_->loop_momenta().nops() == 0; // id = 0
}

bool Sector::operator<(const Sector& other) const {
	if (*integralfamily_->source_integralfamily()
			!= *other.integralfamily_->source_integralfamily())
		return (*integralfamily_->source_integralfamily()
				< *other.integralfamily_->source_integralfamily());
	if (t_ != other.t_)
		return (t_ < other.t_);
	if (id_ != other.id_) {
		if (smaller_sector_id_is_less_)
			return (id_ < other.id_);
		else
			return (id_ > other.id_);
	}
	return (*integralfamily_ < *other.integralfamily_);
}

std::ostream& operator<<(std::ostream& os, const Sector& s) {
	return os << "(" << s.integralfamily_->name() << ":" << s.t_ << ":"
			<< s.id_ << ")";
}

void Sector::print(YAML::Emitter& os) const {
	os << YAML::Flow << YAML::BeginSeq//
			<< (integralfamily_ ? integralfamily_->name() : "") //
			<< id_ //
			<< YAML::EndSeq;
}

int Sector::t_of_id(int sector_id) {
	unsigned t = 0;
	for (unsigned i = sector_id; i != 0; i = i >> 1)
		if (i & 1)
			t++;
	return t;
}

Sector Sector::get_uncrossed_sector() const {
	return Sector(integralfamily_->source_integralfamily(), id_);
}

namespace {
GiNaC::ex get_U_polynomial(const std::vector<Propagator>& props,
		const GiNaC::lst& loopmom, GiNaC::lst& symbs) {
	using namespace GiNaC;
	symbs = create_symbols("x", props.size());
	ex t0(0);
	for (unsigned i = 0; i < props.size(); ++i)
		t0 += pow(props[i].to_explicite_form(), -1) * symbs.op(i);
	t0 = t0.subs(ScalarProduct(wild(1), wild(2)) == wild(1) * wild(2));
	t0 = t0.expand();
	matrix mat(loopmom.nops(), loopmom.nops());
	for (unsigned i = 0; i < loopmom.nops(); ++i) {
		for (unsigned j = i; j < loopmom.nops(); ++j) {
			if (i != j) {
				ex e = t0.coeff(loopmom.op(i), 1).coeff(loopmom.op(j), 1);
				mat(i, j) = numeric(1, 2) * e;
				mat(j, i) = mat(i, j);
			} else {
				mat(i, j) = t0.coeff(loopmom.op(i), 2);
			}
		}
	}
	if (loopmom.nops() == 0)
		return 1;
	// GiNaC's determinant computation is much faster than Fermat's
	return mat.determinant(determinant_algo::laplace);
}

/*
 * A U-polynomial of a given graph is the sum over all spanning trees
 * where each term is a product of the lines x[i] not contained in the tree.
 *
 * A polynomial obtained by the determinant as in get_U_polynomial(...) is
 * a possible graph U-polynomial if:
 *
 * - it is not zero
 * - each term has exactly #loops different factors x[i]
 * - the number of terms is smaller or equal than binomial(#edges, #loops)
 * - every term has the same numerical coefficient
 *
 * NB:
 * Two possible U-polynomials, which differ by a global numerical factor,
 * can correspond to the same graph but with different momentum mappings
 * which can only be identified by a transformation of the loop momenta
 * with determinant not equal to one.
 *
 */

void assert_possible_graph_U_polynomial(const GiNaC::ex& u, unsigned int loops,
		unsigned int edges) {
	using namespace GiNaC;
	string messsage("U-polynomial doesn't correspond to a graph: ");
	ex u_expanded = u.expand();
	if (u_expanded.is_zero())
		throw runtime_error(messsage + "is zero.");
	ASSERT(edges >= loops);
	ex terms = is_a<add> (u_expanded) ? u_expanded : lst(u_expanded);
	if (terms.nops() > binomial(edges, loops))
		throw runtime_error(messsage + "too many terms.");
	exset num_factors;
	for (const_iterator t = terms.begin(); t != terms.end(); ++t) {
		ex fac = is_a<mul> (*t) ? *t : lst(*t);
		exset symb_factors;
		ex num(1);
		for (const_iterator f = fac.begin(); f != fac.end(); ++f) {
			if (is_a<numeric> (*f))
				num *= *f;
			else
				symb_factors.insert(*f);
		}
		num_factors.insert(num);
		if (symb_factors.size() != loops)
			throw runtime_error(messsage
					+ "term with wrong number of lines detected.");
	}
	if (num_factors.size() != 1) // several numeric coefficients in terms
		throw runtime_error(messsage + "several numerical coefficients.");
}

// returns the edges numbers (a set for each connected component) of lines
// not contained in the spanning tree from the U-polynomial 'u'
// the line numbers are the position of the symbols in 'symbs'
// selects lexicographic simplest term from all the possible terms in the U-polynomial
std::set<std::set<int> > get_fundamental_cycles(const GiNaC::ex u,
		const GiNaC::lst& symbs, unsigned loops) {
	using namespace GiNaC;
	set<set<int> > res;
	try {
		assert_possible_graph_U_polynomial(u, loops, symbs.nops()); // may throw
	} catch (exception& e) {
		throw runtime_error(e.what());
	}

	exmap symbs2int;
	int lauf(0);
	for (lst::const_iterator s = symbs.begin(); s != symbs.end(); ++s)
		symbs2int[*s] = lauf++;

	ex u_factor = factor(u);
	u_factor = is_a<mul> (u_factor) ? u_factor : lst(u_factor);
	for (const_iterator f = u_factor.begin(); f != u_factor.end(); ++f) {
		if (is_a<numeric> (*f))
			continue;
		ex f_expanded = f->expand();
		f_expanded = is_a<add> (f_expanded) ? f_expanded : lst(f_expanded);
		set<set<int> > sorted;
		for (const_iterator e = f_expanded.begin(); e != f_expanded.end(); ++e) {
			set<int> tmp;
			for (lst::const_iterator s = symbs.begin(); s != symbs.end(); ++s)
				if (e->has(*s))
					tmp.insert(ex_to<numeric> (symbs2int[*s]).to_int());
			sorted.insert(tmp);
		}
		ASSERT(!sorted.empty());
		res.insert(*sorted.begin());
	}

	unsigned count(0);
	for (set<set<int> >::const_iterator s = res.begin(); s != res.end(); ++s)
		count += s->size();
	ASSERT(count == loops);
	return res;
}

} // unnamed namespace

GiNaC::ex Sector::get_U_polynomial(GiNaC::lst& symbs) const {
	using namespace GiNaC;
	vector<Propagator> props = get_propagators();
	const lst& loopmom = integralfamily_->loop_momenta();
	const lst& extmom = integralfamily_->kinematics()->external_momenta();
	lst zeros;
	for (unsigned i = 0; i < extmom.nops(); ++i)
		zeros.append(0);
	exmap ext2zero = make_translation(extmom, zeros);
	// set masses and external moment< to zero
	for (unsigned i = 0; i < props.size(); ++i) {
		props[i].let_op(0) = props[i].let_op(0).subs(ext2zero,
				GiNaC::subs_options::no_pattern);
		props[i].let_op(1) = props[i].let_op(1).subs(ext2zero,
				GiNaC::subs_options::no_pattern);
		props[i].let_op(2) = 0;
	}
	return Reduze::get_U_polynomial(props, loopmom, symbs);
}

bool Sector::search_graph(Graph& graph, bool& det_non_zero,
		bool construct_minimal_graphs) const {
	using namespace GiNaC;
	const vector<Propagator> props = get_propagators();
	const GiNaC::lst loopmom = integralfamily_->loop_momenta();

	// get fundamental cycles


	try {
		// throws if a propagator doesn't allow access to momentum
		for (vector<Propagator>::const_iterator p = props.begin(); p
				!= props.end(); ++p)
			p->momentum();
	} catch (exception& e) {
		det_non_zero = true;
		return false;
	}

	pair<ex, set<int> > dc = independent_propagators(props, loopmom);

	// det found to be zero -> no graph and zero
	if (dc.first.is_zero()) {
		det_non_zero = false;
		return false;
	}
	det_non_zero = true;

	const Kinematics* kin = integralfamily_->kinematics();
	const set<int>& cycles = dc.second;
	const exmap& mom_cons = kin->rule_momentum_conservation();
	ex in_mom = kin->incoming_momenta().subs(mom_cons);
	ex out_mom = kin->outgoing_momenta().subs(mom_cons);
	const exmap& sp2inv = kin->rules_sp_to_invariants();

	const unsigned num_in = in_mom.nops();
	const unsigned num_out = out_mom.nops();
	const unsigned num_ext = num_in + num_out;

	// propagators with edge id > num_ext
	vector<pair<int, GraphEdge> > edges_tree, edges_cycle;
	ASSERT(loopmom.nops() <= props.size());
	edges_cycle.reserve(loopmom.nops());
	edges_tree.reserve(props.size() - loopmom.nops());
	int cycle_graph_sector_id = 0;
	for (unsigned p = 0; p < props.size(); ++p) {
		GraphEdge gedge(props[p].momentum(), props[p].squaredmass());
		if (cycles.find(p) != cycles.end()) {
			cycle_graph_sector_id += power_int(2, p);
			edges_cycle.push_back(make_pair(num_ext + p + 1, gedge));
		} else {
			edges_tree.push_back(make_pair(num_ext + p + 1, gedge));
		}
	}
	ASSERT(edges_tree.size() == props.size() - loopmom.nops());

	// set up the graph with the fundamental cycles and external legs
	Graph g;
	g.set_loop_momenta(loopmom);
	g.set_name(get_safe_string_for_filename());

	const int root_node = 1;
	int next_dummy_edge = num_ext + props.size() + 1;
	int next_dummy_node = 2 * props.size() + 2 /*- loopmom.nops()*/; // internal + cc - loop

	// attach the fundamental edges to the root_node
	// add them twice, additional degree 2 node will be removed at the end
	vector<int> edges1, edges2;
	edges1.reserve(2 * edges_cycle.size());
	edges2.reserve(2 * edges_cycle.size());
	for (unsigned int i = 0; i < edges_cycle.size(); ++i) {
		int from = root_node;
		int to = next_dummy_node++;
		int edge_id = edges_cycle[i].first;
		edges1.push_back(g.insert_edge(Edge(from, to, edge_id)));
		edges2.push_back(g.insert_edge(Edge(to, from, next_dummy_edge++)));
	}
	for (unsigned int i = 0; i < edges_cycle.size(); ++i) {
		g.set_momentum(edges1[i], edges_cycle[i].second.momentum_);
		g.set_momentum(edges2[i], edges_cycle[i].second.momentum_);
		g.set_squaredmass(edges1[i], edges_cycle[i].second.squaredmass_);
		g.set_squaredmass(edges2[i], edges_cycle[i].second.squaredmass_);
	}

	// attach external legs to the root_node

	vector<int> edges_in;
	edges_in.reserve(num_in);
	for (unsigned int i = 1; i <= num_in; ++i)
		edges_in.push_back(g.insert_edge(Edge(-i, root_node, i)));

	vector<int> edges_out;
	edges_out.reserve(num_out);
	for (unsigned int i = 1; i <= num_out; ++i)
		edges_out.push_back(g.insert_edge(Edge(root_node, -i - num_in, num_in
				+ i)));

	for (unsigned int i = 0; i < num_in; ++i) {
		g.set_momentum(edges_in[i], in_mom.op(i));
		ex sqm = ScalarProduct(in_mom.op(i), in_mom.op(i));
		sqm = sqm.expand().subs(sp2inv).expand();
		g.set_squaredmass(edges_in[i], sqm);
	}
	for (unsigned int i = 0; i < num_out; ++i) {
		g.set_momentum(edges_out[i], out_mom.op(i));
		ex sqm = ScalarProduct(out_mom.op(i), out_mom.op(i));
		sqm = sqm.expand().subs(sp2inv).expand();
		g.set_squaredmass(edges_out[i], sqm);
	}

	// add the edges from the spanning tree to the graph

	try {
		if (construct_minimal_graphs)
			g.insert_graph_edges_unique(edges_tree);
		else
			g.insert_graph_edges(edges_tree);
	} catch (exception& e) {
		return false;
	}

	// clean up

	if (!construct_minimal_graphs) {
		g.remove_degree_2_nodes();
		g.disconnect_vacuum_components();
	}

	graph.swap(g);
	return true;
}

std::vector<Propagator> Sector::get_propagators() const {
	vector<Propagator> allprops = integralfamily_->propagators();
	vector<Propagator> p;
	int tt = t();
	p.reserve(tt);
	for (int j = 0, i = id_; i != 0; ++j, i = i >> 1)
		if (i & 1)
			p.push_back(allprops[j]);
	return p;
}

long int Sector::get_num_integrals_static(int n, int t, int r, int s) {
	long int res;
	if (n > t)
		res = binom_int(r - 1, r - t) * binom_int(n - 1 - t + s, s);
	else
		res = binom_int(r - 1, r - t);
	return res;
}

long int Sector::get_num_integrals(int r, int s) const {
	size_t nprops = integralfamily_->num_propagators();
	return get_num_integrals_static(nprops, t_, r, s);
}

long int Sector::get_num_integrals(int r_min, int r_max, int s_min, int s_max) const {
	long int res = 0;
	for (int i = r_min; i <= r_max; ++i)
		for (int j = s_min; j <= s_max; ++j)
			res += get_num_integrals(i, j);
	return res;
}

// SectorSelection

SectorSelection SectorSelection::find_compact_recursive_selection(
		const std::set<Sector>& secs, bool use_uncrossed) {
	typedef set<Sector, greater<Sector> > secset;
	secset needed, have;
	for (set<Sector>::const_iterator s = secs.begin(); s != secs.end(); ++s) {
		Sector snox = s->get_uncrossed_sector();
		Sector smin = SectorMappings::get_shifted_sector(snox, true);
		if (!SectorMappings::is_zero(smin))
			needed.insert(smin);
	}
	for (secset::const_iterator s = needed.begin(); s != needed.end(); ++s) {
		set<Sector> sub = find_prerequisite_sectors(*s, true);
		for (set<Sector>::const_iterator h = sub.begin(); h != sub.end(); ++h)
			needed.erase(*h);
	}
	SectorSelection sectree;
	sectree.select_recursively_.insert(sectree.select_recursively_.end(),
			needed.begin(), needed.end());
	return sectree;
}

void SectorSelection::read(const YAML::Node& node) {
	verify_yaml_spec(node);
	Sector* pro = 0;
	if (node.FindValue("select"))
		read_proto_ctor(node["select"], select_, pro);
	if (node.FindValue("select_recursively"))
		read_proto_ctor(node["select_recursively"], select_recursively_, pro);
	if (node.FindValue("deselect"))
		read_proto_ctor(node["deselect"], deselect_, pro);
	if (node.FindValue("deselect_recursively"))
		read_proto_ctor(node["deselect_recursively"], deselect_recursively_,
				pro);
	delete pro;
	if (node.FindValue("t_restriction")) {
		node["t_restriction"][0] >> t_min_;
		node["t_restriction"][1] >> t_max_;
	}
}

void SectorSelection::print(YAML::Emitter& os) const {
	using namespace YAML;
	os << BeginMap;
	os << Key << "select" << Value << select_;
	os << Key << "select_recursively" << Value << select_recursively_;
	os << Key << "deselect" << Value << deselect_;
	os << Key << "deselect_recursively" << Value << deselect_recursively_;
	os << Key << "t_restriction" << Value << Flow << BeginSeq << t_min_
			<< t_max_ << EndSeq;
	os << EndMap;
}

/// adds a sector
void SectorSelection::add(const Sector& s) {
	select_.push_back(s);
}

/// adds a sector and prerequisite sectors (subsectors)
void SectorSelection::add_recursively(const Sector& s) {
	select_recursively_.push_back(s);
}

std::set<Sector> SectorSelection::find_prerequisite_sectors(const Sector& sin,
		bool recurse) {
	set<Sector> res; // result
	set<Sector> visited; // needed for a->b->c shifts (see also rev 1410)
	stack<Sector> todo;
	Sector next = sin;
	do {
		// collect first level sub sectors (support a->b->c shift chains)
		// each shift can introduce sub sectors of (intermediate) shift target
		Sector n = next;
		do {
			n = next;
			visited.insert(n);
			const IntegralFamily* f = n.integralfamily();
			set<int> ids = f->get_subsector_equivalents(n.id(), false);
			for (set<int>::const_iterator i = ids.begin(); i != ids.end(); ++i) {
				Sector sub = Sector(f, *i);
				Sector s = SectorMappings::get_shifted_sector(sub, false);
				if (!s.is_obviously_zero() && visited.find(s) == visited.end()) {
					res.insert(SectorMappings::get_shifted_sector(s, true));
					todo.push(sub); // just first shift applied
				}
			}
			next = SectorMappings::get_shifted_sector(n, false);
		} while (next != n);
		if (n.integralfamily()->is_crossed()) {
			Sector nox = n.get_uncrossed_sector();
			if (!nox.is_obviously_zero() && visited.find(nox) == visited.end()) {
				res.insert(nox);
				todo.push(nox);
			}
		}
	} while (recurse && !todo.empty() && (next = todo.top(), todo.pop(), true));
	return res;
}

set<Sector> SectorSelection::find_sectors() const {
	set<Sector> sectors;
	list<Sector>::const_iterator s;
	for (s = select_.begin(); s != select_.end(); ++s) {
		Sector equiv = SectorMappings::get_shifted_sector(*s);
		if (SectorMappings::is_zero(equiv)) {
			LOG("Omitting zero sector " << *s << " from selection");
		} else if (equiv != *s) {
			LOG("Selecting " << equiv << " instead of " << *s);
			sectors.insert(equiv);
		} else {
			sectors.insert(*s);
		}
	}
	for (s = select_recursively_.begin(); s != select_recursively_.end(); ++s) {
		Sector equiv = SectorMappings::get_shifted_sector(*s);
		if (SectorMappings::is_zero(equiv)) {
			LOG("Omitting zero sector " << *s << " from selection");
		} else if (equiv != *s) {
			LOG("Selecting " << equiv << " instead of " << *s);
			sectors.insert(equiv);
		} else {
			sectors.insert(*s);
		}
		set<Sector> sub = find_prerequisite_sectors(*s, true);
		sectors.insert(sub.begin(), sub.end());
		//for (set<Sector>::const_iterator s = sub.begin(); s != sub.end(); ++s)
		//	LOGX("found " << *s << endl);
	}
	for (s = deselect_.begin(); s != deselect_.end(); ++s) {
		sectors.erase(SectorMappings::get_shifted_sector(*s));
	}
	for (s = deselect_recursively_.begin(); s != deselect_recursively_.end(); ++s) {
		sectors.erase(SectorMappings::get_shifted_sector(*s));
		set<Sector> sub = find_prerequisite_sectors(*s, true);
		for (set<Sector>::const_iterator d = sub.begin(); d != sub.end() ; ++d)
			sectors.erase(*d);
	}
	set<Sector> filtered;
	set<Sector>::const_iterator ss;
	for (ss = sectors.begin(); ss != sectors.end(); ++ss) {
		if (!((t_min_ >= 0 && ss->t() < t_min_) || //
				(t_max_ >= 0 && ss->t() > t_max_)))
			filtered.insert(*ss);
	}
	return filtered;
}

// global functions


std::pair<GiNaC::ex, std::set<int> > independent_propagators(const std::vector<
		Propagator>& props, const GiNaC::lst& loopmoms) {

	if (loopmoms.nops() == 0)
		return make_pair(1, set<int> ());

	vector<set<int> > props_by_mom;
	props_by_mom.reserve(loopmoms.nops());
	for (GiNaC::lst::const_iterator l = loopmoms.begin(); l != loopmoms.end(); ++l) {
		ASSERT(GiNaC::is_a<GiNaC::symbol>(*l));
		props_by_mom .push_back(set<int> ());
		for (unsigned i = 0; i < props.size(); ++i) {
			GiNaC::ex momentum(0);
			try {
				momentum = props[i].momentum();
			} catch (exception&) {
				continue;
			}
			if (momentum.has(*l))
				props_by_mom.back().insert(i);
		}
	}

	GiNaC::ex det(0);

	combinations<int> combs(props_by_mom);
	vector<int> r;
	while (combs.get_next(r)) {
		set<int> res(r.begin(), r.end());
		if (res.size() != loopmoms.nops())
			continue;

		GiNaC::matrix mat(loopmoms.nops(), loopmoms.nops());
		for (unsigned i = 0; i < loopmoms.nops(); ++i)
			for (unsigned j = 0; j < loopmoms.nops(); ++j)
				mat(i, j) = props[r[i]].momentum().diff(GiNaC::ex_to<
						GiNaC::symbol>(loopmoms.op(j)));

		det = mat.determinant(GiNaC::determinant_algo::laplace);
		if (!det.is_zero())
			return make_pair(det, res);
	}

	return make_pair(det, set<int> ());
}

}
// namespace Reduze

